package no.uio.ifi.rcos.algebraic.createmodel

import org.scalatest.FlatSpec
import org.scalatest.matchers.ShouldMatchers
import no.uio.ifi.rcos.algebraic.spec.CreateModel
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel
import org.scalatest.BeforeAndAfter
import org.eclipse.uml2.uml.Type
import edu.unu.iist.rcos.spec.antlr.MyDesign
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.TYPE
import org.eclipse.uml2.uml.Property
import sun.security.action.GetBooleanAction
import edu.unu.iist.rcos.spec.antlr.RCOSModelCheckException

class BasicCompositonsTests extends FlatSpec with ShouldMatchers with BeforeAndAfter {

  val modelName = "basic_compositions.uml"
  val fileName = "testcases/modelcreator/basic_compositions.rcosp"
  var createModel : CreateModel = _
    
  before {
    createModel = new CreateModel
    createModel.createModelFromFileSpec("basic_compositions.uml", "basic_compositions", new MyRCOSModel.RCOSFromFile(fileName))
  }
  

  "Creating a model of modelcreator/basic_compositions.rcosp" should "create all basic interfaces" in {
    createModel.model.getInterface("A", null, null) should not be (null)
    createModel.model.getInterface("B", null, null) should not be (null)
    createModel.model.getInterface("Foo", null, null) should not be (null)
    createModel.model.getInterface("V", null, null) should not be (null)
  }

  "Creating a model of modelcreator/basic_compositions.rcosp" should "create all basic components" in {
    createModel.model.getComponent("C", null, null) should not be (null)
    createModel.model.getComponent("D", null, null) should not be (null)
  }

  "Interface A" should "have a field called 'k'" in {

	val t : TYPE = MyDesign.createInt
	val interface = createModel.model.getInterface("A", null, null)
    val prop : Property = interface.getAttribute("k", createModel.model.getTypeFromTYPE(t, interface))
    prop should not be (null)
  }

  "Component C" should "have a field called 'myInt'" in {

	val t : TYPE = MyDesign.createInt
	val component = createModel.model.getComponent("C", null, null)
    val prop : Property = component.getAttribute("myInt", createModel.model.getTypeFromTYPE(t, component))
    prop should not be (null)
  }
  
  "A field in a components provided interface" should " be added to the component " in {
    //Placeholder
    2 should be (2)
  }
  
  "Component D" should "require interface A " in {
	val component = createModel.model.getComponent("D", null, null)
	val interface = component.getRequired("A")
	interface should not be (null)
  }
  
  "Component C" should "provide interface A and B" in {
	val component = createModel.model.getComponent("C", null, null)
	val interfaceA = component.getProvided("A")
	val interfaceB = component.getProvided("B")
	interfaceA should not be (null)
	interfaceB should not be (null)
  }

  "Component C" should "not provide Foo" in {
	val component = createModel.model.getComponent("C", null, null)
	val interfaceFoo = component.getProvided("Foo")
	interfaceFoo should be (null)
  }
  
  "Component E" should "be a component" in {
	val component = createModel.model.getComponent("E", null, null)
	component should not be (null)
  }
  
  "Component E" should "be a parallel composition of component C and D" in {
    val componentE = createModel.model.getComponent("E", null, null)
    componentE.getAppliedStereotype("rCOS::Parallel") should not be (null)
  }

  "Component F" should "be a plugging composition of component C and D" in {
    val componentE = createModel.model.getComponent("F", null, null)
    componentE.getAppliedStereotype("rCOS::Plugging") should not be (null)
  }

  "Component G" should "be a disjoint composition of component X and Y" in {
    val componentE = createModel.model.getComponent("G", null, null)
    componentE.getAppliedStereotype("rCOS::Disjoint") should not be (null)
  }

  "Component XYZ" should "be a parallel composition of component (X o Z) and Y" in {
    val componentE = createModel.model.getComponent("XYZ", null, null)
    componentE.getAppliedStereotype("rCOS::Parallel") should not be (null)
  }

  "if component doesnt exist we " should "get RCOSModelCheckedException " in {
    var name = "D_parallel_L"
    intercept[RCOSModelCheckException] {
      createModel.model.getComponent(name, null, null)
    }
  }

  "Component 'Restricted' " should " be a restriced composition of component C" in {
    var comp  = createModel.model.getComponent("Restricted", null, null)
    comp.getAppliedStereotype("rCOS::Hiding") should not be (null)
  }

  "Component 'Renamed' " should " be a renamed composition of component C" in {
    var comp  = createModel.model.getComponent("Renamed", null, null)
    comp.getAppliedStereotype("rCOS::Renaming") should not be (null)
  }

  "The plugging component created in F = C << D " should "be properly renamed to F" in {
    var compF  = createModel.model.getComponent("F", null, null)
    compF should not be (null)
    intercept[RCOSModelCheckException] {
      createModel.model.getComponent("C_plugged_D", null, null)
    }
  }

  "Context " should "be correctly set in expression in expression XYZ3 = X || Y || Z" in {
    var xyz3 = createModel.model.getComponent("XYZ3", null, null)
    var yparaz  = createModel.model.getComponent("Y_parallel_Z", xyz3, null)
    yparaz should not be (null)
  }

  "Context " should "be correctly set in expression in expression XHiddenY = X \\ {a, b} || Y; " in {
    var xhiddeny = createModel.model.getComponent("XHiddenY", null, null)
    var xhidden  = createModel.model.getComponent("X_restricted", xhiddeny, null)
    xhidden should not be (null)
  }
  
  "The expression FE = F || E " should " be a parallel composition of two non-atomic components E and F" in {
    
    var fe = createModel.model.getComponent("FE", null, null)
    fe.getAppliedStereotype("rCOS::Parallel") should not be (null)

    var e = createModel.model.getComponent("E", null, null)
    var f = createModel.model.getComponent("F", null, null)
    e.getAppliedStereotype("rCOS::Parallel") should not be (null)
    f.getAppliedStereotype("rCOS::Plugging") should not be (null)
  }
}