package no.uio.ifi.rcos.algebraic.extractors

import org.eclipse.uml2.uml.Operation
import org.eclipse.uml2.uml.Classifier
import edu.unu.iist.rcos.mm.util.Utility
import org.eclipse.uml2.uml.Interface
import org.eclipse.uml2.uml.Component
import scala.collection.JavaConversions._
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel
import org.eclipse.uml2.uml.Constraint
import edu.unu.iist.rcos.mm.util.RCOSRuleFeature
import org.eclipse.uml2.uml.OpaqueExpression
import org.eclipse.uml2.uml.UMLPackage

trait CopyOperation {
   /**
   * copy an operation from one classifier to another
   * @author martinhb
   */
  def extractAndCopyOperation(to : Classifier, operation: Operation, name: String, model: MyRCOSModel) : Operation = {

    var op: Operation = null

    //no common supertype for component and interface except Classifer(?), 
    //but classifier doesnt offer createOwnedOperation, so we need casting. argh.
    if (to.isInstanceOf[Interface]) {
	  op = Utility.addMethod(name,
			  operation.getVisibility(),
			  to.asInstanceOf[Interface], 
			  operation.isQuery())
    }
    else { //assume component...
	  op = Utility.addMethod(name,
			  operation.getVisibility(),
			  to.asInstanceOf[Component], 
			  operation.isQuery())
    }
      
    val params = operation.getOwnedParameters()
    for (param <- params) {
      model.addPara(param.getName(), param.getType(), op, param.getDirection())
    }

    //last thing - operation bodies (I hope this is the correct way)
    val rules = operation getOwnedRules

    for (r <- rules) {
		val c: Constraint = op.createOwnedRule("design"); //hardcoded for getRCOS_RULE_ID()
		val ope: OpaqueExpression = c.createSpecification(null, null, UMLPackage.Literals.OPAQUE_EXPRESSION).asInstanceOf[OpaqueExpression]
		ope.getBodies().add(r.getSpecification().stringValue());
    }

    op
  }
}