package no.uio.ifi.rcos.algebraic.spec

import org.antlr.runtime.ANTLRStringStream
import no.uio.ifi.rcos.algebraic.parser.new_rcospLexer
import org.antlr.runtime.CommonTokenStream
import no.uio.ifi.rcos.algebraic.parser.new_rcospParser
import no.uio.ifi.rcos.algebraic.utility.Parse
import org.antlr.runtime.tree.CommonTreeNodeStream
import no.uio.ifi.rcos.algebraic.parser.new_rcospTreeGetClasses
import org.eclipse.emf.common.util.URI
import no.uio.ifi.rcos.algebraic.parser.new_rCOSTreeToUML
import no.uio.ifi.rcos.algebraic.typechecking.AtomicTypeChecker
import no.uio.ifi.rcos.algebraic.typechecking.CompositionExpressionTypeChecker
import org.antlr.runtime.ANTLRInputStream

/**
 *
 * @author martinhb
 */
class CreateModel {
  
  private var theModel : MyExtendedRCOSModel = _
  def model = theModel

  def createModelFromFileSpec(fileName : String, modelName : String, spec : ANTLRStringStream) : Unit = {
    val tree = Parse.getCommonTree(spec)
    createModel(fileName, modelName, tree)
  }

  def createModelFromStream(fileName : String, modelName : String, spec : ANTLRInputStream) : Unit = {
    val tree = Parse.getCommonTree(spec)
    createModel(fileName, modelName, tree)
  }
  
  def createModel(fileName: String, modelName: String, tree: org.antlr.runtime.tree.CommonTree) {

    theModel = new MyExtendedRCOSModel(URI.createFileURI(fileName), modelName)
	var nodes = new CommonTreeNodeStream(tree)
    //Get basic classes (interfaces, components, classes) and add them to the model (interfaces, components, classes)
    val treeWalker = new new_rcospTreeGetClasses(nodes) 
    treeWalker.m = model
    	treeWalker.rCOSSpec()

    	//create uml
    	nodes = new CommonTreeNodeStream(tree)
    	val walker2 = new new_rCOSTreeToUML(nodes)
    	walker2.m = model
    	walker2.typechecker = new AtomicTypeChecker
    	walker2.rCOSSpec()

	}
  
  def save() { model.save }
}