package no.uio.ifi.rcos.algebraic.utility

import org.eclipse.uml2.uml.Operation
import scala.collection.JavaConversions._
import scala.collection.mutable.Set
import scala.collection.mutable.Buffer
/**
 * Custom mutable Set implementation for UML operations
 * @author martinhb
 */
class OperationSet(operations: Buffer[Operation]) extends Set[Operation] with OperationComparator {

  override def contains(key: Operation) : Boolean = {
	for (op <- operations) {
	   if (equalOperations(key, op)) 
	     return true
	}
    false
  } 

  def iterator: Iterator[Operation] = {operations.iterator}
  def +=(elem: Operation) = { 
    if (!contains(elem)) {
      operations += (elem);OperationSet.this 
    } 
    else OperationSet.this  } //Hope this works...

  def -=(elem: Operation) = {operations -= elem; OperationSet.this} 
  
  //diff - so we can compute req/provided sets where this is needed. Destructive operation
  def diff(that: OperationSet) : OperationSet = {
   
    //manually looping to avoid shady NPE caused when removing during iteration
    var pos = 0
    while (pos < operations.size) {
      val op = operations(pos)
      if (that.contains(op)) {
         operations-=(op)
      }
      else {
        pos+=1
      }
    }
    this
  }
  
  def intersect(that: OperationSet) : OperationSet = {

    val set = new OperationSet(Buffer.empty[Operation]) 
    for (op <- operations) {
      if (that.contains(op))
        set += op //NB! add the same operation to the new set - not a copy. 
    }
    set
  }
}