package edu.unu.iist.rcos.spec.antlr;

import static org.junit.Assert.*;

import org.eclipse.emf.common.util.URI;
import org.eclipse.uml2.uml.Operation;
import org.junit.Test;

import edu.unu.iist.rcos.spec.antlr.visitor.expprefix.DesignPrefixFilterVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.expprefix.PrefixFilterException;
import edu.unu.iist.rcos.uml.EmptyModel;

public class Test_DesignPrefixFilter {

	final DesignPrefixFilterVisitor v;
	final EmptyModel m;
	final Operation op;

	public Test_DesignPrefixFilter() throws Exception{
		m = new EmptyModel(URI.createFileURI("test.uml"),"test");
		op = m.rcos.getClassModel().getBase_Package().createOwnedClass("test", false).createOwnedOperation("m", null, null);
		v = new DesignPrefixFilterVisitor(m.rcos,op);
	}

	@Test
	public void test1() throws Exception {
		v.run(Parse.getDesignFromString("skip ; a.b.m(a.c;x)"));
		assertEquals(3,v.getResult().size());
	}

	@Test
	public void test2() throws Exception {
		v.run(Parse.getDesignFromString("var T a; skip ; a.b.m(a.c;x); end a"));
		assertEquals(1,v.getResult().size());
	}

	@Test(expected=PrefixFilterException.class)
	public void test3() throws Exception {
		v.run(Parse.getDesignFromString("if (x.a) then { return 42 } else { g.m() }"));
	}

	@Test
	public void test4() throws Exception {
		v.run(Parse.getDesignFromString("if (x.a) then { g.k() } else { g.m() }"));
		v.print();
	}

	@Test
	public void test5() throws Exception {
		v.run(Parse.getDesignFromString("for (LineItem l : sale.lines) {[ ⊢sum' = sum + l.subtotal ] }"));
		v.print();
		assertEquals(2,v.getResult().size());
		assertNull(v.getChoice());
	}
}