package edu.unu.iist.rcos.spec.antlr;

import static org.junit.Assert.*;

import java.io.*;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.junit.Test;

import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Package;

import edu.unu.iist.rcos.spec.antlr.RCOSModelUtility;
import edu.unu.iist.rcos.spec.antlr.RCOSModelValidation;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import edu.unu.iist.rcos.spec.antlr.commands.ExpertPatternClassCommand;
import edu.unu.iist.rcos.spec.transformer.DesignTransformerEPGetED;
import edu.unu.iist.rcos.spec.transformer.DesignTransformerEPSetED;
import edu.unu.iist.rcos.uml.RCOSLoader;

public class Test_EP {

	@Test
	public void test_EP_D() throws Exception {
		RCOSLoader rl = new RCOSLoader("testcases/rCOS-models/Test-simpleget/TestEP.uml");
		EditingDomain ed = rl.editingDomain;
		List<Throwable> errors = new ArrayList<Throwable>();
		Command cmd = ExpertPatternClassCommand.runExpertPattern(ed, errors, (Class) rl.rcos.getClassModel().getBase_Package().getOwnedMember("D"));
		ed.getCommandStack().execute(cmd);
		assertTrue(cmd.canExecute());;
		assertTrue(new RCOSModelValidation().validate(rl.rcos));
	}

	@Test
	public void testEP_simpleget() throws Exception {
		String model = "Test-simpleget/TestEP.uml";
		RCOSLoader rl = transform_get(model);
		Package cm = (Package)rl.rcos.getClassModel().getBase_Package();
		assertNotNull(cm);
		Class b = (Class)cm.getOwnedMember("B");
		// Checking that all the methods are generated
		assertTrue("No all the methods were generated for B", b.getOwnedOperations().size() == 4);
		Class c = (Class)cm.getOwnedMember("C");
		Operation op = c.getOwnedOperation("foo", null, null);
		final String expectedS = "var int x;\nx := this.b.get_a_f_()";
		DESIGN expectedD = Parse.getDesign(expectedS, op);
		DESIGN d2 = Parse.getDesign(op);
		assertTrue("The design of C.foo() doesn't call the generated getter get_a_f_()", 
				DesignUtility.isDesignEqual(expectedD, d2));
	}
	
	@Test
	public void testEP_simpleset() throws Exception {
		String model = "Test-simpleset/TestEP.uml";
		transform_set(model);
	}

	@Test
	public void testEP_mix_setget() throws Exception {
		String model = "Test-mix_getset/TestEP.uml";
		transform_set(model);
		then_transform_get(model);
	}
	
	@Test
	public void testEP_vs_subclass1() throws Exception {
		// plain test case
		RCOSLoader rl = new RCOSLoader("testcases/expert/ExpertTest.uml");
		Package cm = (Package)rl.rcos.getClassModel().getBase_Package().getOwnedMember("test_subclass1");
		assertNotNull(cm);
		Class c = (Class)cm.getOwnedMember("O");
		assertNotNull(c);
		
		EditingDomain ed = rl.editingDomain;
		DesignTransformerEPSetED dtep = new DesignTransformerEPSetED(ed);		
		Command command = RCOSModelUtility.transform(ed,c, dtep);
		ed.getCommandStack().execute(command);
		assertTrue(command.canExecute());
		
		save(rl,"testcases/expert/out/ExpertTest-subclass1.uml");
//		assertTrue(new RCOSModelValidation().validate(rl.rcos)); // TODO: it's not a ReqM

		Class tgt = (Class)cm.getOwnedMember("C");
		assertNotNull(tgt.getOwnedOperation("set_x", null,null));
	}
	
	@Test
	public void testEP_vs_subclass_opt() throws Exception {
		// Ideally, setter should get created in "A" at the top
		RCOSLoader rl = new RCOSLoader("testcases/expert/ExpertTest.uml");
		Package cm = (Package)rl.rcos.getClassModel().getBase_Package().getOwnedMember("test_subclass1");
		assertNotNull(cm);
		Class c = (Class)cm.getOwnedMember("O");
		assertNotNull(c);
		
		EditingDomain ed = rl.editingDomain;
		DesignTransformerEPSetED dtep = new DesignTransformerEPSetED(ed);
		Command command = RCOSModelUtility.transform(ed,c, dtep);
		ed.getCommandStack().execute(command);
		assertTrue(command.canExecute());
		
		save(rl,"testcases/expert/out/ExpertTest-subclass_opt.uml");
//		assertTrue(new RCOSModelValidation().validate(rl.rcos)); // TODO: see above

		Class tgt = (Class)cm.getOwnedMember("A");
		assertNotNull("The setter could be generated higher up in the hierarchy",tgt.getOwnedOperation("set_x", null,null));
	}
	
	private static RCOSLoader transform_set(String string) throws Exception {
		return transform_set(string, false);
	}

	private static RCOSLoader transform_get(String string) throws Exception {
		return transform_get(string, false);
	}
	
	private static void then_transform_set(String string) throws Exception {
		transform_set(string, true);
	}

	private static void then_transform_get(String string) throws Exception {
		transform_get(string, true);
	}
	
	private static RCOSLoader transform_set(String model, boolean gen) throws Exception{
		// gen indicates if one should load the model already generated or the base model.
		String path = "testcases/rCOS-models/" + (gen?"gen/":"");
		RCOSLoader rl = new RCOSLoader(path + model);
		EditingDomain ed = rl.editingDomain;
		DesignTransformerEPSetED dtep = new DesignTransformerEPSetED(ed);
		Command command = RCOSModelUtility.transformAllClasses(ed,rl.rcos, dtep);
		ed.getCommandStack().execute(command);
		assertTrue(command.canExecute());
		
		save(rl,"testcases/rCOS-models/gen/" + model);
		assertTrue(new RCOSModelValidation().validate(rl.rcos));
		return rl;
	}
	
	private static RCOSLoader transform_get(String model, boolean gen) throws Exception{
		String path = "testcases/rCOS-models/" + (gen?"gen/":"");
		RCOSLoader rl = new RCOSLoader(path + model);
		EditingDomain ed = rl.editingDomain;
		DesignTransformerEPGetED dtep = new DesignTransformerEPGetED(ed);
		Command command = RCOSModelUtility.transformAllClasses(ed,rl.rcos, dtep);
		ed.getCommandStack().execute(command);
		assertTrue(command.canExecute());

		save(rl,"testcases/rCOS-models/gen/" + model);
		assertTrue(new RCOSModelValidation().validate(rl.rcos));
		return rl;
	}
	
	/**
	 * Save a UML model in a new file.
	 * Marked as final so that the RCOSLoader doesn't step on it accidentally.
	 * @param new_file
	 * @throws IOException
	 */
	private static void save(RCOSLoader rl, String new_file) throws IOException{
		File f = new File(new_file);
		OutputStream os = new FileOutputStream(f);
		rl.save(os, null);
	}
}
