package edu.unu.iist.rcos.spec.antlr;

import static org.junit.Assert.*;

import org.eclipse.uml2.uml.Type;
import org.junit.Test;

import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;
import edu.unu.iist.rcos.spec.antlr.visitor.expprefix.ExpPrefixFilterVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.expprefix.PrefixFilterException;
import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import static edu.unu.iist.rcos.spec.antlr.Parse.*;

public class Test_ExpPrefixFilter extends ExpPrefixFilterVisitor {

	public Test_ExpPrefixFilter() {
		super(new Vardecls() {
			@Override
			public Type getType(TYPE t) throws GetTypeException {
				return null;
			}
		});
	}

	private void add(EXP e) throws PrefixFilterException {
		add((FA)e);
	}

	@Test
	public void test1() throws Exception {
		visit(getExpFromString("x"));
		visit(getExpFromString("x"));
		assertEquals(1, getResult().size());
		PREFIX cs = getChoice();
		assertNotNull(cs);
		assertTrue(isExpEqual(cs, getExpFromString("x")));
	}

	@Test
	public void test2() throws Exception {
		add(getExpFromString("x"));
		add(getExpFromString("y"));
		PREFIX cs = getChoice();
		assertNull(cs);
	}

	@Test
	public void test3() throws Exception {
		add(getExpFromString("x.a"));
		add(getExpFromString("x.b"));		
		PREFIX cs = getChoice();
		assertNotNull(cs);
		assertTrue(isExpEqual(cs, getExpFromString("x")));
	}

	@Test
	public void test4() throws Exception {
		EXP fa1;
		add(fa1 = getExpFromString("x.a"));
		add(getExpFromString("x.a.c"));		
		PREFIX cs = getChoice();
		assertNotNull(cs);
		assertTrue(expToString(cs),isExpEqual(cs, fa1));
	}

	@Test
	public void test5() throws Exception {
		visit(getExpFromString("m(3).a"));
		PREFIX cs = getChoice();
		assertNull(cs);
	}

	@Test
	public void v_test6() throws Exception {
		visit(getExpFromString("a.b.m(3,a.b).z"));
		PREFIX cs = getChoice();
		assertNotNull(getResult().toString(),cs);
		assertTrue(isExpEqual(cs, getExpFromString("a.b")));
	}

	@Test
	public void v_test7() throws Exception {
		visit(getExpFromString("a.b.m(3,a.b,a.m).z"));
		PREFIX cs = getChoice();
		assertNotNull(cs);
		assertTrue(isExpEqual(cs, getExpFromString("a")));
		System.err.println(expToString(cs));
	}

	@Test
	public void v_test8() throws Exception {
		visit(getExpFromString("a.b.c.z AND a.b.k"));
		System.err.println("choices:");
		print(getResult());
		PREFIX cs = getChoice();
		assertNotNull(cs);
		assertTrue(expToString(cs),isExpEqual(cs, getExpFromString("a.b")));
	}

	@Test
	public void v_test_reduce() throws Exception {
		visit(getExpFromString("a.b.c.z AND a.b.k"));
		System.err.println("choices:");
		print(getResult());
		assertEquals(2,getResult().size());
	}

	@Test
	public void v_test9() throws Exception {
		visit(getExpFromString("a.f() AND a.g()"));
		System.err.println("choices:");
		print(getResult());
		PREFIX cs = getChoice();
		EXP expected = getExpFromString("a");
		assertTrue(expToString(cs),isExpEqual(cs, expected));
	}
}