package edu.unu.iist.rcos.spec.antlr;

import static org.junit.Assert.*;

import org.junit.*;

import org.eclipse.uml2.uml.Association;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Package;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.mm.util.WellformedDiagnosticException;
import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.profile.RCOSModel;
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel.*;
import edu.unu.iist.rcos.spec.antlr.parser.rCOSParser.DuplicatedMethodName;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;
import edu.unu.iist.rcos.spec.tc.*;
import edu.unu.iist.rcos.spec.tc.RuleTCMembers.*;
import edu.unu.iist.rcos.uml.*;

/**
 * TODO: Need to make sure that ParseTC and validate() find the same errors.
 */
public class Test_TC {

	RuleTCMembers rtc;

	@Test
	public void test_class_generalization() throws Exception{
		// RCOSTOOL-219
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("import lib::Collections; class A{ Set<A> field; }"));
		RCOSModel model = rm.rcos;
		Classifier Aclass = (Classifier) model.getClassModel().getBase_Package().getOwnedMember("A");
		assertNotNull(Aclass);
		Classifier objectClass = null;
		for(Classifier classifier : Aclass.getGenerals()){
			System.out.println(classifier.getName());
			if(classifier.getName().equals("Object"))
				objectClass = classifier;
		}
		assertNotNull("Didn't find expected generalization of Object!",objectClass);
		Association association = Aclass.getAssociations().get(0);
		assertTrue(association != null);
		Property field = null;
		for(Property type : association.getMemberEnds()){
			if(type.getName().equals("field"))
				field = type;
		}
		assertTrue(field != null);
		Type fieldType = field.getType();
		assertTrue(fieldType instanceof Class);
		assertEquals(1,((Class)fieldType).getGenerals().size());
		objectClass = null;
		for(Classifier classifier : ((Class)fieldType).getGenerals()){
			System.out.println(classifier.getName());
			if(classifier.getName().equals("Object"))
				objectClass = classifier;
		}
		assertTrue(objectClass != null);
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_broken_IF_inherit_1() throws Exception {
		translation("testcases/tc/broken-IF-inherit-1.rcosp");		
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_broken_delegation() throws Exception {
		translation("testcases/tc/broken-delegation.rcosp");		
	}

	@Test
	public void test_correct_IF_inherit_1() throws Exception {
		translation("testcases/tc/correct-IF-inherit-1.rcosp");		
	}

	@Test(expected=WellformedDiagnosticException.class)
	public void test_broken_IF_Co_1() throws Exception {
		translationThenValidate("testcases/tc/broken-IF-Co-1.rcosp");		
	}

	@Test
	public void test_correct_IF_Co_1() throws Exception {
		translationThenValidate("testcases/tc/correct-IF-Co-1.rcosp");		
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_attribute_default_1() throws Exception {
		translation("testcases/tc/broken-attribute-default-1.rcosp");		
	}

	@Test(expected=RCOSModelCheckException.class) // we should probably check that the right problem is actually in there...
	public void test_primed_default1() throws Exception {
		translation("testcases/tc/broken-primed-default.rcosp");
	}

	/* The next tests check default attribute initialisation */
	@Test(expected=RCOSModelCheckException.class)
	public void test_default_type1() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class B { public int x = true; }"));
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_default_type2() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class B { public int x = exists Object i : true; }"));
	}

	@Test
	public void test_default_type3() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class B { public int x = exists Object i : x=3; }"));
	}

	public void test_default_type4() throws Exception {
		MyRCOSModel rm = new MyRCOSModel("test.uml", "test", new RCOSFromString("class B { public int x = true; }"));
		Class b = (Class) rm.rcos.getClassModel().getBase_Package().getOwnedMember("B");
		Property p = b.getOwnedAttribute("x", null);
		assertEquals(p.getDefault(), "true");
		assertTrue(rm.validate());
	}
	/* END default attributes */

	@Test(expected=RCOSModelCheckException.class)
	public void test_primed_guard1() throws Exception {
		translation("testcases/tc/broken-primed-guard.rcosp");		
	}

	@Test(expected=WellformedException.class)
	public void test_primed_guard2() throws Throwable {
		MyRCOSModel rm= new MyRCOSModel("test.uml","test", new RCOSFromFile("testcases/tc/broken-primed-guard.rcosp"));
		try {
			assertFalse(validate(rm.rcos));
		} catch (RCOSModelCheckException e) {
			throw e.getCause();
		} 
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_primed_pre1() throws Exception  {
		translation("testcases/tc/broken-primed-pre.rcosp");		
	}

	@Test(expected=WellformedException.class)
	public void test_primed_pre2() throws Throwable {
		MyRCOSModel rm;
		rm = new MyRCOSModel("test.uml","test", new RCOSFromFile("testcases/tc/broken-primed-pre.rcosp"));
		try {
			assertFalse(validate(rm.rcos));
		} catch (RCOSModelCheckException e) {
			throw e.getCause();
		}
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_broken_duplicate_class() throws Exception {
		// has class A twice
		RCOSModel m = translation("testcases/tc/broken-duplicate-class.rcosp");
		assertFalse("Must not validate!",validate(m));
	}

	@Test(expected=DuplicatedMethodName.class)
	public void test_broken_IF_duplicate() throws Exception {
		RCOSModel rm = translation("testcases/grammar/broken_IF_duplicate.rcosp");
		assertFalse("Must not validate!",RCOSLoader.validate(rm));
	}

	@Test(expected=DuplicatedMethodName.class)
	public void test_broken_class_duplicate() throws Exception {
		RCOSModel rm = translation("testcases/tc/broken-duplicate-method.rcosp");
		assertFalse("Must not validate!",RCOSLoader.validate(rm));
	}

	@Test(expected=WellformedException.class)
	public void test_UML_AssocVsAttr_1() throws Exception {
		// At's a bit unclear who's at fault here. Even if we check it, this situation is
		// a bit awkward: from a UML perspective, everything is okay.
		// 
		RCOSLoader rl = new RCOSLoader("testcases/tc/Tests.uml");
		// we don't want things that are not UML
		assertTrue("Does not UML-validate.",rl.validate());
		assertFalse(validate(rl.rcos));
	}

	@Test
	public void test_UML_Params_inout() throws Exception {
		// We accept parameters with direction 'inout', but give you a warning.
		RCOSLoader rl = new RCOSLoader("testcases/tc/Tests.uml");
		Package p = (Package)rl.rcos.getClassModel().getBase_Package().getOwnedMember("test_Params_inout");
		Class c = (Class)p.getOwnedType("B");
		rmv.th = new TypeHandler(rl.rcos);
		assertTrue(rmv.validate(c));		
	}

	@Test(expected=WellformedException.class)
	public void test_UML_Params_retret() throws Exception {
		// We don't want more than one parameter with direction 'return'
		RCOSLoader rl = new RCOSLoader("testcases/tc/Tests.uml");
		Package p = (Package)rl.rcos.getClassModel().getBase_Package().getOwnedMember("test_Params_retret");
		Class c = (Class)p.getOwnedType("C");
		rmv.th = new TypeHandler(rl.rcos);
		assertTrue(rmv.validate(c));
	}

	@Test
	public void test_subtyping() throws Exception {
		RCOSModel rm = translation("testcases/grammar/subtyping.rcosp");
		assertTrue(validate(rm));
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_super_private_field() throws Exception {
		RCOSModel rm = translation("testcases/tc/broken-super-private-field.rcosp");
		assertFalse("Must not validate!",RCOSLoader.validate(rm));
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_super_private_method() throws Exception {
		RCOSModel rm = translation("testcases/tc/broken-super-private-method.rcosp");
		assertFalse("Must not validate!",RCOSLoader.validate(rm));
	}

	@Test
	public void test_Documentation_buffer() throws Exception {
		translation("../documentation/RCOS_Syntax/examples/buffer.rcosp");
	}

	@Test
	public void test_Documentation_buffer2() throws Exception {
		translationThenValidate("../documentation/RCOS_Syntax/examples/buffer.rcosp");
	}

	@Test
	public void test_Constructor1() throws Exception {
		translation("testcases/grammar/Constructor-1.rcosp");
	}

	@Test(expected=WellformedException.class)
	public void test_Constructor2() throws Exception {
		RCOSModel rm = translation("testcases/grammar/Constructor-2.rcosp");
		assertFalse("Must not validate!", validate(rm));
	}

	@Test
	public void test_Constructor_super() throws Exception {
		RCOSModel rm = translation("testcases/grammar/Constructor-super.rcosp");
		assertTrue(validate(rm));
	}

	@Test
	public void test_Constructor_this() throws Exception {
		RCOSModel rm = translation("testcases/grammar/Constructor-this.rcosp");
		assertTrue(validate(rm));
	}

	@Test
	public void test_mcall_super() throws Exception {
		RCOSModel rm = translation("testcases/grammar/mcall-super.rcosp");
		assertTrue(validate(rm));
	}

	@Test(expected=WellformedException.class)
	public void test_Params_twisted() throws Exception {
		RCOSLoader rl = new RCOSLoader("testcases/tc/Tests.uml");
		Package p = (Package)rl.rcos.getClassModel().getBase_Package().getPackagedElement("test_Params_twisted");
		Class d = (Class)p.getOwnedMember("D");
		assertFalse(rmv.checkImplements(d));
	}

	@Test
	public void test_confused_1() throws Exception {
		RCOSModel rm = translation("testcases/tc/tc-confused-1.rcosp");
		assertTrue(validate(rm));
	}

	@Test
	public void test_scope_1() throws Exception {
		RCOSModel rm = translation("testcases/tc/scope-1.rcosp");
		assertTrue(validate(rm));
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_assign_null() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public m() { null := 3 }}"));
		assertTrue(validate(rm.rcos));
	}

	@Test
	public void test_correct_return() throws Exception {
		RCOSModel rm = translation("testcases/tc/return-correct.rcosp");
		assertTrue(validate(rm));
	}

	@Test
	public void test_correct_return2() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public m() { return }}"));
		assertTrue(validate(rm.rcos));
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_broken_return_1() throws Exception {
		RCOSModel rm = translation("testcases/tc/broken-return-1.rcosp");
		assertTrue(validate(rm));
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_broken_return_2() throws Exception {
		RCOSModel rm = translation("testcases/tc/broken-return-2.rcosp");
		assertTrue(validate(rm));
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_broken_return_3() throws Exception {
		RCOSModel rm = translation("testcases/tc/broken-return-3.rcosp");
		assertTrue(validate(rm));
	}

	@Test
	public void tc_this1() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public A a; public m() { a := this } }"));
		assertTrue(validate(rm.rcos));
	}

	@Test
	public void tc_this2() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public A a; public m() { [ |- a' = this] } }"));
		assertTrue(validate(rm.rcos));
	}

	@Test
	public void tc_cast1() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public A a; public m() { a := ((A) this) } }"));
		assertTrue(validate(rm.rcos));
	}

	@Test
	public void tc_cast_mcall() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public m(A x) { m(((A) null)) } }"));
		assertTrue(validate(rm.rcos));
	}

	@Test(expected=WellformedException.class)
	public void tc_broken_circular() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A extends B {} class B extends A {}"));
		assertFalse(validate(rm.rcos));
		// TODO: Another testcase for twin on UML model. Could be done in OCL :)
	}

	@Test
	public void test_vardecl_1() throws Exception {
		try {
			ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public m() { var int b; end b; end b }}"));
			assertFalse(validate(rm.rcos));
		} catch (RCOSModelCheckException e) {
			// Too much knowledge...
			assertTrue(e.getCause().getCause() instanceof GetTypeException);
		}
	}

	@Test
	public void test_vardecl_2() throws Exception {
		try {
			ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public m() { var boolean b; if (b) then { end b } else { skip }; end b }}"));
			assertFalse(validate(rm.rcos));
		} catch (RCOSModelCheckException e) {
			// Too much knowledge...
			assertTrue(e.getCause().getCause() instanceof GetTypeException);
		}
	}

	@Test
	public void set_1() throws Exception {
		RCOSModel rm = translation("testcases/grammar/set-1.rcosp");
		assertTrue(validate(rm));
	}

	@Test
	public void object_extends() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A extends Object {}"));
	}

	@Test
	public void object_subtype() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public A() {} } class B { public m() { var Object x = A.new() } }"));
	}

	@Test
	public void assign_null() throws Exception {
		// XXX: null is compatible with type int?
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A {} class B { public A a; public int x; public m() { a := null; x := null } }"));
	}

	@Test
	public void rel_null() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A {} class B { public A a; public int x; public m() { [ |- a' = null AND x' = null] } }"));
	}

	@Test(expected=RCOSModelCheckException.class)
	public void null_is_not_arith() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A {} class B { public int x; public m() { [ |- x' = null + 1] } }"));
	}

	@Test(expected=RCOSModelCheckException.class)
	public void void_is_not_arith() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A {} class B { public int x; public m() { [ |- x' = null + 1] } public n() { x := m() + 1 }}"));
	}

	@Test // pure.
	public void pure_1() throws Exception {
		ParseRCOSModelTC m = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public pure m() { var int x; x := 25 }}"));
		Class cA = m.rcos.getClass("A", null);
		RuleOperationDesign rod = rtc.new RuleOperationDesign();
		final Operation op = cA.getOperation("m", null, null);
		rod.check(op);
		assertTrue(op.isQuery());
	}

	@Test // pure.
	public void pure_1_b() throws Exception {
		ParseRCOSModelTC m = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public m() { var int x; x := 25 }}"));
		Class cA = m.rcos.getClass("A", null);
		RuleOperationDesign rod = rtc.new RuleOperationDesign();
		final Operation op = cA.getOperation("m", null, null);
		rod.check(op);
		assertTrue(! op.isQuery());
	}

	@Test // pure.
	public void pure_2a() throws Exception {
		ParseRCOSModelTC m = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public pure m(; int x) { x := 25 }}"));
		Class cA = m.rcos.getClass("A", null);
		RuleOperationDesign rod = rtc.new RuleOperationDesign();
		final Operation op = cA.getOperation("m", null, null);
		rod.check(op);
		assertTrue(op.isQuery());
	}

	@Test // pure.
	public void pure_2b() throws Exception {
		ParseRCOSModelTC m = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public pure m(; int x) { [ true |- x' = 25 ] }}"));
		Class cA = m.rcos.getClass("A", null);
		RuleOperationDesign rod = rtc.new RuleOperationDesign();
		final Operation op = cA.getOperation("m", null, null);
		rod.check(op);
		assertTrue(op.isQuery());
	}

	@Test(expected=RCOSModelCheckException.class) // not pure.
	public void pure_3a() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { int y; public pure m(; int x) { [ true |- x' = 25 ] } public pure n(){ m(;y) }}"));
	}

	@Test // Infeasible but pure.
	public void pure_3b() throws Exception {
		ParseRCOSModelTC m = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { int y; public pure m(; int x) { [ true |- x' = 25 ]} public pure n(){ [ |- m(;y) ] }}"));
		Class cA = m.rcos.getClass("A", null);
		RuleOperationDesign rod = rtc.new RuleOperationDesign();
		final Operation op = cA.getOperation("m", null, null);
		rod.check(op);
		assertTrue(op.isQuery());
	}

	@Test(expected=RCOSModelCheckException.class) // Not pure
	public void pure_3c() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { int y; public pure m(; int x) { [ true |- x' = 25 ]} public pure n(){ [ |- m(;y') ] }}"));
	}

	public static RCOSModel translation(String rcosp) throws Exception {
		// TODO: save in a sensible place
		String filename="test.uml";
		ParseRCOSModelTC m = new ParseRCOSModelTC(filename, "test", new RCOSFromFile(rcosp));
		m.save();
		assertTrue("Does not UML-validate.",m.validate());
		return m.rcos;
	}

	public static RCOSModel translationThenValidate(String rcosp) throws Exception {
		String filename="test.uml";
		MyRCOSModel m = new MyRCOSModel(filename, "test", new RCOSFromFile(rcosp));
		m.save();
		assertTrue("Does not UML-validate.", m.validate());
		return m.rcos;
	}

	@Test
	public void test_instanceof() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public m() { VAR boolean b; [ true |- b' = (b instanceof Object) ]; END b }}"));
		assertTrue(validate(rm.rcos));
	}

	@Test
	public void test_instanceof2() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public m() { VAR boolean b; if (b instanceof A) then { [ true |- b' = true ] } else { [ true |- b' = false ]}; END b }}"));
		assertTrue(validate(rm.rcos));
	}	

	@Test
	public void test_postcond1() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("import lib::Collections; class A { public m() { var Set<string> a; VAR string b; [ |- a.contains(b) ] }}"));
		assertTrue(validate(rm.rcos));
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_postcond2() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("import lib::Collections; class A { public m() { var Set<string> a; [ |- a.size() ] }}"));
		assertTrue(validate(rm.rcos));
	}

	@Test
	public void test_postcond3() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public f(; int x) { return 1 } public m() { var int c; f(;c) }}"));
		assertTrue(validate(rm.rcos));
	}

	/**
	 * Should produce an error since it doesn't make sense to call a pure function as a method.
	 * Compare this with {@link #test_postcond1()}.
	 */
	@Test(expected=RCOSModelCheckException.class)
	public void test_postcond4() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("import lib::Collections; class A { public m() { var Set<string> a; VAR string b; a.contains(b) }}"));
		assertTrue(validate(rm.rcos));
	}

	/**
	 * This call is okay even though we call a pure function, since we capture
	 * the return value in a variable.
	 * Compare this with {@link #test_postcond4()}.
	 */
	@Test
	public void test_postcond4b() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("import lib::Collections; class A { public m() { var Set<string> a; VAR string b; var boolean c; a.contains(b;c) }}"));
		assertTrue(validate(rm.rcos));
	}

	// RCOSTOOL-230:
	@Test
	public void test_mcall_null() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public m(Object o) {m(null)} }"));
		assertTrue(validate(rm.rcos));
	}

	@Test
	public void test_new_null1() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public A(Object o) { var A a = A.new(null) } }"));
		assertTrue(validate(rm.rcos));
	}

	@Test(expected=RCOSModelCheckException.class)
	public void test_new_null2() throws Exception {
		// Shouldn't type check, since there are two constructors with similar signature.
		// TODO: pin down the "right" exception.
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public A(Object o) { var A a = A.new(null) } public A(A a) {skip}}"));
		assertFalse(validate(rm.rcos));
	}

	final RCOSModelValidationNoUML rmv = new RCOSModelValidationNoUML();

	protected boolean validate(RCOSModel rm) throws RCOSModelCheckException, WellformedException {
		return rmv.validate(rm);
	}

	@Before
	public void setup() throws NotAnRCOSModelException {
//		super.setup();
		rtc = new RuleTCMembers();
	}	
}
