package edu.unu.iist.rcos.spec.antlr.test;


import static org.junit.Assert.*;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Operation;
import org.junit.Test;

import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import static edu.unu.iist.rcos.spec.antlr.MyDesign.*;
import edu.unu.iist.rcos.spec.antlr.Parse;
import edu.unu.iist.rcos.spec.antlr.ParseRCOSModelTC;
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel.RCOSFromString;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.AOP;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.AOps;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.BINEXP;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.LOGOP;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.LogOps;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.PAR;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.PP;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.RELOP;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.RelOps;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.VAR;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement.EXPToComposition;

public class Test_EXPToComposition extends  EXPToComposition{

	@Test
	public void test_0() throws Exception{
		//---------------[ true |- (a' = c + e') AND c' = e' AND e' = f ]---------------
		VAR v1 = createVar(createVarname("a"),createInt(), true);
		VAR v2 = createVar(createVarname("c"),createInt(), false);
		VAR v7 = createVar(createVarname("e"),createInt(), true);
		AOP ap = createAOp(AOps.PLUS, v2, v7);
		RELOP rop1 = createRelOp(RelOps.EQ, v1, ap);//a' = c + e'
		
		VAR v3 = createVar(createVarname("c"),createInt(), true);
		VAR v4 = createVar(createVarname("e"),createInt(), true);
		RELOP rop2 = createRelOp(RelOps.EQ, v3, v4);//c' = e'
		
		BINEXP bexp = createLogOp(LogOps.AND, rop1, rop2);//a' = c + e' AND c' = e'
		
		VAR v5 = createVar(createVarname("e"),createInt(), true);
		VAR v6 = createVar(createVarname("f"),createInt(), false);
		RELOP rop3 = createRelOp(RelOps.EQ, v5, v6);//e' = f
		
		BINEXP bexp2 = createLogOp(LogOps.AND, bexp, rop3 );//a' = c + e' AND c' = e' AND e' = f

		System.out.println("----------------- [ |-"+expToString(bexp2)+"] ------------------");
		String expect_out = "[ |-e' = f ];\n[ |-a' = (c + e) ];\n[ |-c' = e ]";
		String s = designToString(postToSEQ((LOGOP)bexp2));
		System.out.println(s);
		assertEquals(expect_out,s);
	}

	@Test
	public void test_1() throws Exception{
		ParseRCOSModelTC m = new ParseRCOSModelTC("test.uml", "test", 
				new RCOSFromString("class A { public int a; public int c; public int e; public int f ; "+
						                                           "public m(){ [ true |- (a' = c + e') AND c' = e' AND e' = f ] }}"));				
		Class cA = m.rcos.getClass("A", null);
		Operation op = cA.getOperation("m", null, null);
		PP d = (PP) (Parse.getDesign(op));

		System.out.println("----------------- [ |-"+expToString(d.getPost())+"] ------------------");
		    DESIGN expect_out = Parse.getDesignFromString("[ |-e' = f ];[ |-a' = (c + e) ];[ |-c' = e ]");
			setCl(cA);
			DESIGN dout = postToSEQ((LOGOP)d.getPost());
			System.out.println(designToString(dout));
			assertTrue(isDesignEqual(dout, expect_out));		
	}

	@Test
	public void test_2() throws Exception{
		//---------------[ true |- (a' = c' and e') AND c' = e' AND e' = f ]---------------
		VAR v1 = createVar(createVarname("a"),createBoolean(), true);
		VAR v2 = createVar(createVarname("c"),createBoolean(), true);
		VAR v7 = createVar(createVarname("e"),createBoolean(), true);
		LOGOP lp = createLogOp(LogOps.AND, v2, v7);
		RELOP rop1 = createRelOp(RelOps.EQ, v1, lp);//a' = c' and e'
		
		VAR v3 = createVar(createVarname("c"),createBoolean(), true);
		VAR v4 = createVar(createVarname("e"),createBoolean(), true);
		RELOP rop2 = createRelOp(RelOps.EQ, v3, v4);//c' =e'
		
		BINEXP bexp = createLogOp(LogOps.AND, rop1, rop2); //(a' = c' or e') AND c' = e'
		
		VAR v5 = createVar(createVarname("e"),createBoolean(), true);
		VAR v6 = createVar(createVarname("f"),createBoolean(), false);
		RELOP rop3 = createRelOp(RelOps.EQ, v5, v6);//e' = f
		
		BINEXP bexp2 = createLogOp(LogOps.AND, bexp, rop3 );//(a' = c' or e') AND c' =e' AND e' = f

		System.out.println("----------------- [ |-"+expToString(bexp2)+"] ------------------");
		String expect_out = "[ |-e' = f ];\n[ |-c' = e ];\n[ |-a' = (c AND e) ]";
		String s = designToString(postToSEQ((LOGOP)bexp2));
		System.out.println(s);
		assertEquals(expect_out,s);
	}

    @Test
    public void test_3() throws Exception {
    	//---------------[ true |- (w' - n > x' + m => x' = y') AND m = w' ]---------------
		VAR w2 = createVar(createVarname("w"),createInt(), true); 	
		VAR m2 = createVar(createVarname("m"),createInt(), false);
		RELOP rop3 = createRelOp(RelOps.EQ, m2, w2);//m = w'

		//w' - n >= x' + m
		VAR x1 = createVar(createVarname("x"),createInt(), true); 	
		VAR m1 = createVar(createVarname("m"),createInt(), false);
		AOP ap2 = createAOp(AOps.PLUS, x1, m1);//x' + m
		
		VAR w1 = createVar(createVarname("w"),createInt(), true);
		VAR n1 = createVar(createVarname("n"),createInt(), false);
		AOP ap1 = createAOp(AOps.MINUS, w1, n1);//w' - n
		
		RELOP rop1 = createRelOp(RelOps.GT, ap1, ap2);//w' - n > x' + m
		//w' - n >= x' + m

		VAR x2 = createVar(createVarname("x"),createInt(), true); 	
		VAR y = createVar(createVarname("y"),createInt(), true);
		RELOP rop2 = createRelOp(RelOps.EQ, x2, y);//x' = y'
		
		LOGOP impl = createLogOp(LogOps.IMP, rop1, rop2);//w' - n > x' + m => x' = y'
		
		BINEXP be = createLogOp(LogOps.AND, impl, rop3);
		System.out.println("----------------- [ |-"+expToString(be)+"] ------------------");
		String expect_out = "[ |-(((w' - n) > (x' + m)) IMP (x' = y')) AND (m = w') ]";
		String s = designToString(postToSEQ((LOGOP)be));
		System.out.println(s);
		assertEquals(expect_out,s);
    }
 
    @Test
    public void test_4() throws Exception {
    	//---------------[ true |- y' = x AND x' = y' +y ]---------------
		VAR xp = createVar(createVarname("x"),createInt(), true); 	
		VAR y = createVar(createVarname("y"),createInt(), false);
		VAR yy = createVar(createVarname("y"),createInt(), true);
		AOP ap = createAOp(AOps.PLUS, yy, y);
		RELOP rop1 = createRelOp(RelOps.EQ, xp, ap);//x' = y' + y

		VAR yp = createVar(createVarname("y"),createInt(), true); 	
		VAR x = createVar(createVarname("x"),createInt(), false);
		RELOP rop2 = createRelOp(RelOps.EQ, yp, x);//y' = x
		
		BINEXP be = createLogOp(LogOps.AND, rop2, rop1);
		System.out.println("----------------- [ |-"+expToString(be)+"] ------------------");
		String expect_out = "var int y_b;\n[ |-y_b' = y ];\n[ |-y' = x ];\n[ |-x' = (y + y_b) ];\nend y_b";
		String s = designToString(postToSEQ((LOGOP)be));
		System.out.println(s);
		assertEquals(expect_out,s);
    }

    @Test
    public void test_5() throws Exception{
    	//---------------[ true |- w' - n > x' + m AND x' = y' AND m = w' ]---------------
		VAR w2 = createVar(createVarname("w"),createInt(), true); 	
		VAR m2 = createVar(createVarname("m"),createInt(), false);
		RELOP rop3 = createRelOp(RelOps.EQ, m2, w2);//m = w'

		//w' - n >= x' + m
		VAR x1 = createVar(createVarname("x"),createInt(), true); 	
		VAR m1 = createVar(createVarname("m"),createInt(), false);
		AOP ap2 = createAOp(AOps.PLUS, x1, m1);//x' + m
		
		VAR w1 = createVar(createVarname("w"),createInt(), true);
		VAR n1 = createVar(createVarname("n"),createInt(), false);
		AOP ap1 = createAOp(AOps.MINUS, w1, n1);//w' - n
		
		RELOP rop1 = createRelOp(RelOps.GT, ap1, ap2);//w' - n > x' + m
		//w' - n >= x' + m

		VAR x2 = createVar(createVarname("x"),createInt(), true); 	
		VAR y = createVar(createVarname("y"),createInt(), true);
		RELOP rop2 = createRelOp(RelOps.EQ, x2, y);//x' = y'
		
		LOGOP andop = createLogOp(LogOps.AND, rop1, rop2);//w' - n > x' + m AND x' = y'
		
		BINEXP be = createLogOp(LogOps.AND, andop, rop3);
		System.out.println("----------------- [ |-"+expToString(be)+"] ------------------");
		String expect_out = "[ |-m = w' ];\n[ |-(w - n) > (x' + m) ];\n[ |-x = y' ]";
		String s = designToString(postToSEQ((LOGOP)be));
		System.out.println(s);
		assertEquals(expect_out,s);
    }
    
    @Test
    public void test_6() throws Exception{
    	//---------------[ true |- x' >= y + y' AND z > x'  ]---------------
		VAR xp = createVar(createVarname("x"),createInt(), true); 	
		VAR y = createVar(createVarname("y"),createInt(), false);
		VAR yy = createVar(createVarname("y"),createInt(), true);
		AOP ap = createAOp(AOps.PLUS, y, yy);
		RELOP rop1 = createRelOp(RelOps.GEQ, xp, ap);//x' >= y + y' 

		VAR z = createVar(createVarname("z"),createInt(), false); 	
		VAR x = createVar(createVarname("x"),createInt(), true);
		RELOP rop2 = createRelOp(RelOps.GT, z, x);//z > x'
		
		BINEXP be = createLogOp(LogOps.AND, rop1, rop2);
		System.out.println("----------------- [ |-"+expToString(be)+"] ------------------");
		String expect_out = "[ |-z > x' ];\n[ |-x >= (y + y') ]";
		String s = designToString(postToSEQ((LOGOP)be));
		System.out.println(s);
		assertEquals(expect_out,s);
    }

    @Test
    public void test_62() throws Exception{
    	//---------------[ true |- x' > y AND z > x'  ]---------------
		VAR xp = createVar(createVarname("x"),createInt(), true); 	
		VAR y = createVar(createVarname("y"),createInt(), false);
		RELOP rop1 = createRelOp(RelOps.GT, xp, y);//x' >= y 

		VAR z = createVar(createVarname("z"),createInt(), false); 	
		VAR x = createVar(createVarname("x"),createInt(), true);
		RELOP rop2 = createRelOp(RelOps.GT, z, x);//z > x'
		
		BINEXP be = createLogOp(LogOps.AND, rop1, rop2);
		System.out.println("----------------- [ |-"+expToString(be)+"] ------------------");
		String expect_out = "[ |-(x' > y) AND (z > x') ]";
		String s = designToString(postToSEQ((LOGOP)be));
		System.out.println(s);
		assertEquals(expect_out,s);
    }

//    @Test
//    public void testFieldAccess(){
//    	System.out.println("---------------[ true |- e.a.a + g.a - g.a' >= f.a' AND g.a' = f.a ]---------------");
//   
//    	PREFIX pr_fp = createVar(createVarname("f"),null); 
//    	FA fa_fp = createFA(pr_fp, createVarname("a"),true);//f.a'
//
//    	VAR pr_e = createVar(createVarname("e"), null);
//    	FA fa_e = createFA(pr_e, createVarname("a"),false);//e.a
//    	FA x = createFA(fa_e, createVarname("a"),false);//e.a.a
//
//		PREFIX gg0 = createVar(createVarname("g"),null); 
//    	FA gg1 = createFA(gg0, createVarname("a"),true);//g.a'
//
//		PREFIX gg2 = createVar(createVarname("g"),null); 
//    	FA gg3 = createFA(gg2, createVarname("a"),false);//g.a
//
//    	AOP aop1 = createAOp(AOps.MINUS, gg3, gg1);
//    	AOP aop2 = createAOp(AOps.PLUS, x, aop1);
//		RELOP rop1 = createRelOp(RelOps.GEQ, aop2, fa_fp);//f'.a = e.a.x + g.a' - g.a
//
//
//		PREFIX pr_gp = createVar(createVarname("g"),null); 
//    	FA fa_gp = createFA(pr_gp, createVarname("a"),true);//g.a'
//
//    	PREFIX pr_h = createVar(createVarname("f"),null); 
//    	FA fa_h = createFA(pr_h, createVarname("a"),false);//f.a
//
//		RELOP rop2 = createRelOp(RelOps.EQ, fa_gp, fa_h);//g.a' = f.a
//		
//		BINEXP be = createLogOp(LogOps.AND, rop1, rop2);//f.a' = e.a.x AND g.a' = f.a
//		//print(be);
//		//System.out.println("");
//		try {
//			System.out.println(designToString(postToSEQ((LOGOP)be)));
//		} catch (RefinementException e) {
//			e.printStackTrace();
//		}		    
//    }
    
    @Test
    public void testPAR() throws Exception{//(((b * b') = a') AND (a = b'))
    	VAR a = createVar(createVarname("a"), createInt(), true);
    	PAR x = createPar(createVarname("b"), createInt(), false);
    	PAR bb = createPar(createVarname("b"), createInt(), true);
    	AOP aop = createAOp(AOps.MULT, x, bb);
    	RELOP rop1 = createRelOp(RelOps.EQ, aop, a);

    	PAR b = createPar(createVarname("b"), createInt(), true);	
    	VAR a2 = createVar(createVarname("a"), createInt(), false);
    	RELOP rop2 = createRelOp(RelOps.EQ, a2, b);
    	
		BINEXP be = createLogOp(LogOps.AND, rop1, rop2);
		System.out.println("----------------- [ |-"+expToString(be)+"] ------------------");
		String expect_out = "var int b_b;\n[ |-b_b' = b ];\n[ |-a = b' ];\n[ |-(b_b * b) = a' ];\nend b_b";
		String s = designToString(postToSEQ((LOGOP)be));
		System.out.println(s);
		assertEquals(expect_out,s);
    }
    
	@Test
	public void test_PAR2() throws Exception{
		ParseRCOSModelTC m = new ParseRCOSModelTC("test.uml", "test", 
				new RCOSFromString("class A { public int a; boolean b;"+
						                                           "public m(int i; int j){ [ true |- a' + i >= j' - j AND i < j' + a AND b' = (j' <8)] }}"));				
		Class cA = m.rcos.getClass("A", null);
		Operation op = cA.getOperation("m", null, null);
		PP d = (PP) (Parse.getDesign(op));

		System.out.println("----------------- [ |-"+expToString(d.getPost())+"] ------------------");
		ParseRCOSModelTC m2 = new ParseRCOSModelTC("test.uml", "test", 
				new RCOSFromString("class A { public int a; boolean b;"+
                        "public m(int i; int j){ var int j_b;[ |-(j_b' = j) ];[ |-(i < (j' + this.a)) ];"+
                        "[ |-(this.b' = (j < 8)) ];[ |-((this.a' + i) >= (j - j_b)) ];end j_b }}"));    
		Class cA2 = m2.rcos.getClass("A", null);
		Operation op2 = cA2.getOperation("m", null, null);
		DESIGN expect_out = Parse.getDesign(op2);

		setCl(cA);
		DESIGN dout = postToSEQ((LOGOP)d.getPost());
		System.out.println(designToString(dout));
		assertTrue(isDesignEqual(dout, expect_out));		
	}
    
    @Test
    public void testPAR_FA() throws Exception{//y' = buff AND buff' = y
		ParseRCOSModelTC m = new ParseRCOSModelTC("test.uml", "test", 
				new RCOSFromString("class A { public int f;"+
						                                           "public m(; int y){ [ true |- y' = f AND f' = y ] }}"));				
		Class cA = m.rcos.getClass("A", null);
		Operation op = cA.getOperation("m", null, null);
		PP d = (PP) (Parse.getDesign(op));
		
		System.out.println("----------------[ |- "+expToString(d.getPost())+" ]----------------");
		ParseRCOSModelTC m2 = new ParseRCOSModelTC("test.uml", "test", 
				new RCOSFromString("class A { public int f;"+
						                                           "public m(; int y){ var int y_b; [ true |- y_b' = y ; y' = f ; f' = y_b ];end y_b }}"));				
		Class cA2 = m2.rcos.getClass("A", null);
		Operation op2 = cA2.getOperation("m", null, null);
		DESIGN expect_out = Parse.getDesign(op2);

		setCl(cA);
		DESIGN dout = postToSEQ((LOGOP)d.getPost());
		System.out.println(designToString(dout));
		assertTrue(isDesignEqual(dout, expect_out));		
    }
    
}