package edu.unu.iist.rcos.spec.antlr.test;

import java.util.HashSet;
import java.util.Set;

import org.antlr.runtime.RecognitionException;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.*;
import org.junit.Test;
import static org.junit.Assert.*;

import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import static edu.unu.iist.rcos.spec.antlr.Parse.*;
import edu.unu.iist.rcos.profile.ContractInterface;
import edu.unu.iist.rcos.profile.RCOSModel;
import edu.unu.iist.rcos.spec.antlr.*;
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel.RCOSFromFile;
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel.RCOSFromString;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.parser.rCOSParser.ConflictWithKeyword;
import edu.unu.iist.rcos.spec.antlr.parser.rCOSParser.DuplicatedMethodName;
import edu.unu.iist.rcos.spec.antlr.visitor.ExpPrinter;
import edu.unu.iist.rcos.uml.RCOSLoader;

public class Test_Grammar {

	@Test
	public void test_Expression1() throws Exception {
		translation("testcases/grammar/Expression-1.rcosp");
	}
	
	@Test
	public void test_ProvidedContract() throws Exception {
		translation("testcases/grammar/Provided-contract.rcosp");
	}
	
	@Test
	public void test_composition2() throws Exception {
		translation("testcases/grammar/composition-2.rcosp");
	}
	
	@Test
	public void test_composition3() throws Exception {		
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("interface I {} contract Ctr of I {} component Co { provided Ctr by A; provided Ctr by A; class A implements I {}}"));
		Component co = rm.getComponent("Co", null, null);
		assertNotNull(co);
		assertEquals(2, co.getProvideds().size());
		assertFalse("Expected duplicate contract error!", rm.validate() /* new RCOSModelValidation().validate(rm.rcos) */);
	}	

	@Test
	public void test_LocalVarDecl1() throws Exception {
		translation("testcases/grammar/LocalVarDecl-1.rcosp");
	}

	@Test
	public void test_LocalVarDecl2() throws Exception {
		translation("testcases/grammar/LocalVarDecl-2.rcosp");
	}
	
	@Test
	public void test_Protocol_1() throws Exception {
		translation("testcases/grammar/protocol-range-1.rcosp");
	}

	@Test(expected=RecognitionException.class)
	public void test_SetOfSets() throws Exception {
		// we don't handle sets of sets for now, mostly because we don't have them in UML either
		translation("testcases/grammar/SetOfSets.rcosp");
	}
	
	@Test(expected=RecognitionException.class)
	public void test_broken_IF_methodname() throws Exception {
		translation("testcases/grammar/broken_IF_methodname.rcosp");
	}
	
	@Test(expected=DuplicatedMethodName.class)
	public void test_broken_duplicate() throws Exception {
		translation("testcases/grammar/broken_IF_duplicate.rcosp");
	}
	
	@Test
	public void generic_types() throws Exception {
			translation("testcases/grammar/Generic_types.rcosp");
	}
	
	@Test
	public void test_context() throws Exception {
			translation("testcases/grammar/Context.rcosp");
	}

	@Test
	public void test_VarDefault1() throws Exception {
		RCOSModel rm = translation("testcases/grammar/VarDefault-1.rcosp");
		Class c = (Class)rm.getClassModel().getBase_Package().getOwnedMember("Foo");
		Property p = c.getOwnedAttribute("i", null);
		assertEquals("42",p.getDefault());
	}
	
	@Test
	public void test_import1() throws Exception {
		RCOSModel rm = translation("testcases/grammar/import-1.rcosp");
		assertEquals(2,rm.getBase_Model().getImportedPackages().size());
	}



	@Test
	public void test_Documentation_exampleComposition() throws Exception {
		RCOSModel rm = translation("../documentation/RCOS_Syntax/examples/exampleComposition.rcosp");
		assertTrue(RCOSLoader.validate(rm));
	}
	
	@Test
	public void test_Documentation_fact() throws Exception {
		RCOSModel rm = translation("../documentation/RCOS_Syntax/examples/fact.rcosp");
		assertTrue(RCOSLoader.validate(rm));
	}
	
	@Test
	public void test_Constructor1() throws Exception {
		translation("testcases/grammar/Constructor-1.rcosp");
	}

	@Test
	public void test_Composition1() throws Exception {
		translation("testcases/grammar/composition-1.rcosp");
	}
	
	@Test
	public void test_unicode_1() throws Exception {
		translation("testcases/grammar/unicode-1.rcosp");
	}
	
	@Test
	public void test_return_exp() throws Exception {
		translation("testcases/grammar/return-exp.rcosp");
	}
	
	@Test
	public void test_visibility_method() throws Exception {
		RCOSModel rm = translation("testcases/grammar/visibility_method.rcosp");
		Class i = (Class) rm.getClassModel().getBase_Package().getOwnedMember("I");
		assertNotNull(getDesign(i.getOperation("transform", null, null)));
	}
	
	@Test(expected=RecognitionException.class)
	public void broken_design_spaceprime() throws Exception {
		getDesignFromString("[ |- y ' = 2]");
	}
	
	@Test(expected=RecognitionException.class)
	public void broken_mcall_space() throws Exception {
		getDesignFromString("[ |- foo. bar()]");
	}
	
	@Test(expected=RecognitionException.class)
	public void this_prime() throws Exception {
		DESIGN d = getDesignFromString("[ |- this'.contains(0)]");
		assertFalse(designToString(d), true);
	}
	
	@Test
	public void this_test() throws Exception {
		assertTrue(expToString(getExpFromString("this.x")).equals("this.x"));
	}

	@Test(expected=RecognitionException.class)
	public void broken_design_1() throws Exception {
		getDesignFromString("[ |- true] x");
	}

	@Test
	public void broken_design_mcall() throws Exception {
		MCALL d = (MCALL) getDesignFromString("m(;a)");
		assertEquals(1,d.getOut().size());
	}

	@Test
	public void default_lost() throws Exception {
		DESIGN d = getDesignFromString("var int x = 25");
		assertTrue("Result should be of type SEQ, but is of type "+d.getClass().getSimpleName(),d instanceof SEQ);
		final String ds = designToString(d);
		assertEquals("var int x;\nx := 25",ds);
	}

	public static RCOSModel translation(String rcosp) throws Exception {
		String filename="test.uml";
		MyRCOSModel m = new MyRCOSModel(filename, "test", new RCOSFromFile(rcosp));
		m.save();
		return m.rcos;
	}
	
	@Test
	public void test_test1() throws Exception {
		RCOSModel rm = translation("testcases/grammar/kewei/test1.rcosp");
		assertTrue(RCOSLoader.validate(rm));
	}

	@Test
	public void test_bridge() throws Exception {
		RCOSModel rm = translation("testcases/grammar/kewei/bridge.rcosp");
		assertTrue(RCOSLoader.validate(rm));
	}

	@Test
	public void test_peano() throws Exception {
		RCOSModel rm = translation("testcases/grammar/kewei/peano.rcosp");
		assertTrue(RCOSLoader.validate(rm));
	}

	@Test
	public void test_minlist() throws Exception {
		RCOSModel rm = translation("testcases/grammar/kewei/minlist.rcosp");
		assertTrue(RCOSLoader.validate(rm));
	}

	@Test
	public void test_linked_list() throws Exception {
		RCOSModel rm = translation("testcases/grammar/kewei/linked_list.rcosp");
		assertTrue(RCOSLoader.validate(rm));
	}
	
	@Test
	public void precedence1() throws Exception {
		// RCOSTOOL-184
		DESIGN d1 = getDesignFromString("SKIP \\/ (SKIP ; SKIP)");
		DESIGN d2 = getDesignFromString(designToString(d1));
		assertTrue(designToString(d2), isDesignEqual(d1, d2));
	}
	
	@Test(expected=RecognitionException.class)
	public void broken_design_postNotBool1() throws Exception {
		getDesignFromString("[ |- 2]");
	}
	
	@Test(expected=ConflictWithKeyword.class)
	public void broken_keyword_vardecl() throws Exception {
		getDesignFromString("var int process");
	}
	
	@Test(expected=ConflictWithKeyword.class)
	public void broken_attr_process() throws Exception {
		new MyRCOSModel("test.uml", "test", new RCOSFromString("class A {int process;}"));
	}
	
	@Test(expected=ConflictWithKeyword.class)	
	public void broken_meth_super() throws Exception {
		new MyRCOSModel("test.uml", "test", new RCOSFromString("class A {public super(int x) { skip}}"));
	}
	
	@Test(expected=ConflictWithKeyword.class)	
	public void broken_meth_process() throws Exception {
		new MyRCOSModel("test.uml", "test", new RCOSFromString("class A {public process(int x); }"));
	}
	
	
	@Test(expected=ConflictWithKeyword.class)
	public void broken_attr_this1() throws Exception {
		new MyRCOSModel("test.uml", "test", new RCOSFromString("class A {int this;}"));
	}
	
	@Test(expected=ConflictWithKeyword.class)
	public void broken_attr_this2a() throws Exception {
		// Ideally the parser should already reject this:
		new MyRCOSModel("test.uml", "test", new RCOSFromString("class A {int a; public m() {a.this := 1}}"));
	}

	@Test(expected=ConflictWithKeyword.class)
	public void broken_attr_this2b() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A {int a; public m() {a.this := 1}}"));
	}

	@Test(expected=ConflictWithKeyword.class)
	public void broken_attr_this3a() throws Exception {
		new MyRCOSModel("test.uml", "test", new RCOSFromString("class A { public m() { var int this }}"));
	}

	@Test(expected=ConflictWithKeyword.class)
	public void broken_attr_this3b() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("class A { public m() { var int this }}"));
	}

	@Test
	public void test_enumDecl1() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("enum E { E1,E2 }"));
	}
	@Test
	public void test_enumDecl2() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("enum E { E1,E2 } class A {public m(; E out) { out := E1 }}"));
	}
	@Test
	public void test_enumDecl3() throws Exception { // TODO: do we want this?
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("enum E { E1,E2 } class A {public m(; E out) { out := null }}"));
	}
	@Test(expected=RCOSModelCheckException.class)
	public void test_enumDecl4() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("enum E { E1,E2 } class A {public m(; E out) { out := 1 }}"));
	}
	@Test(expected=RCOSModelCheckException.class)
	public void test_enum_tc1() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("enum E { E1,E2 } enum F { F1,F2} class A {public m(; boolean out) { out := E1 = F2 }}"));
	}
	@Test
	public void test_enum_tc2() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("enum E { E1,E2 } enum F { F1,F2} class A {public m(; boolean out) { out := E1 = E2 }}"));
	}
	@Test(expected=RCOSModelCheckException.class)
	public void test_enum_tc3() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("enum E { E1,E2 } enum F { F1,F2} class A {public m(; F out) { out := E1 }}"));
	}
	@Test(expected=RCOSModelCheckException.class)
	public void test_enumDecl_unique1() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("enum E { E1,E2 } enum F { E1 }"));
	}
	@Test(expected=RCOSModelCheckException.class)
	public void test_enumDecl_unique2() throws Exception {
		new ParseRCOSModelTC("test.uml", "test", new RCOSFromString("enum E { E1,E2 } enum E { A }"));
	}

	@Test
	public void test_exp_component_navigation1() throws Exception {
		EXP e = Parse.getExpFromString("C1.C2.store = s ");
		assertNotNull(e);
	}
	
	@Test
	public void test_exp_component_navigation2() throws Exception {
		EXP e = Parse.getExpFromString("\\exists T s : C1.C2.store = s ");
		assertNotNull(e);
	}

	@Test
	public void test_nestedQuantSymbol() throws Exception {
		EXP e = Parse.getExpFromString("\\exists T s : ((\\exists T s : s != null) AND (s = null))");
		assertNotNull(e);
		String s = expToString(e);
		// quantSymbols are not properly scoped. Rewrite 
		assertFalse(s,s.contains("this.s"));
	}
	
	@Test
	public void test_precedence() throws Exception {
		MyRCOSModel rm = new ParseRCOSModelTC("test.uml","Test", new MyRCOSModel.RCOSFromString("class A {}"));
		Class ca = rm.th.getClass("A", null, null);
		assertNotNull(ca);
		EXP e = Parse.getExpFromString("(\\forall Object o : true) && (\\forall Object o : false)");
		String eOut = new ExpPrinter(ca).run(e);
		System.err.println(eOut);
		EXP e2 = Parse.getExpFromString(eOut);
		assertTrue(expToString(e2),DesignUtility.isExpEqual(e, e2));
	}

	/* Proxies for less warning markers */
	@SuppressWarnings("deprecation")
	private DESIGN getDesignFromString(String s) throws RecognitionException {
		return Parse.getDesignFromString(s);
	}
	@SuppressWarnings("deprecation")
	private String designToString(DESIGN d) {
		return DesignUtility.designToString(d);
	}
}
