package edu.unu.iist.rcos.spec.antlr.test;

import static org.junit.Assert.*;

import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Component;
import org.eclipse.uml2.uml.Interface;
import org.eclipse.uml2.uml.NamedElement;
import org.junit.*;

import edu.unu.iist.rcos.mm.util.Utility;
import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.profile.ContractInterface;
import edu.unu.iist.rcos.profile.Stereotypes;
import edu.unu.iist.rcos.spec.antlr.RCOSModelValidation;
import edu.unu.iist.rcos.spec.antlr.RCOSModelValidationNoUML;
import edu.unu.iist.rcos.uml.RCOSLoader;

public class Test_RCOSModelValidation {

	@Test
	public void test_OnePlaceBuffer() throws Exception {
		RCOSLoader rl = new RCOSLoader("../examples/OnePlaceBuffer/OnePlaceBuffer.uml");
		rl.validate();
	}

	@Test(expected=WellformedException.class)
	public void test_broken_ClassNotImplementing() throws Exception {
		RCOSLoader rl = new RCOSLoader("../examples/cocome/CoCoME/CoCoME.uml");
		NamedElement ctr = rl.rcos.getComponentModel().getBase_Package().getOwnedMember("Ctr_CD");
		assertNotNull(ctr);
		ContractInterface ctrI = Stereotypes.isContractInterface((Interface) ctr);
		Component co = (Component) rl.rcos.getComponentModel().getBase_Package().getOwnedMember("ProcessSale");
		assertNotNull(co);
		Class i = (Class) rl.rcos.getClassModel().getBase_Package().getOwnedMember("Item");
		assertNotNull(i);
		Utility.setControllerClass(co, ctrI, i);
		RCOSModelValidation rmv = new RCOSModelValidation();
		assertFalse("Class Item does not implement the contract interface Ctr_CD.", rmv.validate(rl.rcos));
	}

	@Test(expected=RCOSModelValidationNoUML.InvalidComponentComposition.class)
	public void test_ComponentComposition() throws Exception {
		RCOSLoader rl = new RCOSLoader("testcases/rCOS-models/Test_ComponentComposition.uml");
		assertFalse(
				"Unwrapped component composition at toplevel not detected!\nRCOSTOOL-88",
				new RCOSModelValidationNoUML().validate(rl.rcos)
		);
	}
}
