package edu.unu.iist.rcos.spec.antlr.test;

import org.antlr.runtime.RecognitionException;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.*;
import org.junit.Test;
import static org.junit.Assert.*;

import edu.unu.iist.rcos.mm.util.Designs;
import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel;
import static edu.unu.iist.rcos.spec.antlr.Parse.*;
import edu.unu.iist.rcos.spec.antlr.ParseRCOSModelTC;
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel.RCOSFromString;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import edu.unu.iist.rcos.spec.tc.RuleTCMembers;

public class Test_simplification {

	@Test
	public void test_Design1() throws Exception {
		assertTrue(checkDesign("skip; x:=3", "x:=3"));
	}
	
	@Test
	public void test_Design2() throws Exception {
		assertTrue(checkDesign("x:=3;skip","x:=3"));
	}
	@Test
	public void test_Design3() throws Exception {
		assertTrue(checkDesign("skip \\/ x:=3","x:=3"));
	}
	@Test
	public void test_Design4() throws Exception {
		assertTrue(checkDesign("if (true) then {x:= 3} else {x:= 2}", "x:=3"));
	}
	@Test
	public void test_Design5() throws Exception {
		assertTrue(checkDesign("if (false) then {x:= 3} else {x:= 2}", "x:=2"));
	}
	@Test
	public void test_Design6() throws Exception {
		assertTrue(checkDesign("if ((true AND true)) then { skip \\/ skip } else { skip }","skip"));
	}
	
	@Test
	public void test_Expression1() throws Exception {
		assertTrue(checkExp("x AND false", "false"));
	}

	@Test
	public void test_Expression2() throws Exception {
		assertTrue(checkExp("x AND true","x"));
	}

	@Test
	public void test_SeqAssoc1() throws Exception {
		DESIGN d = getDesignFromString("(x:=1; y:=2); z:=3");
		DESIGN dOut = simplifyDesign(d);
		assertTrue(checkDesign(dOut,"x:=1;y:=2;z:=3"));
	}

	@Test
	public void test_SeqAssoc2() throws Exception {
		DESIGN d = getDesignFromString("x:=1; (y:=2; z:=3)");
		DESIGN dOut = simplifyDesign(d);
		assertTrue(checkDesign(dOut,"x:=1;y:=2;z:=3"));
	}

	@Test
	public void test_pullUpIf1() throws Exception {
		MyRCOSModel rl = new ParseRCOSModelTC("test.uml","Test", new RCOSFromString("class A { public m(boolean b) {var int c; if (b) then { skip } else { chaos }; end c}}"));
		Class cA = (Class) rl.rcos.getClassModel().getBase_Package().getOwnedMember("A");
		Operation op = cA.getOwnedOperation("m", null, null);
		DESIGN d = getDesign(op);
		DESIGN d2 = simplify(d,cA);
		final String dS = designToString(op,d2);
		System.err.println(dS);
		Designs.iNSTANCE.update(op, dS);
		RuleTCMembers rtc = new RuleTCMembers();
		rtc.new RuleOperationDesign().check(op);
	}

	private static boolean checkDesign(DESIGN d1,String s2) throws Exception {
		return checkDesign(designToString(d1),s2);
	}

	/**
	 * @author Charles Morisset
	 * @param str1
	 * @param str2
	 * @return true if simplify(str1) == str2
	 * @throws RecognitionException
	 */
	private static boolean checkDesign(String str1, String str2) throws Exception {
		DESIGN d1 = simplifyDesign(getDesignFromString(str1));
		System.err.println(designToString(d1));
		return isDesignEqual(d1, 
				getDesignFromString(str2));
	}
	
	/**
	 * @author Charles Morisset
	 * @param str1
	 * @param str2
	 * @return true if simplify(str1) == str2
	 * @throws RecognitionException
	 */
	static boolean checkExp(String str1, String str2) throws Exception {
		return isExpEqual(simplifyExpression(getExpFromString(str1)), 
				getExpFromString(str2));
	}
	
}
