package edu.unu.iist.rcos.spec.antlr.test.charles;

import java.util.ArrayList;
import java.util.List;

import org.antlr.runtime.RecognitionException;
import org.eclipse.uml2.uml.Operation;
import org.junit.Before;
import org.junit.Test;

import edu.unu.iist.rcos.mm.resources.tests.CoCoME_Loader;
import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.PrinterException;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement.*;
import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import static edu.unu.iist.rcos.spec.antlr.Parse.*;
import static org.junit.Assert.*;

public class Test_DesignRefinement extends CoCoME_Loader {

	List<DESIGN> l	= new ArrayList<DESIGN>(20);
	private Operation finishSale;

	@Before
	public void local_setup() throws Exception {
		finishSale = cashdesk.getOwnedOperation("finishSale", null, null);
		assertNotNull(finishSale);
		finishSale.setIsQuery(true); // eeek
		l.add(getDesignFromString("[ (\\forall int x : startSale()) AND finishSale() |- true ]"));
		l.add(getDesignFromString("[ (\\forall int x : startSale()) OR finishSale() |- true ]"));
		l.add(getDesignFromString("[ (\\forall int x : startSale()) AND ((\\exist int y : enableExpress()) OR finishSale()) |- true ]"));
		l.add(getDesignFromString("[ true |- (2 = 2) => (3 = 3) => x'= 3 ] ; [(4 = 4) |- y' = 5]"));
		l.add(getDesignFromString("[ true |- (2 = 2) => (3 = 3) => x'= 3 ] ; [(4 = 4) |- y' = 5]"));
		l.add(getDesignFromString("[ true |- (\\forall int x : x > 1) => (3 = 3) => x'= 3 ] ; [(4 = 4) |- y' = 5]"));
		l.add(getDesignFromString("[ x > 2 |- false ]"));
//		l.add(getDesignFromString("[ x > 2 |- \\forall Item x : store.catalog.contains(x) => (y'=x + 2 AND foo(x))]"));
		l.add(getDesignFromString("[ x > 2 |- finishSale() ]"));
		l.add(getDesignFromString("[\\forall Item  x : store.catalog.contains(x) => x > 3 |- finishSale()]"));
		l.add(getDesignFromString("[\\exist Item x : store.catalog.contains(x) AND \\exists Item y : store.catalog.contains(y) AND p(x,y) |- finishSale() ]"));
		l.add(getDesignFromString("var boolean b ; [\\exist Item i : store.catalog.contains(i) && i.barcode = code && qty > 0 && sale != null |-  " +
				"line' = LineItem.new(code, qty) ; " +
				"\\exist Item j : store.catalog.contains(j) && j.barcode = code && " +
		" line.subtotal' = j.price * qty; sale.lines.add(line)]"));
		l.add(getDesignFromString("[\\exist Item i : store.catalog.contains(i) AND i.barcode = code && qty > 0 && sale != null |-  " +
				"line' = LineItem.new(code, qty) ; " +
				"\\exist Item j : store.catalog.contains(j) && j.barcode = code && " +
		" line.subtotal' = j.price * qty; sale.lines.add(line)]"));
		l.add(getDesignFromString("[\\exist Item i : store.catalog.contains(i) AND forall Item j : store.catalog.contains(j) => (i.barcode = code && qty > 0 && sale != null) |-  " +
				"line' = LineItem.new(code, qty) ; " +
				"\\exist Item j : store.catalog.contains(j) && j.barcode = code && " +
		" line.subtotal' = j.price * qty; sale.lines.add(line)]"));
		l.add(getDesignFromString("[\\exist Item i : store.catalog.contains(i) && i.barcode = code && qty > 0 && sale != null |-  true ]"));
		l.add(getDesignFromString("[ sale != null && bank.authorize(c, sale.total) |- sale.pay' = CardPayment.new(c); store.sales.add(sale); \\forall LineItem l: sale.lines.contains(l) => \\forall Item p : store.catalog.contains(p) => (p.barcode = l.barcode) => p.stock' = p.stock - l.quantity ]"));
		l.add(getDesignFromString("[|- \\exist Item j : store.catalog.contains(j) && j.barcode = code && " +
		" line.subtotal' = j.price * qty; sale.lines.add(line)]"));
		l.add(getDesignFromString("[|- \\forall Item j : store.catalog.contains(j) => (j.barcode = code && " +
		" line.subtotal' = j.price * qty); sale.lines.add(line)]"));
		l.add(getDesignFromString("[|- \\forall Item j : store.catalog.contains(j) => (j.barcode = code => " +
		" line.subtotal' = j.price * qty); sale.lines.add(line)]"));
	}

	@Test
	public void testExtractPredicate() throws Exception{

		DesignExtractPredicate dep = new DesignExtractPredicate(cashdesk);
		System.out.println("\n********************************** Start of design splitting ***************************************\n");

		for (DESIGN e : l){
			String new_des = designToString(dep.run(e));
			System.out.println("The design : \n" + designToString(e) + "\n is rewritten to  \n" +
					new_des + "\n");
			getDesignFromString(new_des);
		}

		System.out.println("\n********************************** End of design splitting ***************************************\n");
	}

	@Test
	public void testSplitPP() throws Exception{

		DesignSplitPP dspp = new DesignSplitPP(cashdesk);
		System.out.println("\n********************************** Start of design splitting ***************************************\n");

		for (DESIGN e : l){
			String new_des = designToString(dspp.run(e));
			System.out.println("The design : \n" + designToString(e) + "\n is rewritten to  \n" +
					new_des + "\n");
			getDesignFromString(new_des);
		}

		System.out.println("\n********************************** End of design splitting ***************************************\n");
	}

	@Test
	public void test_RelToAssign1() throws Exception {
		DesignRelToAssign da = new DesignRelToAssign(cashdesk);
		DESIGN d = getDesignFromString("[|- x']");
		DESIGN dExp = getDesignFromString("x:=true");
		DESIGN dOut;
		String out = designToString(dOut = da.run(d));
		assertTrue(out,isDesignEqual(dOut, dExp));
	}

	@Test
	public void test_RelToAssign2() throws Exception {
		DesignRelToAssign da = new DesignRelToAssign(cashdesk);
		DESIGN d = getDesignFromString("[|- !x']");
		DESIGN dExp = getDesignFromString("x:=false");
		DESIGN dOut;
		String out = designToString(dOut = da.run(d));
		assertTrue(out,isDesignEqual(dOut, dExp));
	}

	@Test
	public void testSplitPPAssign() throws Exception{

		DesignSplitPP dspp = new DesignSplitPP(cashdesk);
		DesignRelToAssign da = new DesignRelToAssign(cashdesk);

		System.out.println("\n********************************** 2 ***************************************\n");

		for (DESIGN e : l){
			DESIGN d = dspp.run(e);
			String new_des = designToString(da.run(d));

			System.out.println("The design : \n" + designToString(e) + "\n is rewritten to  \n" +
					new_des + "\n");

			// Check the design is still a correct expression
			getDesignFromString(new_des);
		}

		System.out.println("\n********************************** 2 ***************************************\n");
	}

	@Test
	public void testSplitPPAssignMethods() throws Exception{

		DesignSplitPP dspp = new DesignSplitPP(cashdesk);
		DesignRelToAssign da = new DesignRelToAssign(cashdesk);
		DesignRefineMethods drm = new DesignRefineMethods(cashdesk);

		System.out.println("\n********************************** 3 ***************************************\n");

		for (DESIGN e : l){
			DESIGN d = dspp.run(e);
			d = da.run(d);
			d = drm.run(d);
			String new_des = designToString(d);

			System.out.println("The design : \n" + designToString(e) + "\n is rewritten to  \n" +
					new_des + "\n");

			// Check the design is still a correct expression
			getDesignFromString(new_des);
		}

		System.out.println("\n********************************** 3 ***************************************\n");
	}

	
	
	@Test
	public void testRefine() throws Exception{


		System.out.println("\n********************************** 10 ***************************************\n");

		for (DESIGN e : l){

			System.out.println("The design : \n" + designToString(e) + "\n is rewritten to  \n" +
					designToString(refine(e, cashdesk)) + "\n");

		}

		System.out.println("\n********************************** 10 ***************************************\n");
	}
	
	@Test
	public void testSplitPPAssignMethodsAssert() throws Exception{

		DesignSplitPP dspp = new DesignSplitPP(cashdesk);
		DesignRelToAssign da = new DesignRelToAssign(cashdesk);
		DesignRefineMethods drm = new DesignRefineMethods(cashdesk);
		DesignRefineAssert dra = new DesignRefineAssert(cashdesk);

		System.out.println("\n********************************** 4 ***************************************\n");

		for (DESIGN e : l){
			DESIGN d = dspp.run(e);
			d = da.run(d);
			d = drm.run(d);
			d = dra.run(d);
			String new_des = designToString(d);

			System.out.println("The design : \n" + designToString(e) + "\n is rewritten to  \n" +
					new_des + "\n");

			// Check the design is still a correct expression
			getDesignFromString(new_des);
		}

		System.out.println("\n********************************** 4 ***************************************\n");
	}

	@Test
	public void testSplitPPre() throws Exception{

		DesignExtractPredicate dsp = new DesignExtractPredicate(cashdesk);
		DesignSplitPP dspp = new DesignSplitPP(cashdesk);
		DesignRelToAssign da = new DesignRelToAssign(cashdesk);
		DesignRefineMethods drm = new DesignRefineMethods(cashdesk);
		DesignRefineAssert dra = new DesignRefineAssert(cashdesk);

		System.out.println("\n********************************** 5 ***************************************\n");

		for (DESIGN e : l){

			System.out.println("The design : \n" + designToString(e) + "\n is rewritten to  \n");

			DESIGN d = dspp.run(e);
			d = dsp.run(d);
			d = da.run(d);
			d = drm.run(d);
			d = dra.run(d);
			d = dspp.run(d);
			d = da.run(d);
			d = dsp.run(d);


			String new_des = designToString(d);

			System.out.println(new_des + "\n");
			
			// Check the design is still a correct expression
			getDesignFromString(new_des);
		}
		System.out.println("\n********************************** 5 ***************************************\n");
	}

	private String designToString(DESIGN d) throws PrinterException {
		return DesignUtility.designToString(cashdesk,d);
	}
	private DESIGN getDesignFromString(String d) throws RecognitionException, WellformedException {
		return getDesign(d, finishSale);
	}
}
