package edu.unu.iist.rcos.spec.antlr.test.charles;

import static org.junit.Assert.*;

import java.util.ArrayList;
import java.util.List;

import org.antlr.runtime.RecognitionException;
import org.junit.Test;

import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.Parse;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.EXP;

public class Test_equals {

	@Test
	public void testDesignParsing() throws RecognitionException{

		List<String> l	= new ArrayList<String>(20);
		l.add("if (b) then {  	skip  } else { 	chaos  }; end b; this.line := LineItem.new(code, qty);  this.sale.lines.add(this.line)");
		l.add("quantity := 2");
		l.add("var C x; x := 2");
		l.add("x:= x; y := 2; x:= 2");
		l.add("x := y");
		l.add("if (b) then {x := 3} else {y := 4}");
		l.add("while (b) {x := 3; x := y}");
		l.add("x:= 2 \\/ x := y");
		l.add("m(a, b; x, y)");
		l.add("m(c', d; x, y)");
		l.add("m(e', f'; x, y)");
		l.add("skip");
		l.add("n()");
		l.add("skip; x:= 2");
		l.add("[ x' = 2 |- x' = y AND x' = 3]");
		l.add("return 3");
		l.add("skip; return 2");

		List<String> l2	= new ArrayList<String>(20);
		l2.add("var boolean b;\nb := false;\nfor (Item i : this.store.catalog) { \n	if (((i.barcode = code) AND ((qty > 0) AND (this.sale != null)))) then {\n		b := true\n 	} else {\n		skip\n 	}\n};\nif (b) then {\n	skip\n } else {\n	chaos\n };\n end b;this.line := LineItem.new(code, qty);\n [ true|-\\exists Item j:(this.store.catalog.contains(j) AND ((j.barcode = code) AND (this.line.subtotal' = (j.price * qty)))) ];\n this.sale.lines.add(this.line)");
		l2.add("quantity := 4");
		l2.add("var C x");
		l2.add("x:= x; z := 2; x:= 2");
		l2.add("y := x");
		l2.add("if (b) then {x := 4} else {y := 3}");
		l2.add("while (true) {x := 3; x := y}");
		l2.add("x:= y \\/ x := 2");
		l2.add("m(a, b; z, y)");
		l2.add("m(c', e; x, y)");
		l2.add("m(g', f'; x, y)");
		l2.add("chaos");
		l2.add("n(x)");
		l2.add("skip");
		l2.add("[ y' = 2 |- y' = y AND y' = 3]");
		l2.add("return 4");
		l2.add("return 2");
		//		
		for (String e : l) {
			assertTrue("The Design '"  + e + "' shouldn be equal to itself ",  
					DesignUtility.isDesignEqual(Parse.getDesignFromString(e), Parse.getDesignFromString(e)));
		}
		for (String e : l2) {
			assertTrue("The Design '"  + e + "' shouldn be equal to itself ",  
					DesignUtility.isDesignEqual(Parse.getDesignFromString(e), Parse.getDesignFromString(e)));
		}
		
		for (int i = 0; i < l.size(); i++){
			assertFalse("The Designs '"  + l.get(i) + "' and '" + l2.get(i) + "' shouldn't be equal",  
					DesignUtility.isDesignEqual(Parse.getDesignFromString(l.get(i)), Parse.getDesignFromString(l2.get(i))));
		}
	}

	@Test
	public void testExpression() throws RecognitionException{
		
		List<String> l = new ArrayList<String>(20);
		l.add("LineItem.new(code, qty)");
		l.add("0");
		l.add("true");
		l.add("1 + 2");
		l.add("x + 3");
		l.add("z + y");
		l.add("w = 5");
		l.add("p' = 5");
		l.add("m(1)");
		l.add("m(x)");
		l.add("m(a, b)");
		l.add("m(c', d; x)");
		l.add("m(e', f'; x, y)");
		l.add("\\forall T x : x' = 8");
		l.add("\\forall T x : x = 8 AND y > 9");
		l.add("x.y' + x.z");
		l.add("x' = 2 AND y = 3 AND z' = 7 OR x' = 1");

		List<String> l2 = new ArrayList<String>(20);
		l2.add("LineItem.new(code, x)");
		l2.add("8");
		l2.add("false");
		l2.add("1 + 4");
		l2.add("x' + 3");
		l2.add("z + y + 5");
		l2.add("w = 7");
		l2.add("p' = x");
		l2.add("m(4)");
		l2.add("m(y)");
		l2.add("m(e, f)");
		l2.add("m(d', d; x)");
		l2.add("m(e', f'; w, y)");
		l2.add("\\forall T y : y' = 8");
		l2.add("\\forall T y : y = 8 AND y > 9");
		l2.add("x.z' + x.z");
		l2.add("x' = 2 AND y = 4 AND z' = 7 OR x' = 1");
		
		for (String e : l) {
			final EXP expE = Parse.getExpFromString(e);
			assertTrue("The expression '"  + e + "' should be equal to itself ",  
					DesignUtility.isExpEqual(expE, expE));
		}
		for (String e : l2) {
			final EXP expE = Parse.getExpFromString(e);
			assertTrue("The expression '"  + e + "' should be equal to itself ",  
					DesignUtility.isExpEqual(expE, expE));
		}
		
		for (int i = 0; i < l.size(); i++){
			assertFalse("The expressions '"  + l.get(i) + "' and '" + l2.get(i) + "' shouldn't be equal",  
					DesignUtility.isExpEqual(Parse.getExpFromString(l.get(i)), Parse.getExpFromString(l2.get(i))));
		}
	}
	
	@Test
	public void testWeakEqual() throws RecognitionException{
		final EXP xD = Parse.getExpFromString("x");
		assertTrue(DesignUtility.isExpEqual(xD, xD));
		assertFalse(DesignUtility.isExpEqual(xD, Parse.getExpFromString("x'")));
		assertTrue(DesignUtility.isExpWeakEqual(xD, xD));
		assertTrue(DesignUtility.isExpWeakEqual(xD, Parse.getExpFromString("x'")));
		assertTrue(DesignUtility.isExpWeakEqual(Parse.getExpFromString("x = 2"), Parse.getExpFromString("x' = 2")));
		
		final DESIGN x2D = Parse.getDesignFromString("[ |- x = 2 ]");
		assertTrue(DesignUtility.isDesignEqual(x2D, x2D));
		assertFalse(DesignUtility.isDesignEqual(x2D, Parse.getDesignFromString("[ |- x' = 2 ]")));
		assertTrue(DesignUtility.isDesignWeakEqual(x2D, x2D));
		assertTrue(DesignUtility.isDesignWeakEqual(x2D, Parse.getDesignFromString("[ |- x' = 2 ]")));
	}
	
	@Test
	public void test_equals_vardecl1() throws Exception {
		DESIGN oldD = Parse.getDesignFromString("var int x = 0");
		DESIGN newD = Parse.getDesignFromString(DesignUtility.designToString(oldD));
		assertTrue(DesignUtility.isDesignEqual(oldD, newD));
	}

	@Test
	public void test_equals_vardecl2() throws Exception {
		/* This produces (SEQ (SEQ VAR ASSIGN) END) != (SEQ VAR (SEQ ASSIGN END) */
		DESIGN oldD = Parse.getDesignFromString("var int x = 0 ;end x");
		DESIGN newD = Parse.getDesignFromString(DesignUtility.designToString(oldD));
		assertTrue(DesignUtility.isDesignEqual(oldD, newD));
	}
}
