package edu.unu.iist.rcos.spec.antlr.visitor.transformation.delegation;

import static org.junit.Assert.*;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Operation;
import org.junit.Test;

import edu.unu.iist.rcos.mm.util.EFRCOSRuleFeature;
import edu.unu.iist.rcos.modeler.translation.RCOSpecTranslation;
import edu.unu.iist.rcos.spec.antlr.*;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.delegation.DesignDelegator;

public class Test_DesignDelegator {

	final EditingDomain ed = EFRCOSRuleFeature.initEditingDomain();
	
	@Test
	public void test1() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new MyRCOSModel.RCOSFromString(
		"class D { D d; C c; int v; } class C { D d; public m(int j, D k) { d.d.v := j+j; d.c.m(j,k) } }"));
		Class c = rm.getClass("C", null, null);
		assertNotNull(c);
		Operation op = c.getOwnedOperation("m", null, null);
		assertNotNull(op);
		FA fa = (FA) Parse.getGuard("d", op);
		assertNotNull(fa);
		DesignDelegator dd = new DesignDelegator(ed, op, fa, new VardeclsFIS(op));
		DESIGN d = Parse.getDesign(op);
		assertNotNull(d);
		DESIGN d2 = dd.run(d);
		assertNotNull(d2);
		Command cmd = dd.mkOperation("foo",d2);		
		ed.getCommandStack().execute(cmd);
		assertTrue(cmd.canExecute());
		System.err.println(new RCOSpecTranslation().translate(rm.rcos));
		assertTrue(rm.validate());
	}
	
	@Test
	public void test2() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new MyRCOSModel.RCOSFromString(
		"class D { D d; C c; int v; } class C { D d; public m(int j, D k) { k.d.v := j+j; k.d.c.m(j,k) } }"));
		Class c = rm.getClass("C", null, null);
		assertNotNull(c);
		Operation op = c.getOwnedOperation("m", null, null);
		assertNotNull(op);
		REF fa = (REF) Parse.getGuard("k", op);
		assertNotNull(fa);
		DesignDelegator dd = new DesignDelegator(ed,op, fa, new VardeclsFIS(op));
		DESIGN d = Parse.getDesign(op);
		assertNotNull(d);
		DESIGN d2 = dd.run(d);
		assertNotNull(d2);
		Command cmd = dd.mkOperation("foo",d2);		
		ed.getCommandStack().execute(cmd);
		assertTrue(cmd.canExecute());
		System.err.println(new RCOSpecTranslation().translate(rm.rcos));
		assertTrue(rm.validate());
	}
	
	@Test
	public void test3() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new MyRCOSModel.RCOSFromString(
		"class D { D d; C c; int v; } class C { D d; public m(int j, D k) { k.d.v := j+j; k.d.c.m(j,k) } }"));
		Class c = rm.getClass("C", null, null);
		assertNotNull(c);
		Operation op = c.getOwnedOperation("m", null, null);
		assertNotNull(op);
		REF fa = (REF) Parse.getGuard("k.d", op);
		assertNotNull(fa);
		DesignDelegator dd = new DesignDelegator(ed,op, fa, new VardeclsFIS(op));
		DESIGN d = Parse.getDesign(op);
		assertNotNull(d);
		DESIGN d2 = dd.run(d);
		assertNotNull(d2);
		Command cmd = dd.mkOperation("foo",d2);		
		ed.getCommandStack().execute(cmd);
		assertTrue(cmd.canExecute());
		System.err.println(new RCOSpecTranslation().translate(rm.rcos));
		assertTrue(rm.validate());
	}

	@Test
	public void test5() throws Exception {
		ParseRCOSModelTC rm = new ParseRCOSModelTC("test.uml", "test", new MyRCOSModel.RCOSFromString(
		"class E{} class D { D d; C c; int v; } class C { E e; D d; public m(int j, D k) { k.d.v := j+j; k.d.c.m(j,k) } }"));
		Class c = rm.getClass("C", null, null);
		assertNotNull(c);
		Operation op = c.getOwnedOperation("m", null, null);
		assertNotNull(op);
		REF fa = (REF) Parse.getGuard("e", op);
		assertNotNull(fa);
		DesignDelegator dd = new DesignDelegator(ed,op, fa, new VardeclsFIS(op));
		DESIGN d = Parse.getDesign(op);
		assertNotNull(d);
		DESIGN d2 = dd.run(d);
		assertNotNull(d2);
		Command cmd = dd.mkOperation("foo",d2);		
		ed.getCommandStack().execute(cmd);
		assertTrue(cmd.canExecute());
		System.err.println(new RCOSpecTranslation().translate(rm.rcos));
		assertTrue(rm.validate());
	}
}