package edu.unu.iist.rcos.modeler.translation;

import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Map.Entry;

import org.antlr.runtime.RecognitionException;
import org.eclipse.uml2.uml.*;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Package;

import edu.unu.iist.rcos.mm.util.Invariants;
import edu.unu.iist.rcos.mm.util.Utility;
import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.profile.*;
import edu.unu.iist.rcos.spec.antlr.Parse;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;

public abstract class RCOSTranslation<T1, T2> extends Translation {

	public static String COMPONENT_MODEL = "_ComponentModel";

	public static String CLASS_MODEL = "_ClassModel";

	public static String NEW_LINE = "\n";

	public static String TAB = "\t";

	public static String BLANK = " ";

	public static String LBRACKET = "{";

	public static String RBRACKET = "}";

	public static String INTERFACE = "interface ";

	public static String CONTRACT = "contract ";

	public static String OF = "of ";

	public static String PROVIDED_INTERFACE = "provided ";

	public static String REQUIRED_INTERFACE = "required ";

	public static String IMPLEMENTED_BY = "by ";

	public static String SERVICE_COMPONENT = "component";

	public static String PROCESS_COMPONENT = "process";

	public static String PUBLIC = "public";

	public static String PRIVATE = "private";

	public static String PROTECTED = "protected";

	public static String SC = ";";
	
	public static String COLON = ":";

	public static String LLBRACKET = "(";

	public static String RLBRACKET = ")";

	public static String EXTENDS = "extends";

	public static String IMPLEMENTS = "implements";	

	public static String CLASS = "class";

	public static String PRE = "pre";

	public static String POST = "post";

	public static String DESIGN = "design";

	public static String INVARIANTS = "invariants";

	public static String INVARIANT = "invariant";	

	public static String PROTOCOL = "protocol";

	public static String COMPOSITION = "composition";

	public static String COMPOSITION_STEREOTYPE = "Composition";

	public static String DELEGATION_STEREOTYPE = "Delegation";

	public static String LOCATIONS = "locations";

	public static String INIT = "init";

	public static String ACTION = "action";

	public static String GUARD = "guard";

	public static String TRANSITION = "->";

	public static String CONF_FILE = ".loc";
	
	public static String IMPORT_SEP = "::";
	
	public static String LOOP_WHILE = "while";
	
	public static String LOOP_FOR = "for";
	
	public static String CONDITION_IF = "if";
	
	public static String CONDITION_ELSE = "else";
	
	public static String RETURN = "return";
	
	public static String NEW = "new";

	public static String VOID = "void ";
	
	public static String ASSIGN = "=";

	public static String DOT = ".";

	public static String COMMA = ",";

	// A list containing all the class realizing an interface
	// i.e. the classes defined in a component 
	protected ArrayList<Class> listOfRealizingClasses = new ArrayList<Class>();

	protected static String getTab(int tab){
		String str = "";
		for (int i = 0 ; i < tab; i++){
			str += "\t";
		}
		return str;
	}
	
	protected static void append(StringBuffer buf, String str, int tab){
		buf.append(getTab(tab));
		buf.append(str);
	}

	public String translate(RCOSModel rm) throws ExportException {
		// From here we are sure it is a requirement model
		Model model = rm.getBase_Model();
		String modelName = model.getName();

		StringBuffer spec = new StringBuffer();

		ComponentModel com = rm.getComponentModel();
		Package compPkg = com.getBase_Package();

		if (compPkg == null)
			throw new ExportException("Inconsistent environment: couldn't get Package:"+modelName+COMPONENT_MODEL);

		// We first translate the import classes 
		translateImportedClasses(model);
		
		// We then translate all the interfaces at top level
		for (PackageableElement e : compPkg.getPackagedElements()){
			if (e instanceof Interface && Stereotypes.isContractInterface((Interface) e) == null){
				Interface i = (Interface) e;
				try {
					translateInterface(i);
				} catch (WellformedException e1) {
					throw new ExportException(e1);
				} catch (RecognitionException e2) {
					throw new ExportException(e2);
				}
			}
		}

		// Then the different contracts
		for (PackageableElement e : compPkg.getPackagedElements()){
			if (e instanceof Interface && Stereotypes.isContractInterface((Interface) e) != null){
				ContractInterface ctr = Stereotypes.isContractInterface((Interface) e);
				translateContract(ctr);
			}
		}
		// Then we translate each component
		for (PackageableElement e : compPkg.getPackagedElements()){
			if (e instanceof Component){
				try {
					translateComponent((Component)e);
				} catch (WellformedException e1) {
					throw new ExportException(e1);
				} catch (RecognitionException e2) {
					throw new ExportException(e2);
				}
			}
		}
		// At this point, everything is translated in the buffer spec, 
		// except the classes which are in the class diagram but not in 
		// any component.
		// All the classes inside a component have been added to 
		// listOfRealizingClasses with the method translateComponentClass

		// We get the class diagram and call the method translateOtherClasses
		// which does everything, including the definition of the "dummy" component

		
		ClassModel cm = rm.getClassModel();	
		Package classPkg = cm.getBase_Package();
		
		try {
			translateOtherClasses(classPkg);
		} catch (WellformedException e1) {
			throw new ExportException(e1.toString());
		} catch (RecognitionException e) {
			throw new ExportException(e.toString());
		}

		return spec.toString();
	}

	protected void translateOtherClasses(Package base_Package) throws WellformedException, RecognitionException, ExportException{
		for (PackageableElement e : base_Package.getPackagedElements()){
			if (e instanceof Class && ! listOfRealizingClasses.contains((Class)e)){
				translateClass((Class)e);
			} else if (e instanceof Package) {
				translateOtherClasses((Package)e);
			} else if ((e instanceof Comment) || (e instanceof Association)) {
				// Nada -- skip those
			} else
				assert(false); // XXX			
		}
	}

	protected void translateClass(Class c) throws WellformedException, RecognitionException, ExportException {

		// If it's a binding, do not translate
		if (c.getEAnnotation(Utility.BINDING_DNS) != null) return;

		if (c.getVisibility() == null) 
			;
		else
			;

		//signature
		translateSignature(c.getOwnedTemplateSignature());
		
		//generalization
		translateGeneralization(c);

		//Generate interfaces it realizes
		translateInterfaceRealization(c);

		// Generate invariant specification
		Invariants.iNSTANCE.get(c);

		//fields and methods
		translateFields(c);
		translateMethods(c);
	}

	protected void translateMethods(Class c) throws RecognitionException, WellformedException {
		for(Operation op : c.getOperations()){
			translateMethod(c, op);
		}
	}

	protected void translateMethod(Class c, Operation op) throws RecognitionException, WellformedException {
		for (Parameter p : op.getOwnedParameters())
			translateParameter(p);
		DESIGN d = Parse.getDesign(op);
		translateDesign(c, d);
	}

	protected void translateFields(Classifier c) throws ExportException, RecognitionException {
		for(Property p : c.getAllAttributes()){
			translateField(c, p);
		}
	}

	protected void translateInterfaceRealization(Class c) throws WellformedException, ExportException, RecognitionException {
		for (Interface i : c.getImplementedInterfaces()){
			translateImplementedInterface(c, i);
		}
	}

	protected void translateImplementedInterface(Class c, Interface i) throws WellformedException, ExportException, RecognitionException {
		translateFields(i);
		translateMethods(i);
	}

	protected void translateMethods(Interface i) throws WellformedException {
		for (Operation op : i.getOperations()){
			translateMethod(i, op);
		}
	}

	protected void translateMethod(Interface i, Operation op) throws WellformedException {
		for (Parameter p : op.getOwnedParameters())
			translateParameter(p);
	}

	abstract protected void translateDesign(Class c, DESIGN d);
	abstract protected T1 translateField(Classifier c, Property p) throws ExportException, RecognitionException;
	abstract protected T1 translateParameter(Parameter p) throws WellformedException;
	abstract protected T1 translateGeneralization(Classifier c);
	abstract protected T1 translateSignature(TemplateSignature ownedTemplateSignature);
	abstract protected T2 translateComponent(Component e) throws ExportException, WellformedException, RecognitionException;
	abstract protected T1 translateContract(ContractInterface ctr);
	abstract protected T1 translateInterface(Interface i) throws ExportException, WellformedException, RecognitionException;
	abstract protected T1 translateImportedClasses(Model model) throws ExportException;
	
	/*
	 * 
	 */
	protected class Composition {
		final public Component requiredComponent;
		final public ContractInterface clientInterface;
		
		final public Component providedComponent;
		final public ContractInterface supplierInterface;
		
		public Composition(Component rC, ContractInterface cI, Component pC, ContractInterface sI) {
			this.requiredComponent = rC;
			this.clientInterface = cI;
			this.providedComponent = pC;
			this.supplierInterface = sI;
		}
	}
	
	protected List<Composition> findComponentComposition(Component comp) throws WellformedException {
		List<Composition> result = new ArrayList<Composition>();
		for (Element e : comp.getOwnedElements()){
			edu.unu.iist.rcos.profile.Composition composition;
			if (e instanceof Dependency && (composition = Stereotypes.isComposition((Dependency) e)) != null){
				ContractInterface clientInterface = composition.getRequired();
				Component requiredComponent = composition.getRequiredInterfaceComponent();
				ContractInterface supplierInterface = composition.getProvided();
				Component providedComponent = composition.getProvidedInterfaceComponent();
				if (requiredComponent != null && providedComponent != null) {
					result.add(new Composition(requiredComponent, clientInterface, providedComponent, supplierInterface));
				} else
					assert false; // XXX
			}
		}
		return result;
	}
	
	protected Class findInterfaceImplement(Interface i, Component comp){
		for (Dependency d: comp.getClientDependencies()){
			boolean flag = false;
			for (NamedElement ne : d.getSuppliers()){
				if (ne.getName().equals(i.getName())){
					flag = true;
					break;
				}
			}
			if (flag == false)
				continue;
			for (NamedElement ne : d.getClients()){
				if (ne instanceof Class && !(ne instanceof Component))
					return (Class)ne;
			}
		}
		return null;
	}

	/*
	 * parameter Component: input
	 * parameter providedByClass
	 * parameter providedByDelegation
	 * parameter interfaceNameMapping: output, a contract interface's name may be different from the interface it references.
	 * parameter interfaceImplementMapping: output
	 * parameter interfaceDelegationMapping: output
	 */
	protected void findProvidedInterfaces(Component comp, List<Interface> providedByClass, List<Interface> providedByDelegation, Hashtable<String, Class> interfaceImplementMapping, Hashtable<String, Entry<Component, Interface>> interfaceDelegationMapping) throws ExportException{
		for (Interface i: comp.getProvideds()){
			Interface ri = null;
			ContractInterface ctr = Stereotypes.isContractInterface(i);
			if (ctr == null)
				ri = i;
			else {
				//get the interface of a contract
				Interface bi = ctr.getInterface();
				if (bi == null)
					ri = i;
				else {
					ri = bi;
				}
			}
			if (ri == null)
				continue;
			Class cl = findInterfaceImplement(i, comp);
			if (cl == null) {
				//the interface is not implemented by class
				//then we check if it is a delegation
				Entry<Component, Interface> delegation = findInterfaceDelegation(i, comp);
				if (delegation == null)
					continue;
				if (providedByDelegation.contains(ri))
					continue;
				providedByDelegation.add(ri);
				interfaceDelegationMapping.put(ri.getName(), delegation);
			}
			else {
				if (providedByClass.contains(ri))
					continue;
				providedByClass.add(ri);
				interfaceImplementMapping.put(ri.getName(), cl);
			}
		}
	}
	
	protected Entry<Component, Interface> findInterfaceDelegation(Interface i, Component comp) throws ExportException{
		Hashtable<Component, Interface> result = new Hashtable<Component, Interface>();
		
		for (Element e : comp.getOwnedElements()){
			if (!(e instanceof Dependency))
				continue;
			
			Dependency d = (Dependency) e;
			Delegation del;
			if ((del = Stereotypes.isDelegation(d)) == null)
				continue;
			try {
				final ContractInterface i3 = del.getClient();
				// We check that the delegation corresponds to the outside interface 
				Delegation dl = Stereotypes.isDelegation((Dependency) e);
				Interface i2 = dl.getInterface().getBase_Interface(); 
				if ( i2.getName().equals(i.getName())){
					Component co = i3.getProvidedInterfaceComponent(comp); // XXX
					if (co == null)
						throw new ExportException("Cannot find the inner component providing the interface " + i3.getBase_Interface().getLabel());
					result.put(co, i3.getBase_Interface());
					break;
				}
			} catch (WellformedException e1) {
				throw new ExportException(e1);
			}
		}
		
		assert result.entrySet().size() <= 1; // TODO: Review
		if (result.entrySet().size() > 0)
			return result.entrySet().iterator().next();
		else return null;
	}
}