package edu.unu.iist.rcos.modeler.translation;

import java.io.ByteArrayInputStream;
import java.io.UnsupportedEncodingException;
import java.util.*;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.uml2.uml.*;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Enumeration;
import org.eclipse.uml2.uml.Package;

import edu.unu.iist.rcos.mm.ocl.Class_OCL;
import edu.unu.iist.rcos.mm.ocl.Component_OCL;
import edu.unu.iist.rcos.mm.ocl.RuleI;
import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.profile.*;
import edu.unu.iist.rcos.spec.antlr.FreshNameGenerator;
import edu.unu.iist.rcos.spec.antlr.TypeHandler;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.TYPE;

import static edu.unu.iist.rcos.mm.util.Utility.*;

public class RCOSpecTranslation extends Translation {

	private final static String COMPONENT_MODEL = "_ComponentModel";

	public final static String NEW_LINE = "\n";

	public final static String TAB = "\t";

	public final static String BLANK = " ";

	public final static String LBRACKET = "{";

	public final static String RBRACKET = "}";

	public final static String INTERFACE = "interface ";

	public final static String CONTRACT = "contract ";

	public final static String OF = "of ";

	public final static String PROVIDED_INTERFACE = "provided ";

	public final static String REQUIRED_INTERFACE = "required ";

	public final static String IMPLEMENTED_BY = "by ";

	public final static String SERVICE_COMPONENT = "component";

	public final static String PROCESS_COMPONENT = "process";

	public final static String PUBLIC = "public";

	public final static String PRIVATE = "private";

	public final static String PROTECTED = "protected";

	public final static String SC = ";";

	public final static String LLBRACKET = "(";

	public final static String RLBRACKET = ")";

	public final static String EXTENDS = "extends";

	public final static String IMPLEMENTS = "implements";	

	public final static String CLASS = "class";

	private final static String INVARIANT = "invariant";	

	public final static String PROTOCOL = "protocol";

	public final static String COMPOSITION = "composition";

	public final static String COMPOSITION_STEREOTYPE = "Composition";

	public final static String DELEGATION_STEREOTYPE = "Delegation";

	public final static String LOCATIONS = "locations";

	public final static String INIT = "init";

	public final static String ACTION = "action";

	public final static String TRANSITION = "->";

	public final static String CONF_FILE = ".loc";

	public final static String IMPORT_SEP = "::";

	private static final boolean RCOS_MARKER_DELEGATION_OWNER = true;

	private TypeHandler th;

	// A list containing all the class realizing an interface
	// i.e. the classes defined in a component 
	//	private ArrayList<Class> listOfRealizingClasses = new ArrayList<Class>();

	private static void getTab(StringBuffer str, int tab){
		for (int i = 0 ; i < tab; i++){
			str.append("\t");
		}
	}

	private static void append(StringBuffer buf, String str, int tab){
		getTab(buf,tab);
		buf.append(str);
	}

	public void translate(RCOSModel model, IFile file, IProgressMonitor monitor) throws CoreException, ExportException {
		try {
			String source = translate(model);
			ByteArrayInputStream bis = new ByteArrayInputStream(source.getBytes("utf-8"));

			if (!file.exists()){
				file.create(bis, true, monitor);
			} else {
				file.setContents(bis, true, true, monitor);
			}
			file.setCharset("utf-8", monitor);
		} catch (UnsupportedEncodingException e) {
			throw new ExportException(e);
		}
	}

	public String translate(RCOSModel rm) throws ExportException {
		try {
			th = new TypeHandler(rm);
			// From here we are sure it is a requirement model
			Model model = rm.getBase_Model();
			String modelName = model.getName();

			StringBuffer spec = new StringBuffer();

			ComponentModel com = rm.getComponentModel();		
			Package compPkg = com.getBase_Package();

			if (compPkg == null)
				throw new ExportException("Inconsistent environment: couldn't get Package:"+modelName+COMPONENT_MODEL);

			// We first translate the import statements 
			for (Package p : rm.getBase_Model().getImportedPackages()) {
				// If it's a package from the standard lib
				StringBuffer imp = new StringBuffer();
				URI uri = p.eResource().getURI();
				// TODO: String literal
				if (uri.scheme().equals("pathmap")) {
					if(uri.devicePath().startsWith("//RCOS_LIBRARIES/")) {
						String filename = uri.lastSegment();
						// Get rid of the extension
						imp = new StringBuffer("lib" + IMPORT_SEP + filename.substring(0, filename.length() - 4)); 
					} else {
						throw new ExportWellformedException("Import statement not supported: " + uri.toString(),p);
					}
				} else if (uri.scheme().equals("platform")){					
					List<String> filename = uri.segmentsList();
					boolean first = true;
					for (String s : filename) {
						if (first)
							first = false;
						else
							imp.append(s + IMPORT_SEP);
					}
					imp = new StringBuffer(imp.substring(0, imp.length() - IMPORT_SEP.length())); 
				} else 
					throw new ExportWellformedException("Import statement not supported: " + uri.toString(),rm.getBase_Model());
				spec.append("import " + imp + SC + NEW_LINE + NEW_LINE);
			}

			// We then translate all the interfaces
			for (Element e : compPkg.allOwnedElements()){
				final Interface i;
				if (e instanceof Interface && Stereotypes.isContractInterface(i = (Interface) e) == null){
					translateInterface(i, spec, 0);
				}
			}

			// Then the different contracts
			for (PackageableElement e : compPkg.getPackagedElements()){
				final ContractInterface ctr;
				if (e instanceof Interface && (ctr = Stereotypes.isContractInterface((Interface) e)) != null){
					translateContract(ctr, spec, 0);
				}
			}
			// Then we translate each component
			for (PackageableElement e : compPkg.getPackagedElements()){
				if (e instanceof Component){
					translateComponent((Component)e, spec, 0);
					spec.append(NEW_LINE);
					spec.append(NEW_LINE);
				}
			}
			// At this point, everything is translated in the buffer spec, 
			// except the classes which are in the class diagram but not in 
			// any component.
			// All the classes inside a component have been added to 
			// listOfRealizingClasses with the method translateComponentClass

			// We get the class diagram and call the method translateOtherClasses
			// which does everything, including the definition of the "dummy" component
			ClassModel cm = rm.getClassModel();

			translateOtherClasses(cm.getBase_Package(), spec, 0);

			//System.out.println(spec);

			return spec.toString();
		} catch (WellformedException ex) {
			throw new ExportException(ex);
		}
	}

	/**
	 * @author Charles Morisset
	 * @param cl an Interface or Class
	 * @return the string " extends Cl1, Cl2, ..., Cln " or the empty string when no generalization. 
	 */
	private String translateGeneralization(Classifier cl) {
		String extendString = "";
		boolean first = true;
		for (Generalization g : cl.getGeneralizations()) {
			String current = g.getGeneral().getName();
			if(current != null && !"".equals(current) && !current.equals("Object")){ // XXX use PrimitiveTypes
				if (first)
					first = false;
				else
					extendString += ", ";
				extendString += current;
			}
		}
		if(!"".equals(extendString)){
			extendString = " " + EXTENDS + " " + extendString;
		}
		return extendString;
	}

	private void translateInterface(Interface i, StringBuffer buf, int tab) throws WellformedException, ExportException {
		append(buf, INTERFACE + i.getName(), tab);
		buf.append(translateGeneralization(i));
		buf.append(BLANK + LBRACKET + NEW_LINE);	

		translateFields(i, buf, tab);
		translateMethods(i, buf, tab, false);
		append(buf, RBRACKET + NEW_LINE + NEW_LINE, tab);
	}

	private void translateContract(ContractInterface ctr, StringBuffer buf, int tab) throws WellformedException, ExportException {
		// i the UML element representing the contract NOT the type of the contract
		Interface i = ctr.getBase_Interface();

		Interface type = ctr.getInterface();
		if (!Component_OCL.CONTRACTINTERFACE_HAS_INTERFACE_R.check(ctr))
			throw new ExportWellformedException(Component_OCL.CONTRACTINTERFACE_HAS_INTERFACE_R, ctr.getBase_Interface());
		append(buf, CONTRACT + i.getName() + BLANK + OF + type.getName(), tab);
		buf.append(translateGeneralization(i));
		buf.append(BLANK + LBRACKET);

		translateFields(i, buf, tab);
		translateMethods(i, buf, tab, true);

		if (ctr.getProtocol() != null) {
			if (ctr.getProtocol().getSequencediagram() != null){

				append(buf, "/* Protocol */" + NEW_LINE, tab + 1);

				Collaboration c = ctr.getProtocol().getSequencediagram();
				if (c.getOwnedBehaviors().size() > 0 && c.getOwnedBehaviors().get(0) instanceof Interaction){
					Interaction inter = (Interaction)c.getOwnedBehaviors().get(0);
					if (inter.getOwnedRule("expression") != null){
						append(buf, PROTOCOL + BLANK + LBRACKET + BLANK, tab + 1);
						buf.append(((OpaqueExpression)inter.getOwnedRule("expression").getSpecification()).getBodies().get(0));
						buf.append(BLANK + RBRACKET + NEW_LINE);
					}
				}
			}
		}
		append(buf, RBRACKET + NEW_LINE + NEW_LINE, tab);

	}	

	/**
	 * @author Charles Morisset
	 * @param o a UML Operation
	 * @param buf the current StringBuffer
	 * @param tab the tabulation
	 * @param withDesign is true if the body of the method also has to be translated, false is just the method declaration is needed
	 * @throws WellformedException
	 * @throws ExportException 
	 */
	private void translateOperation(Operation o, StringBuffer buf, int tab, boolean withDesign) throws WellformedException, ExportException {

		String pure = o.isQuery() ? ("pure"+BLANK) : "";
		append(buf, PUBLIC + BLANK + pure + o.getName() + LLBRACKET, tab);

		translateParaList(o.getOwnedParameters(), buf);
		buf.append(RLBRACKET);

		if (! withDesign){
			buf.append(SC + NEW_LINE);
		} else {
			String des = Designs.iNSTANCE.get(o); // don't use directly, see below!
			if (des != null){

				// RCOSTOOL-115
				// fix linefeeds from SWT specification tab
				// Also indent correctly.				
				{
					StringBuffer desBuf = new StringBuffer();
					des = des.replace('\r', '\n');
					for (String l : des.split("\n")) {
						append(desBuf,l+'\n',tab+1); // must match tab+n below
					}
					// first line without TABs:
					des = desBuf.toString().trim();
				}

				buf.append(LBRACKET + NEW_LINE);

				{ // print a guard if present
					String g = Guards.iNSTANCE.get(o);
					if (g != null && !"".equals(g))
						append(buf, g + " & " + des + NEW_LINE, tab+1);
					else
						append(buf, des + NEW_LINE, tab+1);
				}	

				append(buf, RBRACKET + NEW_LINE, tab);
			} else {
				final String g;
				if ((g = Guards.iNSTANCE.get(o)) == null || (g != null && "".equals(g)))
					buf.append(LBRACKET + RBRACKET + NEW_LINE);
				else {
					buf.append(LBRACKET + NEW_LINE);
					append(buf, g + " & skip" + NEW_LINE, tab+1);
					append(buf, RBRACKET + NEW_LINE, tab);
				}
			}
		}
	}

	private void translateProperty(Property p, StringBuffer buf, int tab) throws ExportException,WellformedException {
		if (p.getVisibility() == null) 
			append(buf, PUBLIC, tab);
		else
			append(buf, p.getVisibility().getName(), tab);

		buf.append(BLANK);

		if (p.getType() == null)
			throw new ExportWellformedException("No type for:"+p.getQualifiedName()+"["+p.getLower()+".."+p.getUpper()+"]",p);

		{
			// Handle multiplicity
			buf.append(parameterTypeWithMultiplicity(p.getName(), p, p.getType()));

			buf.append(BLANK + p.getName());
			if (p.getDefault() != null )
				buf.append(" = " + p.getDefault());
			buf.append(SC + NEW_LINE);
		}
	}

	private void translateComponent(Component comp, StringBuffer buf, int tab) throws WellformedException, ExportException{

		if (Stereotypes.isServiceComponent(comp) != null) {
			append(buf, SERVICE_COMPONENT, tab);
		} else if (Stereotypes.isProcessComponent(comp) != null){
			append(buf, PROCESS_COMPONENT, tab);
		}

		buf.append(BLANK + comp.getName() + BLANK + LBRACKET);
		buf.append(NEW_LINE + NEW_LINE);

		// We first translate the composition 
		translateComponentComposition(comp, buf, tab + 1);

		{
			// Then the invariant
			String invStr = Invariants.iNSTANCE.get(comp);
			if (invStr != null){
				if (invStr.length() > 0){
					String[] strList = invStr.split(";");
					for (int i=0; i < strList.length; i++){
						append(buf, INVARIANT + " : " + strList[i], tab + 1);
						buf.append(";" + NEW_LINE + NEW_LINE);
					}
				}
			}
		}

		// Then all the provided interfaces
		translateProvidedInterfaces(comp, buf, tab + 1);

		// Then all the required interfaces 
		translateRequiredInterfaces(comp, buf, tab + 1);

		// Then all the declared interfaces inside the component
		//		translateComponentInterfaces(comp, buf, tab + 1);
		for (Element e : comp.getOwnedElements()){
			if (e instanceof Interface){
				ContractInterface ctr = Stereotypes.isContractInterface((Interface) e);
				if (ctr == null){
					translateInterface((Interface) e, buf, tab + 1);
				} else {
					translateContract(ctr, buf, tab + 1);
				}
			}
		}

		// Then all the classes
		translateComponentClass(comp, buf, tab + 1);

		// Then all the nested component
		//		for (Element p : comp.getOwnedElements()){
		for (Classifier c : comp.getNestedClassifiers()){
			if (c instanceof Component){
				translateComponent((Component)c, buf, tab + 1);
			}
		}

		append(buf, RBRACKET + NEW_LINE + NEW_LINE, tab);
	}

	//	/**
	//	 * Utility method for translateComponentInterfaces
	//	 * @param name : the name of the interface 
	//	 * @param l : the list of interfaces
	//	 * @return the first interface in the list with the name "name", null otherwise
	//	 */
	//	private Interface getInterfaceWithName(String name, List<Interface> l){
	//		for (Interface i : l){
	//			if (i.getName()().equals(name))
	//				return i;
	//		}
	//		return null;
	//	}

	//	/**
	//	 * Translate all the interfaces defined inside a component.
	//	 * If two interfaces have the same name, then only the non-empty is translated. If they are both non-empty, an 
	//	 * RuntimeException is thrown.
	//	 * @param comp : a component
	//	 * @param buf : the string buffer
	//	 * @param tab : the tabulation
	//	 */
	//	private void translateComponentInterfaces(Component comp, StringBuffer buf, int tab) {
	//		// We first store in a list l all the interfaces of the component, such that there isn't two 
	//		// different interfaces with the same name
	//		List <Interface> l = new ArrayList<Interface>();
	//		for (Element e : comp.getOwnedElements()){
	//			if (e instanceof Interface){
	//				Interface i = (Interface) e;
	//				Interface i2 = getInterfaceWithName(i.getName()(), l); 
	//				if ( i2 != null){
	//					// There is already an interface with the same name
	//					if (i2.getOperations().size() > 0) {
	//						if (i.getOperations().size() > 0)
	//							throw new RuntimeException("Interface " + i.getName()() + " is defined twice"); // WF
	//						} else if (i.getOperations().size() > 0) {
	//							l.remove(i2);
	//							l.add(i);
	//						}
	//				} else {
	//					l.add(i);
	//				}
	//			}
	//		}
	//		// Then we actually translate all the stored interfaces.
	//		for (Interface i3 : l){
	//			translateInterface(i3, buf, tab);
	//		}
	//	}

	private void translateRequiredInterfaces(Component comp, StringBuffer buf, int tab) throws ExportWellformedException {
		// The delegated required interfaces are managed in translateComponentComposition
		try {
			Collection<String> reqs = new HashSet<String>();
			for (ContractInterface ci : Utility.getRequiredContracts(comp)) {
				String name = ci.getBase_Interface().getName();
				append(buf, REQUIRED_INTERFACE + name + SC + NEW_LINE, tab);
				if (!reqs.add(name)) // WF RCOSTOOL-229
					throw new ExportWellformedException("Component "+comp.getQualifiedName()+" unfortunately requires the same interface multiple times, sorry :-/",comp);
			}
		} catch (WellformedException e1) {
			throw new ExportWellformedException("Component "+comp.getQualifiedName()+":\n"+e1.getMessage(),e1,comp);
		}
		//		for (Element e : comp.getOwnedElements()){
		//			if (e instanceof Usage){
		//				Usage u = (Usage) e;
		//				// We get the interface, which is the supplier
		//				if (u.getSuppliers().size() != 1) throw new ExportException("The Usage " + u.getName() + " doesn't have exactly one supplier" );
		//				else{
		//					Interface i = (Interface) u.getSuppliers().get(0);
		//					append(buf, REQUIRED_INTERFACE + i.getName() + SC + NEW_LINE + NEW_LINE, tab);
		//				}
		//
		//			}
		//		}

	}

	private void translateProvidedInterfaces(Component comp, StringBuffer buf, int tab) throws WellformedException {
		try {
			for (InterfaceRealization ir : comp.getInterfaceRealizations()){
				// We first get the interface, which is the supplier
				Interface i = Utility.getSupplier(ir);
				if (!Class_OCL.ELEMENT_MUST_HAVE_NAME_R.check(i))
					throw new ExportWellformedException(Class_OCL.ELEMENT_MUST_HAVE_NAME_R.getDescription(),i);
				append(buf, PROVIDED_INTERFACE + i.getName() + BLANK + IMPLEMENTED_BY, tab);
				// Then we have to go through the clients which can be either 
				// - the component and the class implementing the interface
				// - the component and the port and the class implementing the interface
				// - the component and only the port if it is a delegated interface
				Class c = Utility.getInterfaceImplementingClass(ir);
				if (c != null)
					buf.append(c.getName() + SC + NEW_LINE);
				else {
					boolean appended = false;
					//case 3
					// We look through the delegations of the component to identify which one is connected to the port
					Delegation dl = getDelegation(comp,ir);

					// We check that the delegation corresponds to the outside interface
					if (!Component_OCL.DELEGATION_HAS_INTERFACE_R.check(dl))
						throw new ExportWellformedException(Component_OCL.DELEGATION_HAS_INTERFACE_R.getDescription(), dl.getBase_Dependency());
					Interface i2 = dl.getInterface().getBase_Interface(); 
					if ( i2.getName().equals(i.getName())){ // XXX: Equality-by-name?
						buf.append(dl.getQualifiedProvidedInterfaceName() + SC + NEW_LINE + NEW_LINE);
						appended = true;
						// XXX break???
					} else
						assert false : "While rendering "+ir.getQualifiedName()+"\n"+i2.getQualifiedName() +" <-> "+i.getQualifiedName(); // XXX

					if (!appended)
						// Component name is added by outer catch.
						throw new WellformedException("Looks like I forgot to print the provided interface for "+ir.getQualifiedName()+" :-(\n Did "+comp.getQualifiedName()+" forget to declare any Delegations (or they are in the wrong place) or the implementing class?");
				}
			}

		} catch (WellformedException e2) {
			throw new ExportWellformedException(e2,comp);
		}
	}

	private Delegation getDelegation(Component c, InterfaceRealization ir) throws WellformedException {
		assert ir.getContract() != null;
		// TODO: I'm not sure if we should use the type of ports in this way...
		Port port = c.getOwnedPort(null, ir.getContract());
		if (port == null)
			throw new WellformedException("No port with contract "+ir.getContract().getQualifiedName()+" in "+c.getQualifiedName());
		Delegation dl = null;
		for (DirectedRelationship d : port.getTargetDirectedRelationships(UMLPackage.eINSTANCE.getDependency())) {
			Delegation dl2;
			if ((dl2 = Stereotypes.isDelegation((Dependency) d)) != null) {
				if (dl != null)
					throw new WellformedException("Two dependencies?!");
				else
					dl = dl2;
				// TODO: OCL + break
			}
		}
		if (dl != null) {
			assert dl.getBase_Dependency().getOwner() == c;
		}
		return dl;
	}

	public static void translateComponentComposition(Component comp,
			StringBuffer buf, int tab) throws WellformedException {

		// If the component is atomic (no nested component), then return
		if (checkAtomicComponent(comp)) return;

		// At this point, there is at least one nested component.
		// We first translate : composition : (C1 | C2 | ... | Cn)
		append(buf, COMPOSITION + " : ", tab);

		// first is only used for pretty-printing issues
		boolean first = true;
		for (Classifier p : comp.getNestedClassifiers()){
			if (p instanceof Component){
				if (first) {
					buf.append(p.getName());
					first = false;
				} else {
					buf.append(" | " + p.getName());
				}
			} 
		}
		buf.append(NEW_LINE);

		// We can now translate the "interfaces plugging" : [(C1.I1 <- C2.I2), ...]
		// Each pair (C1.I1 <- C2.I2) corresponds to one composition.
		append(buf, "              [", tab);

		first = true;
		for (Element e : comp.getOwnedElements()){
			Composition composition;
			if (e instanceof Dependency && (composition = Stereotypes.isComposition((Dependency) e)) != null){

				if (!first){ 
					//					buf.append("," + NEW_LINE);
					buf.append("" + NEW_LINE);
					append(buf, "               ", tab);
				}
				buf.append("(");

				// The client is the required interface (C1.I1)
				// there should be exactly one.
				{
					final Component co = composition.getRequiredInterfaceComponent();
					assert co != null : composition.getBase_Dependency().getQualifiedName(); // WF XXX
					buf.append(co.getName() + "." + composition.getRequiredName());
				}
				buf.append(" <- ");

				// The supplier is the provided interface (C2.I2)
				// there should be exactly one.
				{
					final Component co = composition.getProvidedInterfaceComponent();
					if (co == null)
						buf.append(composition.getProvidedName());
					else
						buf.append(co.getName() + "." + composition.getProvidedName());
				}
				buf.append(")");
				first = false;
			}	
		}

		first = true;
		// We now translate the delegated required interfaces
		for (Element e : comp.getOwnedElements()){
			if (e instanceof Dependency && Stereotypes.isDelegation((Dependency) e) != null){
				if (!first){ 
					//					buf.append("," + NEW_LINE);
					buf.append("" + NEW_LINE);
					append(buf, "               ", tab);
				} 

				Dependency dp = (Dependency) e;
				Delegation dl = Stereotypes.isDelegation(dp);

				// We need to check that this delegation is for a required interface. 
				// so we see if the port attached to the delegation use the interface of the delegation
				for (Dependency d2 : dl.getSupplier().getClientDependencies()){
					if (d2 instanceof Usage){
						Usage u = (Usage) d2;
						Interface i2 = getSupplier(u);
						assert i2 != null;
						ContractInterface ci2 = Stereotypes.isContractInterface(i2);
						assert ci2 != null;
						//						if (i2.getName().equals(dl.getInterface().getBase_Interface().getName())){
						if (i2 == dl.getInterface().getBase_Interface()){
							ContractInterface i = dl.getClient();
							Component co = dl.getRequiredInterfaceComponent();
							buf.append("(" + co.getName() + "." + i.getBase_Interface().getName() + " <- " + i2.getName() + ")");
						}
					}
				}
			}
		}
		buf.append("]" + NEW_LINE + NEW_LINE);
	}

	/**
	 * Also print nested components
	 */
	public static void translateComponentCompositionHierarchical(Component comp,
			StringBuffer buf, int tab) throws WellformedException {
		if (Utility.checkAtomicComponent(comp)) {
			buf.append(comp.getName()+" = <<atomic>>\n");
			return;
		} else {
			buf.append(comp.getName()+" =\n");
			translateComponentComposition(comp, buf, tab);
			for (Classifier p : comp.getNestedClassifiers()){
				if (p instanceof Component){
					translateComponentCompositionHierarchical((Component) p, buf, tab);
				}
			}
		}
	}

	//	private void translateComponentClass(Component comp, StringBuffer buf, int tab) throws ExportException, WellformedException {
	//
	//		if (comp.getRealizations().size() == 0) return;
	//
	//		ComponentRealization cr = comp.getRealizations().get(0);
	//
	//		for ( NamedElement ne : cr.getSuppliers()){
	//			if (ne instanceof Class) {
	//				Class cl = (Class) ne;
	//				if (cl.getVisibility().getName().equals("private")) {
	//					listOfRealizingClasses.add((Class)ne);
	//					translateClass((Class)ne, buf, tab);
	//					append(buf, NEW_LINE + NEW_LINE, tab);
	//				}
	//			}
	//		}
	//
	//	}

	private void translateComponentClass(Component comp, StringBuffer buf, int tab) throws WellformedException, ExportException {
		for (Element e : comp.getOwnedElements()) {
			if (e instanceof Class && !(e instanceof Component)) {
				Class cl = (Class) e;
				if (! cl.getVisibility().equals(VisibilityKind.PRIVATE_LITERAL))
					throw new WellformedException("The component " + comp.getLabel() + " contains a public class. " +
					"Either change the visibility of the class to private or move the class to the ClassModel package");
				translateClass(cl, buf, tab);
			}
		}
	}
	private void translateClass(Class c, StringBuffer buf, int tab) throws WellformedException, ExportException {
		StringBuffer b = new StringBuffer();

		// If it's a binding, do not translate
		if (c.getEAnnotation(Utility.BINDING_DNS) != null) return;

		if (c.getVisibility() == null) 
			b.append(PUBLIC);
		else
			b.append(c.getVisibility().getName());

		b.append(BLANK + CLASS + BLANK + c.getName());

		b.append(translateSignature(c.getOwnedTemplateSignature()));
		b.append(translateGeneralization(c));

		if (c.getInterfaceRealizations().size() > 0){
			b.append(BLANK + IMPLEMENTS + BLANK);

			for (int i = 0; i < c.getInterfaceRealizations().size(); i++){
				InterfaceRealization ir = c.getInterfaceRealizations().get(i);
				if (i > 0)
					b.append(", ");
				b.append(ir.getSuppliers().get(0).getName());
			}
		}

		b.append(BLANK + LBRACKET + NEW_LINE);

		append(buf, b.toString(), tab);


		{
			// Generate invariant specification
			String invStr = Invariants.iNSTANCE.get(c);
			if ( invStr != null){
				if (invStr.length() > 0){
					buf.append(NEW_LINE);
					String[] strList = invStr.split(";");
					for (int i=0; i < strList.length; i++){
						append(buf, INVARIANT + " : " + strList[i] + SC , tab + 1);
						buf.append(NEW_LINE + NEW_LINE);
					}
				}
			}
		}

		translateFields(c, buf, tab);
		translateMethods(c, buf, tab, true);

		append(buf, RBRACKET + NEW_LINE + NEW_LINE, tab);
	}

	/**
	 * @author Charles Morisset
	 * @param c a classifier
	 * @param buf the current StringBuffer
	 * @param tab the tabulation
	 * @param withDesign is true if the body of the method also has to be translated, false is just the method declaration is needed
	 * @throws WellformedException
	 * @throws ExportException 
	 */
	private void translateMethods(Classifier c, StringBuffer buf, int tab, boolean withDesign) throws WellformedException, ExportException {
		if (c.getOperations().size() > 0) {
			buf.append(NEW_LINE);
			append(buf, "/* Methods */" + NEW_LINE, tab + 1);
			for (Operation o : c.getOperations()){
				translateOperation(o, buf, tab + 1, withDesign);
			}
			buf.append(NEW_LINE);
		}
	}

	private void translateFields(Classifier c, StringBuffer buf, int tab) throws WellformedException, ExportException {

		EList<Property> l = new BasicEList<Property>();
		l.addAll(c.getAttributes());

		//Generate Association
		for (Association a : c.getAssociations()){
			for (Property p : a.getMemberEnds()){
				if (!p.getType().getName().equals(c.getName())){
					if (a.getNavigableOwnedEnds().indexOf(p) >= 0){
						//means it is a field
						l.add(p);
					}
				}
			}
		}

		if (l.size() > 0) {
			append(buf, "/* Fields */" + NEW_LINE, tab + 1);
			for (Property p : l)
				translateProperty(p, buf, tab + 1);
		}
	}

	private String translateSignature(TemplateSignature ts) throws ExportException {
		String sig = "";
		if (ts != null && ts.getOwnedParameters().size() > 0) {
			boolean isFirst = true;
			sig += "<";
			for (TemplateParameter tp : ts.getOwnedParameters()) {
				if (isFirst) {
					sig += translateType((Type) tp.getOwnedParameteredElement());
					isFirst = false;
				} else {
					sig += SC + BLANK + translateType((Type) tp.getOwnedParameteredElement());
				}
			}
			sig += ">";
		}
		return sig;
	}

	/**
	 * @author Charles Morisset
	 * @param Type
	 * @return
	 * @throws WellformedException 
	 */
	public String translateType(Type t) throws ExportException {
		assert t != null;
		// TODO: this does not look too efficient...
		TYPE t1;
		try {
			t1 = th.createTYPE(t);
			return TypeHandler.printTYPE(t1);
		} catch (WellformedException e) {
			throw new ExportException(e);
		}
	}

	// XXX: generate the description of the state machine, as introduced by Zhenbang. 
	// Currently disconnected. 
	private void translateStateSpec(Class cls, Region region, StringBuffer buf, int tab, IFile file) throws WellformedException, ExportException {

		//first generate location declaration and map
		append(buf, LOCATIONS + " : " + LBRACKET, tab);

		Map<State, String> map = new HashMap<State, String>();

		boolean isFirst = true;
		int i = 1;
		Pseudostate s = null;
		for (Vertex v : region.getSubvertices()){
			if (v instanceof State){
				if (!isFirst){
					buf.append(", ");
				} else isFirst = false;
				String name = "l" + i; 
				buf.append(name);
				map.put((State)v, name);
				i++;

			}
			if (v instanceof Pseudostate){
				s = (Pseudostate)v;
			}
		}
		buf.append(RBRACKET + NEW_LINE + NEW_LINE);

		// Generate init specification
		if (s != null && s.getOutgoings().size() == 1){
			Transition t = s.getOutgoings().get(0);
			if (t.getTarget() instanceof State){
				append(buf, INIT + " : ", tab);
				{
					String g = Guards.iNSTANCE.get(t);
					if (g != null)
						buf.append(g);
					else
						buf.append("true");
				}
				buf.append(NEW_LINE);
				State initState = (State)t.getTarget();
				append(buf, " : ", tab + 1);
				buf.append(map.get(initState));
				buf.append(BLANK + SC + NEW_LINE + NEW_LINE);
			}
		}

		// Generate the actions
		for (Transition t : region.getTransitions()){
			Vertex source = t.getSource();
			Vertex target = t.getTarget();
			if (source instanceof State && target instanceof State){
				append(buf, ACTION + " : ", tab);
				{
					String g = Guards.iNSTANCE.get(t);
					if (g == null)
						buf.append("true");
					else
						buf.append(g);
				}

				buf.append(NEW_LINE);
				append(buf, BLANK + BLANK, tab + 1);
				// find operation by name
				Operation op = cls.getOperation(t.getName(), null, null);
				String str = "";
				if (op != null){
					StringBuffer sb = new StringBuffer();
					translateParaList(op.getOwnedParameters(), sb);
					str = t.getName() + LLBRACKET + sb.toString() + RLBRACKET;
				} else {
					str = t.getName() + LLBRACKET + RLBRACKET;
				}
				buf.append(" : " + str + NEW_LINE);
				append(buf, BLANK + BLANK + " : ", tab + 1);
				buf.append(map.get(source) + BLANK + TRANSITION + BLANK + map.get(target) + BLANK + SC);
				buf.append(NEW_LINE + NEW_LINE);
			}
		}

		// Generate the configuration file
		if (file != null){

			StringBuffer fileContent = new StringBuffer(); 

			for (Transition t : region.getTransitions()){
				Vertex source = t.getSource();
				Vertex target = t.getTarget();
				if (source instanceof State && target instanceof State){
					String key = map.get(source) + TRANSITION + map.get(target) + "-" + t.getName();
					fileContent.append(key + BLANK + t.eResource().getURIFragment(t) + NEW_LINE);
				} else if (source instanceof Pseudostate && target instanceof State) {
					fileContent.append(map.get(target) + BLANK + t.eResource().getURIFragment(t) + NEW_LINE);
				}
			}

			String confFileName = cls.getName() + CONF_FILE;
			IFile confFile = file.getParent().getFile(new Path(confFileName));

			ByteArrayInputStream bis = new ByteArrayInputStream(fileContent.toString().getBytes());

			try{
				if (!confFile.exists()){
					confFile.create(bis, true, new NullProgressMonitor());
				} else {
					confFile.setContents(bis, true, true, new NullProgressMonitor());
				}
			} catch (Exception e){
				e.printStackTrace();
				// XXX
			}
		}
	}

	private void translateParaList(List<Parameter> c, StringBuffer buf) throws WellformedException, ExportException{
		String strIn="", strOut="";
		boolean infirst = false, outfirst =false;
		for ( Parameter p : c){
			if (p.getDirection().getValue() == ParameterDirectionKind.IN){
				if (infirst) 
					strIn += ", ";
				else
					infirst = true;
				strIn +=
					translatePara(p);
			} else if (p.getDirection().getValue() == ParameterDirectionKind.OUT){
				if (outfirst) 
					strOut += ", ";
				else
					outfirst = true;
				strOut +=
					translatePara(p);
			} else if (p.getDirection().getValue() == ParameterDirectionKind.RETURN){
				if (outfirst) 
					strOut += ", ";
				else
					outfirst = true;
				strOut +=
					translatePara(p);
			} else
				throw new WellformedException(p.getQualifiedName()); // WF
		}

		if (strIn.length() > 0)
			buf.append(strIn);
		if (strOut.length() > 0)
			buf.append(SC + BLANK + strOut);
	}

	private String translatePara(Parameter p) throws WellformedException, ExportException {
		Type t;
		if ((t = p.getType()) == null)
			throw new WellformedException("The parameter " + p.getQualifiedName() + " does not have a type");
		assert (!(t instanceof Classifier) || ((Classifier)t).getOwnedTemplateSignature() == null) : "Generic parameter '"+t.getName()+"<..>'"+" on "+p.getQualifiedName()+" not allowed!"; // WF?
		String name = "";
		if (p.getName() == null || p.getName().equals("")) {
			// the name for the return type can be absent in the UML model, but not in rCOSp
			if (p.getDirection().getValue() == ParameterDirectionKind.RETURN) {
				FreshNameGenerator pg = new FreshNameGenerator(th,(Classifier) p.getOperation().getOwner());
				name = pg.get(t);
			} else
				throw new WellformedException("The parameter " + p.getQualifiedName() + " does not have a name");
		} else {
			name = p.getName();
		}
//		try {
			assert p.getType() != null;
			String type_name = parameterTypeWithMultiplicity(p.getName(),p,p.getType());
			assert type_name != null;
			return (type_name + BLANK + name); 
//		} catch (WellformedException e) {
//			throw new WellformedException("While translating parameter "+p.getQualifiedName()+":",e);
//		}
	}

	/**
	 * Dump package contents (sub-packages/classes only?)
	 * @author vs
	 * @param p
	 * @return
	 * @throws ExportException	 
	 */
	public String translatePackage(Package p) throws ExportException {
		StringBuffer buf = new StringBuffer();
		try {
			translateOtherClasses(p,buf,0);
		} catch (WellformedException e) {
			throw new ExportException(e);
		}
		return buf.toString();
	}

	/**
	 * @author charles
	 * @author vs
	 * @throws WellformedException 
	 */
	private void translateOtherClasses(Package p, StringBuffer buf, int tab) throws ExportException, WellformedException{
		StringBuffer temp = new StringBuffer();

		for (PackageableElement e : p.getPackagedElements()){
			//			if (e instanceof Class && ! listOfRealizingClasses.contains((Class)e)){
			if (e instanceof Class){
				translateClass((Class)e, temp, tab);
			} else if (e instanceof Package) {
				translateOtherClasses((Package)e, temp, tab);
			} else if ((e instanceof Comment) || (e instanceof Association)) {
				// Nada -- skip those
			} else if (e instanceof Enumeration) {
				Enumeration en = (Enumeration) e;
				buf.append("enum "+en.getName()+" { ");
				Iterator<EnumerationLiteral> it = en.getOwnedLiterals().iterator();
				while (it.hasNext()) {
					buf.append(it.next().getName());
					if (it.hasNext())
						buf.append(',');
				}
				buf.append("}\n\n");
			} else
				throw new ExportException("I don't know how to translate\n"+e.toString()+"\nto .rcosp, sorry :-(");
		}
		buf.append(temp);		
	}

	class ExportWellformedException extends WellformedException {
		private static final long serialVersionUID = 6447667054825276137L;

		public ExportWellformedException(String description, NamedElement e) {
			super((e.getQualifiedName() == null ? "<unnamed element>" : e.getQualifiedName())+": "+description);
		}
		public ExportWellformedException(RuleI rule, NamedElement e) {
			super((e.getQualifiedName() == null ? "<unnamed element>" : e.getQualifiedName())+": "+rule.getDescription());
		}
		public ExportWellformedException(Throwable t, NamedElement e) {
			super((e.getQualifiedName() == null ? "<unnamed element>" : e.getQualifiedName())+": "+t.getMessage(),t);
		}
		public ExportWellformedException(String description, Throwable t, NamedElement e) {
			super((e.getQualifiedName() == null ? "<unnamed element>" : e.getQualifiedName())+": "+description,t);
		}
	}
}