package edu.unu.iist.rcos.modeler.translation;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.uml2.uml.MultiplicityElement;
import org.eclipse.uml2.uml.Type;

import edu.unu.iist.rcos.modeler.translation.ExportException;
import edu.unu.iist.rcos.profile.RCOSModel;

public abstract class Translation {
	
	/**
	 * @param model RCOSModel to export.
	 * @param f File to write output to. If it exists, it will be overwritten.
	 * @throws CoreException
	 * @throws ExportException
	 */
	public abstract void translate(RCOSModel model, IFile f, IProgressMonitor monitor)  throws CoreException, ExportException;
	
	/**
	 * @param model RCOSModel to export.
	 * @return The textual representation of the model.
	 * @throws ExportException
	 */
	public abstract String translate(RCOSModel model) throws ExportException;

	protected abstract String translateType(Type t) throws ExportException;

	public String parameterTypeWithMultiplicity(String name, MultiplicityElement p, Type t) throws ExportException {
		// TODO: check stereotypes {ordered}...arrays?
		// 0..* => Set<T>
		if ((p.getLower() == 0) && (p.getUpper() == -1)) { // Set
			// TODO: check if isUnique() is true, to distinguish between Sets and Bags 
			return "Set<" + translateType(t)+">";
		}

		// 1..1 => T
		if ((p.getLower() == 1) && (p.getUpper() == 1)) {
			return translateType(t);
		}

		// 0..1 => Optional type, not handled yet
		if ((p.getLower() == 0) && (p.getUpper() == 1)) {
			return translateType(t);
		}
		throw new ExportException("Don't know how to generate declaration for:"+name+"/"+t.getName()+"["+p.getLower()+".."+p.getUpper()+"]");
	}
}