package edu.unu.iist.rcos.spec.antlr;

import java.util.*;

import org.apache.commons.collections.*;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.util.SyntaxTree_ClassModelSwitch;
import edu.unu.iist.rcos.spec.antlr.visitor.*;
import edu.unu.iist.rcos.spec.antlr.visitor.check.*;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.*;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement.*;
import edu.unu.iist.rcos.spec.antlr.visitor.check.ExpGetVariables;

/**
 * @author stolz
 *
 */
public class DesignUtility /* stupid static factories: implements DesignExpPrinter */ {

	public static Collection<Operation> getAllOperations(DESIGN d, Classifier cl) throws GetTypeException {
		return new DesignGetAllOperations(cl).run(d);
	}

	public static Collection<Operation> getAllOperations(EXP e, Classifier cl) throws GetTypeException {
		return new ExpGetAllOperations(cl).run(e); // is this safe? if an operation changes...
	}

	/**
	 * @author Charles Morisset
	 * @param op The operation to use as base for parameters and fields. Note that this operation must already be contained in the model!
	 * @param d a design
	 * @return the pretty-print of d using the rCOS syntax
	 */
	public static String designToString(final Operation op, DESIGN d) {
		assert op.getOwner() != null : "Operation is not yet contained in the model!";
		VardeclsFIS v = new VardeclsFIS(op);
		return v.designToString(d);
	}

	@Deprecated
	public static String designToString(TypeHandler th, final Operation op, DESIGN d) {
		assert op.getOwner() != null : "Operation is not yet contained in the model!";
		VardeclsFIS v = new VardeclsFIS(op,th);
		return v.designToString(d);
	}

	/**
	 * Factory to configure ExpPrinter
	 */
	public static boolean Unicode = true;

	/**
	 * Can't be used on elements not in the model.
	 */
	public static String designToString(final Classifier c, DESIGN d) {
		VardeclsFIS v = new VardeclsFIS(c);
		return v.designToString(d);
	}

	@Deprecated
	public static String designToString(DESIGN d) {
		class DP extends DesignPrinter<ExpPrinter> {
			public DP() { super(new ExpPrinter()); }
		}
		try {
			return new DP().run(d);
		} catch (PrinterException e) {
			// TODO: relevant for rCOS 123
			throw new RuntimeException("Couldn't print design",e);
		}		
	}

	/**
	 * @author Charles Morisset
	 * @param e an expression
	 * @return the pretty-print of e using the rCOS syntax
	 */
	@Deprecated
	public static String expToString(EXP e) { 
		try {
			return new ExpPrinter().run(e);
		} catch (PrinterException e1) {
			// TODO: relevant for rCOS 123
			throw new RuntimeException("Couldn't print expression",e1);
		}
	}
	
	public static String expToString(Operation op, EXP e) { 
		assert op != null;
		return new VardeclsFIS(op).expToString(e);
	}

	public static String expToString(Classifier c, EXP e) {
		assert c != null;
		return new VardeclsFIS(c).expToString(e);
	}

	/**
	 * @author Charles Morisset
	 * @param e an expression
	 * @return the pretty-print of e using the rCOS syntax
	 */
	public static String typeToString(TYPE e) {
		return TypeHandler.printTYPE(e);
	}

	/**
	 * @author Charles Morisset
	 * @param e an expression
	 * @return false is e contains a forall or exists, true otherwise
	 */
	public static boolean isSimpleExpression(EXP e) {
		return new ExpIsSimpleExpression().run(e,null);
	}

	/**
	 * @author Charles Morisset
	 * @param d a design
	 * @return the list of primed fields accesses, parameters or local variables of d
	 */
	public static Collection<DASHABLE> getPrimedVariables(DESIGN d) {
		return new DesignGetPrimedVariables().run(d,null);
	}

	/**
	 * @author Charles Morisset
	 * @param e an expression
	 * @return the list of primed fields accesses, parameters or local variables of e
	 */
	public static Collection<DASHABLE> getPrimedVariables(EXP e) {
		return new ExpGetPrimedVariables().run(e,null);
	}

	public static Collection<DASHABLE> getVariables(EXP e) {
		return new ExpGetVariables().run(e,null);
	}

	/**
	 * @author Charles Morisset
	 * @param d an expression
	 * @return true if e contains primed variables, false otherwise
	 */
	public static boolean hasPrimedVariables(EXP e) {
		return (!getPrimedVariables(e).isEmpty());
	}

	/**
	 * Get the contained free variables from an expression.
	 * Note that field accesses are prefixed by "this.".
	 * @author Charles Morisset, Volker Stolz
	 * @param e an expression
	 * @return the list of free identifiers (variables, parameters and field accesses) of e
	 */
	@SuppressWarnings("unchecked")
	public static Collection<String> getListId(EXP e) {
		Collection<DASHABLE> ds = new ExpGetVariables().run(e,null);
		return CollectionUtils.collect(ds, new Transformer(){
			SyntaxTree_ClassModelSwitch<String> s = new SyntaxTree_ClassModelSwitch<String>(){
				@Override public String caseVAR(VAR v) {
					return v.getName().getName();
				}
				@Override public String casePAR(PAR v) {
					return v.getName().getName();
				}
				@Override public String caseFA(FA v) {
					return DesignUtility.expToString(v);
				}
				@Override public String defaultCase(EObject eo) {
					throw new RuntimeException(); // NOT REACHED
				}
			};

			public String transform(Object input) {
				return s.doSwitch((DASHABLE)input);
			}});
	}

	/**
	 * @author Charles Morisset
	 * @param e1 an expression
	 * @param e2 an expression
	 * @return true if e1 and e2 are syntactically equals
	 */
	public static boolean isExpEqual(EXP e1, EXP e2){
		try {
			return new ExpEqual(e2, true).run(e1);
		} catch (VisitorException e) {
			// TODO: PO NotReached?
			throw new RuntimeException(e);
		}
	}

	/**
	 * @author Charles Morisset
	 * @param e1 an expression
	 * @param e2 an expression
	 * @return true if e1 and e2 are syntactically equals, without differentiating primed and unprimed variables 
	 */
	public static boolean isExpWeakEqual(EXP e1, EXP e2){
		try {
			return new ExpEqual(e2, false).run(e1);
		} catch (VisitorException e) {
			// TODO: PO NotReached?
			throw new RuntimeException(e);
		}
	}

	/**
	 * @author Charles Morisset
	 * @param d1 a design
	 * @param d2 a design
	 * @return true is d1 and d2 are syntactically equals
	 */
	public static boolean isDesignEqual(DESIGN d1, DESIGN d2){
		return new DesignEqual(d2, true).run(d1,false);
	}
	
	/**
	 * @author Charles Morisset
	 * @param d1 a design
	 * @param d2 a design
	 * @return true is d1 and d2 are syntactically equals, without differentiating primed and unprimed variables 
	 */
	public static boolean isDesignWeakEqual(DESIGN d1, DESIGN d2){
		return new DesignEqual(d2, false).run(d1,false);
	}

	/**
	 * @author Charles Morisset
	 * @param d a design
	 * @return a deep copy of d 
	 */
	public static DESIGN copyDesign(DESIGN d) {
		return (DESIGN) EcoreUtil.copy(d);
	}

	/**
	 * @author Charles Morisset
	 * @param e an expression
	 * @return a deep copy of e
	 */
	public static EXP copyExpression(EXP e) {
		return (EXP) EcoreUtil.copy(e);
	}

	/**
	 * @author Charles Morisset
	 * @param d a design
	 * @return a deep copy of d 
	 * @throws VisitorException
	 * @Deprecated Use {@link #simplify(DESIGN, Classifier)} instead 
	 */
	@Deprecated
	public static DESIGN simplifyDesign(DESIGN d) throws VisitorException {
		return new DesignSimplification().run(d);
	}

	/**
	 * @author Charles Morisset
	 * @param e an expression
	 * @return a deep copy of e
	 * @throws VisitorException 
	 */
	public static EXP simplifyExpression(EXP e) throws VisitorException {
		return new ExpSimplification().run(e);
	}

	public static DESIGN simplify(DESIGN d, Classifier cl) throws RefinementException {
		assert cl != null;
		assert cl.eContainer() != null : cl.getName() + " not contained in the model yet!";
		List<DesignRefinement<RefinementException, ?>> l_ref = new ArrayList<DesignRefinement<RefinementException, ?>>(); 
		l_ref.add(new DesignSimplification(cl));    // There's a strategy here: first simply the design, incl. normalizing SEQ-chains.
		// TODO: pattern does not respect scoping:
		// l_ref.add(new DesignPullUpIfSkipChaos(null)); // Then, let's see if there is anything to pull up.
		return iterate(d,cl,l_ref);
	}

	/**
	 * @author Charles Morisset
	 */
	public static DESIGN refine(DESIGN d, Classifier cl) throws RefinementException {
		List<DesignRefinement<RefinementException, ?>> l_ref = new ArrayList<DesignRefinement<RefinementException, ?>>();
		// First simplify. TODO: shared list?
		l_ref.add(new DesignSimplification(cl));    // There's a strategy here: first simply the design, incl. normalizing SEQ-chains.
//		l_ref.add(new DesignPullUpIfSkipChaos(null)); // Then, let's see if there is anything to pull up.

		l_ref.add(new DesignExtractPredicate(cl));
		l_ref.add(new DesignSplitPP(cl));
		l_ref.add(new DesignRelToAssign(cl));
		l_ref.add(new DesignRefineMethods(cl));
		l_ref.add(new DesignRefineAssert(cl));
		return iterate(d,cl,l_ref);
	}

	public static DESIGN iterate(DESIGN d, Classifier cl, List<DesignRefinement<RefinementException, ?>> l_ref) throws RefinementException {
		DESIGN d2 = copyDesign(d);

		// Just to prevent infinite loops
		int count = MAX_SIMPLIFICATIONS;
		
		boolean modified;
		do {
			modified = false;
			for (DesignRefinement<RefinementException, ?> des : l_ref) {
				/* [vs] is wondering how much state those visitors keep, and if they should be properly
				 * reset/reinstatiated here.
				 */
				des.unsetModified();
				d2 = des.run(d2);
				modified |= des.isModified();
			}
		} while(modified && count-- > 0); // modified is false iff none of l_ref modified the design
		if (modified)
			throw new RuntimeException("Number of iterations in design simplifier exceeded (currently set to "+MAX_SIMPLIFICATIONS+")."); // XXX
		return d2;
	}

	public static final int MAX_SIMPLIFICATIONS = 200;

	/**
	 * Warning: still experimental
	 * @author Charles Morisset
	 * @param d a design
	 * @return a copy of d where every set-related operation has been replaced by the corresponding operation over integers
	 * @throws VisitorException 
	 */
	@Deprecated // just used in testcase
	public static DESIGN abstractSetOpsDesign(DESIGN d) throws VisitorException {
		return new DesignAbstractionSetOps().run(d);
	}

	/**
	 * @author Charles Morisset
	 * @param e an expression
	 * @param cl the class in which e is defined
	 * @return the UML type of e 
	 * @throws TypingException
	 * @throws GetTypeException
	 * @deprecated Use {@link TypeHandler#getType(EXP, Classifier)} instead. 
	 */
	@Deprecated
	public static Type getTypeExp(EXP e, Classifier cl) throws GetTypeException {
		return new ExpGetType(cl).run(e);
	}

	/**
	 * Warning: the typing rules are still incomplete
	 * @author Charles Morisset
	 * @param d a design
	 * @param op the Operation to which the design belongs
	 * @return true if d is well-typed, false otherwise
	 */
	static boolean isWellTypedDesign(DESIGN d, Operation op){
		try {
			return new DesignCheckType(op).run(d,false);
		} catch (GetTypeException e) {
			return false;
		}
	}
	
	/**
	 * @author Charles Morisset
	 * @param d a design
	 * @return true if d is translatable (e.g. to Java), false otherwise
	 */
	static boolean isDesignTranslatable(DESIGN d){
		return new DesignIsTranslatable().run(d, false);		
	}

	/**
	 * @author Charles Morisset
	 * @param e an expression
	 * @param cl the class in which e is defined 
	 * @return true if e is of boolean type, false otherwise and throws an exception if e is not well-typed
	 * @throws TypingException
	 * @throws GetTypeException
	 * @deprecated Use {@link TypeHandler#checkBoolean(EXP, Classifier)} 
	 */
	public static boolean checkBoolean(EXP e, Classifier cl) throws GetTypeException {
		ExpGetType egt = new ExpGetType(cl);
		egt.visit(e);
		return (egt.isBooleanType()); 
	}

	/**
	 * @author Charles Morisset
	 * @param e an expression
	 * @param cl the class in which e is defined 
	 * @return true if e is of boolean type, false otherwise
	 * @deprecated Use {@link TypeHandler#isBooleanExp(EXP, Classifier)}
	 */
	public static boolean isBooleanExp(EXP e,  Classifier cl){	
		try {
			return checkBoolean(e, cl); 
		} catch (GetTypeException exc) {
			return false;			
		}
	}

	/**
	 * @author Charles Morisset
	 * @param type
	 * @param type2
	 * @return
	 */
	public static boolean isTypeEqual(TYPE type, TYPE type2) {
		return new TypeEqual(type).run(type2,false);
	}

	// Purity
	public static Set<String> isPure(DESIGN d, Classifier scope) throws GetTypeException {
		assert scope != null;
		final Collection<Operation> ops = DesignUtility.getAllOperations(d, scope);
		return getUnpureMCalls(ops);
	}

	public static boolean isPureSafe(EXP e, Classifier scope, boolean voidOk) {
		try {
			return isPure(e,scope, voidOk);
		} catch (GetTypeException e1) {
			return false;
		}
	}

	public static boolean isPure(EXP e, Classifier scope) throws GetTypeException {
		return isPure(e,scope,false /* XXX */);
	}

	public static boolean isPure(EXP e, Classifier scope, boolean voidOk) throws GetTypeException {
		assert scope != null;
		final Collection<Operation> ops = new ExpGetAllOperations(scope, voidOk).run(e);
		final Set<String> filteredList = getUnpureMCalls(ops);
		if (filteredList.size() > 0)
			throw new GetTypeException("The expression '" + DesignUtility.expToString(scope,e)  +"' calls non-pure method(s):\n"+filteredList, e);
		return true;
	}

	public static Set<String> getUnpureMCalls(Collection<Operation> ops)
			throws GetTypeException {
		final Set<String> filteredList = new HashSet<String>();

		for (Operation op : ops) {
			if (!op.isQuery())
				filteredList.add(op.getQualifiedName());
		}
		return filteredList;
	}
	// END Purity
	
	public static boolean isDeadBranch(DESIGN d) {
		return d instanceof CHAOS
		|| d instanceof RET
		|| d instanceof STOP;
	}

	public static String dexpToString(VardeclsFIS vdcls, DESIGNEXP de) {
		if (de instanceof DESIGN)
			return vdcls.designToString((DESIGN) de);
		else
			return vdcls.expToString((EXP) de);
	}
}
