package edu.unu.iist.rcos.spec.antlr;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Property;
import org.eclipse.uml2.uml.Type;

import edu.unu.iist.rcos.profile.PrimitiveTypes;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.parser.rCOSParser;
import edu.unu.iist.rcos.spec.antlr.visitor.check.DesignGetVarNames;

/**
 * @author charles
 * This class generates a name given a type, trying to avoid scoping or name conflict issues.
 *  It is used in the Expert-Pattern when generating new methods with parameters. 
 *
 */
public class FreshNameGenerator {
	
	 /**
	 * Comes from rCOSParser.java (slightly modified)
	 */
//	public static final String[] tokenNames = new String[] {
//	         "process", "{", "}", "component", "composition", ":", "|", "[", "]", "(", ".", "<-", ")", "interface", "extends", "public", ";", "protocol", "^", "provided", "contract", "as", "implemented", "by", "required", "of", "class", "implements", ",", "invariant", "locations", "init", "action", "->", "&", "if", "then", "else", "while", "for", "pre", "post", "return", "VAR", "Var", "var", "END", "End", "end", ":=", "instanceof", "new", "\\", "!", "\\u00ac", "in", "Set<", "int", "double", "boolean", "string", "protected", "private"
//	    };
//	
	private static final String[] tokenNames = { "this" };
	final Classifier cl;
	final Map<String, String> m;
	final List<String> list;
	final PrimitiveTypes pt;

	public FreshNameGenerator(PrimitiveTypes pt, Classifier c){
		cl = c;
		m = new HashMap<String, String>();
		m.put("int", "x");
		m.put("String", "str");
		m.put("double", "d");
		//FIXME
		m.put("boolean", "b");
		
		list = new ArrayList<String>();
		this.pt = pt;
	}

	@Deprecated
	public FreshNameGenerator(Classifier c) {
		this(PrimitiveTypes.getPrimitiveTypes(c),c);
	}
	
	/**
	 * Create a new FNG seeded with variables in design 'd'
	 * @param c Classifier
	 * @param d DESIGN to extract identifiers from.
	 */
	public FreshNameGenerator(PrimitiveTypes pt, Classifier c, DESIGN d) {
		this(pt, c);
		list.addAll(new DesignGetVarNames().run(d, null));
	}

	public String get(Type t){
		
		String para = m.get(t.getLabel());
		
		if (para == null) { 
			// Not a primitive type, so we just get the lowercase of the class name
			para = t.getLabel().toLowerCase();
		}
		
		// To avoid to have twice the same parameter, or to have a conflict with the name of a field
		// or with a keyword, we may have to append "_i"
		int i = 0;
		String para_copy = para;
		
		List<String> list_fields = new ArrayList<String>();
		
		if (cl != null)
			for (Property p : cl.getAllAttributes()){
				list_fields.add(p.getLabel());
			}
		
		List<String> list_keywords = new ArrayList<String>();
		for (int j=0; j < rCOSParser.tokenNames.length; j++) list_keywords.add(rCOSParser.tokenNames[j]); 

		while (list.contains(para) || list_fields.contains(para) || list_keywords.contains(para)) {
			para = para_copy + "_" + i++;
		}
		list.add(para);
		return para;
	}
	
	public String getBoolean() {
		return get(pt.getBoolean());
	}

	public static boolean isReservedIdentifier(String name) {
		// TODO: quite slow, I'd guess.
		for (String s : tokenNames) {
			if (s.equals(name))
				return true;
		}
		for (String s : rCOSParser.tokenNames) {
			if (s.equals(name))
				return true;
		}
		return false;
	}
}
