package edu.unu.iist.rcos.spec.antlr;

import java.util.*;

import org.antlr.runtime.tree.CommonTree;

import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;

public class MyDesign {

	public static VARNAME createVarname(String name, CommonTree token){
		assert name != null;
		if (FreshNameGenerator.isReservedIdentifier(name))
			throw new IllegalArgumentException("Reserved identifier: "+name);
		VARNAME v = SyntaxTree_ClassModelFactory.eINSTANCE.createVARNAME(name);
		v.setToken(token);
		return v;
	}
	
	public static VARNAME createVarname(String name){ 
		return createVarname(name, null);
	}

	public static VARNAME createVarname(VARNAME v){ 
		return createVarname(v.getName(), null);
	}

	public static CNAME createCname(String name, CommonTree token){
		assert name != null;
		if (FreshNameGenerator.isReservedIdentifier(name))
			throw new IllegalArgumentException("Reserved identifier: "+name);
		CNAME v = SyntaxTree_ClassModelFactory.eINSTANCE.createCNAME();
		v.setName(name);
		v.setToken(token);
		return v;
	}
	
	public static CNAME createCname(String name){
		return createCname(name, null);
	}	
	
	public static THIS createThis(CommonTree token){
		THIS v = SyntaxTree_ClassModelFactory.eINSTANCE.createTHIS();
		v.setToken(token);
		return v;		
	}

	public static THIS createThis(){
		return createThis(null);
	}	

	public static SUPER createSuper(CommonTree token) {
		SUPER s = SyntaxTree_ClassModelFactory.eINSTANCE.createSUPER();
		s.setToken(token);
		return s;
	}

	public static SUPER createSuper() {
		return createSuper(null);
	}

	public static NULL createNull(CommonTree token){
		return SyntaxTree_ClassModelFactory.eINSTANCE.createNULL();
	}

	public static NULL createNull(){
		return createNull(null);
	}

	/*     Types */
	public static INT createInt(){
		INT t = SyntaxTree_ClassModelFactory.eINSTANCE.createINT();
		t.setToken(null);
		return t;
	}
	
	public static NAT createNat(){
		NAT t = SyntaxTree_ClassModelFactory.eINSTANCE.createNAT();
		t.setToken(null);
		return t;
	}
	
	public static DOUBLE createDouble(){
		DOUBLE t = SyntaxTree_ClassModelFactory.eINSTANCE.createDOUBLE();
		t.setToken(null);
		return t;
	}
	
	public static CHAR createChar(){
		CHAR t = SyntaxTree_ClassModelFactory.eINSTANCE.createCHAR();
		t.setToken(null);
		return t;
	}
	
	public static LONG createLong(){
		LONG t = SyntaxTree_ClassModelFactory.eINSTANCE.createLONG();
		t.setToken(null);
		return t;
	}
	
	public static BOOLEAN createBoolean(){
		BOOLEAN t = SyntaxTree_ClassModelFactory.eINSTANCE.createBOOLEAN();
		t.setToken(null);
		return t;
	}
	
	public static STRING createString(){
		// TODO: remove the set and just let the string
		STRING t = SyntaxTree_ClassModelFactory.eINSTANCE.createSTRING();
		t.setToken(null);
		return t;
	}

	public static GTYPE createGenericType(String name, List<TYPE> par, CommonTree token) {
		GTYPE t = SyntaxTree_ClassModelFactory.eINSTANCE.createGTYPE();
		t.setName(createCname(name));
		t.setToken(token);
		if (par != null)
			for (TYPE t1 : par) {
				t.getPar().add(t1);
			}
		return t;
	}

	public static BASEENUM createEnumLit(String es, CommonTree token){
		BASEENUM e = SyntaxTree_ClassModelFactory.eINSTANCE.createBASEENUM();
		e.setVal(es);
		e.setToken(token);
		return e;
	}
	public static BASEBOOL createBool(boolean b, CommonTree token){
		BASEBOOL t = SyntaxTree_ClassModelFactory.eINSTANCE.createBASEBOOL();
		t.setVal(Boolean.toString(b));
		t.setBVal(b);
		t.setToken(token);
		return t;
	}
	public static BASEBOOL createTrue(){
		return createBool(true,null);
	}

	public static BASEBOOL createFalse(){
		return createBool(false,null);
	}

	public static BASEINT createBaseInt(String s, CommonTree token){
		BASEINT i = SyntaxTree_ClassModelFactory.eINSTANCE.createBASEINT();
		i.setVal(s);
		i.setToken(token);
		return i;
	}
	public static BASEINT createBaseInt(String s){
		return createBaseInt(s, null);
	}
	
	public static BASESTRING createBaseString(String s, CommonTree token){
		BASESTRING i = SyntaxTree_ClassModelFactory.eINSTANCE.createBASESTRING();
		i.setVal(s);
		i.setToken(token);
		return i;
	}
	public static BASESTRING createBaseString(String s){
		return createBaseString(s, null);
	}

	public static BASEDOUBLE createBaseDouble(String s, CommonTree token){
		BASEDOUBLE i = SyntaxTree_ClassModelFactory.eINSTANCE.createBASEDOUBLE();
		i.setVal(s);
		i.setToken(null);
		return i;
	}
	public static BASEDOUBLE createBaseDouble(String s){
		return createBaseDouble(s, null);
	}
	public static BASEENUM createBaseEnum(String s) {
		return createEnumLit(s, null);
	}

	public static CAST createCast(TYPE t, VAL v, CommonTree token){
		CAST c = SyntaxTree_ClassModelFactory.eINSTANCE.createCAST();
		c.setType(t);
		c.setVal(v);
		c.setToken(token);
		return c;
	}
	public static CAST createCast(TYPE t, VAL v){
		return createCast(t, v, null);
	}

	public static FA createFA(PREFIX path, String name, boolean dash, CommonTree token){
		return createFA(path, createVarname(name),dash,token);
	}
	/**
	 * WARNING: This method WILL change the container of 'name'! Use
	 * {@link #createFA(PREFIX, String, boolean, CommonTree)} if you want to create a COPY
	 * of the name!
	 * TODO: Review--maybe it might be better to always create a copy instead?
	 */
	public static FA createFA(PREFIX path, VARNAME name, boolean dash, CommonTree token){
		FA v = SyntaxTree_ClassModelFactory.eINSTANCE.createFA(name,path);
		v.setDash(false);
		v.setToken(token);
		v.setDash(dash);
		return v;
	}
	public static FA createFA(PREFIX path, String name, CommonTree token){
		return createFA(path, name, false, token);
	}
	
	public static MCALL createMcall(PREFIX path, String name, List<EXP>in, List<REF>out, CommonTree token){
		return createMcall(path, createVarname(name), in, out);
	}
	public static MCALL createMcall(PREFIX path, VARNAME name, List<EXP>in, List<REF>out, CommonTree token){
		MCALL v = SyntaxTree_ClassModelFactory.eINSTANCE.createMCALL();
		v.setOwner(path);
		v.setName(name);
		if (in != null){
			for (EXP e : in)
				v.getIn().add(e);
		}
		if (out != null){
			for (REF v1 : out)
				v.getOut().add(v1);
		}
		v.setToken(token);
		return v;
	}
	
	public static MCALL createMcall(PREFIX path, String name, List<EXP>in, List<REF>out){
		return createMcall(path, createVarname(name), in, out, null);
	}
	public static MCALL createMcall(PREFIX path, VARNAME name, List<EXP>in, List<REF>out){
		return createMcall(path, name, in, out, null);
	}

	public static NEW createNew(GTYPE i, List<EXP>in, List<REF>out, CommonTree token){
		NEW v = SyntaxTree_ClassModelFactory.eINSTANCE.createNEW();
		v.setName(i);
		if (in != null) 
			for (EXP e : in)
				v.getIn().add(e);
		if (out != null)
			for (REF v1 : out) {
				// a constructor with return values? XXX
				assert(false);
				v.getOut().add(v1);
			}
		v.setToken(token);
		return v;
	}
	public static NEW createNew(GTYPE i, List<EXP>in, List<REF>out){
		return createNew(i, in, out, null);
	}

	public static VAR createVar(String n, TYPE t, boolean dash, CommonTree token){
		return createVar(createVarname(n),t,dash,token);
	}
	public static VAR createVar(VARNAME n, TYPE t, boolean dash, CommonTree token){
		VAR v = SyntaxTree_ClassModelFactory.eINSTANCE.createVAR();
		v.setName(n);
		v.setType(t);
		v.setDash(dash);
		v.setToken(token);
		return v;
	}	
	public static VAR createVar(String n, TYPE t){
		return createVar(n, t, false, null);
	}
	public static VAR createVar(String n, TYPE t, CommonTree token){
		return createVar(n, t, false, token);
	}
	public static VAR createVar(String n, TYPE t, boolean dash){
		return createVar(createVarname(n),t,dash);
	}
	public static VAR createVar(VARNAME n, TYPE t, boolean dash){
		return createVar(n, t, dash, null);
	}

	public static PAR createPar(String n, TYPE t, boolean dash, CommonTree token){
		return createPar(createVarname(n),t,dash,token);
	}
	public static PAR createPar(VARNAME n, TYPE t, boolean dash, CommonTree token){
		PAR v = SyntaxTree_ClassModelFactory.eINSTANCE.createPAR();
		v.setName(n);
		v.setType(t);
		v.setDash(dash);
		v.setToken(token);
		return v;
	}
	public static PAR createPar(String n, TYPE t){
		return createPar(n, t, false, null);
	}
	public static PAR createPar(String n, TYPE t, CommonTree token){
		return createPar(n, t, false, token);
	}
	public static PAR createPar(VARNAME n, TYPE t, boolean dash){
		return createPar(n, t, dash, null);
	}
	
	// Designs
	public static SEQ createSeq(DESIGN d1, DESIGN d2, CommonTree token){
		assert d1 != null;
		assert d2 != null;
		SEQ s = SyntaxTree_ClassModelFactory.eINSTANCE.createSEQ();
		s.setLeft(d1);
		s.setRight(d2);
		s.setToken(token);
		return s;
	}
	public static SEQ createSeq(DESIGN d1, DESIGN d2){
		return createSeq(d1, d2, null);
	}
	
	public static CHOICE createChoice(DESIGN d1, DESIGN d2, CommonTree token){
		CHOICE s = SyntaxTree_ClassModelFactory.eINSTANCE.createCHOICE();
		s.setLeft(d1);
		s.setRight(d2);
		s.setToken(token);
		return s;
	}
	public static CHOICE createChoice(DESIGN d1, DESIGN d2){
		return createChoice(d1, d2, null);
	}

	public static IF createIf(EXP e, DESIGN d1, DESIGN d2, CommonTree token){
		assert e != null;
		assert d1 != null;
		assert d2 != null;
		IF s = SyntaxTree_ClassModelFactory.eINSTANCE.createIF();
		s.setB(e);
		s.setLeft(d1);
		s.setRight(d2);
		s.setToken(token);
		return s;
	}
	public static IF createIf(EXP e, DESIGN d1, DESIGN d2){
		return createIf(e, d1, d2, null);
	}

	public static WHILE createWhile(EXP e, DESIGN d, CommonTree token){
		WHILE s = SyntaxTree_ClassModelFactory.eINSTANCE.createWHILE();
		s.setB(e);
		s.setDo(d);
		s.setToken(token);
		return s;
	}
	public static WHILE createWhile(EXP e, DESIGN d){
		return createWhile(e, d, null);
	}

	public static FOR createFor(TYPE t, String v, VAL va,  DESIGN d, CommonTree token){
		return createFor(t, createVarname(v),va,d,token);
	}
	public static FOR createFor(TYPE t, VARNAME v, VAL va,  DESIGN d, CommonTree token){
		FOR s = SyntaxTree_ClassModelFactory.eINSTANCE.createFOR();
		s.setType(t);
		s.setVar(v);
		s.setVal(va);
		s.setDo(d);
		s.setToken(token);
		return s;
	}
	public static FOR createFor(TYPE t, String v, VAL va,  DESIGN d){
		return createFor(t, createVarname(v), va, d, null);
	}
	public static FOR createFor(TYPE t, VARNAME v, VAL va,  DESIGN d){
		return createFor(t, v, va, d, null);
	}

	public static PP createPP(EXP p1, EXP p2, CommonTree token){
		PP s = SyntaxTree_ClassModelFactory.eINSTANCE.createPP();
		s.setPre(p1);
		s.setPost(p2);
		s.setToken(token);
		return s;
	}
	public static PP createPP(EXP p1, EXP p2){
		return createPP(p1, p2, null);
	}

	// Atomic commands
	public static SKIP createSkip(CommonTree token){
		SKIP s = SyntaxTree_ClassModelFactory.eINSTANCE.createSKIP();
		s.setToken(token);
		return s;
	}
	public static SKIP createSkip(){
		return createSkip(null);
	}
	
	public static CHAOS createChaos(CommonTree token){
		CHAOS c = SyntaxTree_ClassModelFactory.eINSTANCE.createCHAOS();
		c.setToken(token);
		return c;
	}
	public static CHAOS createChaos(){
		return createChaos(null);
	}
	
	public static MAGIC createMagic(){
		return SyntaxTree_ClassModelFactory.eINSTANCE.createMAGIC();
	}

	public static STOP createStop(){
		return SyntaxTree_ClassModelFactory.eINSTANCE.createSTOP();
	}

	public static VARDECL createVarDecl(TYPE t, String v, CommonTree token){
		return createVarDecl(t,createVarname(v),token);
	}
	public static VARDECL createVarDecl(TYPE t, VARNAME v, CommonTree token){
		VARDECL vd = SyntaxTree_ClassModelFactory.eINSTANCE.createVARDECL();
		vd.setType(t);
		vd.setVar(v);
		vd.setToken(token);
		return vd;
	}
	public static VARDECL createVarDecl(TYPE t, String v){
		return createVarDecl(t, createVarname(v));
	}
	public static ENDDECL createEndDecl(String v){
		return createEndDecl(createVarname(v));
	}
	public static VARDECL createVarDecl(TYPE t, VARNAME v){
		return createVarDecl(t, v, null);
	}

	public static ENDDECL createEndDecl(String v, CommonTree token){
		return createEndDecl(createVarname(v), token);
	}
	public static ENDDECL createEndDecl(VARNAME v, CommonTree token){
		ENDDECL vd = SyntaxTree_ClassModelFactory.eINSTANCE.createENDDECL();
		vd.setVar(v);
		vd.setToken(token);
		return vd;
	}
	public static ENDDECL createEndDecl(VARNAME v){
		return createEndDecl(v, null);
	}

	public static ASSIGN createAssign(REF v, EXP e, CommonTree token){
		ASSIGN a = SyntaxTree_ClassModelFactory.eINSTANCE.createASSIGN();
		a.setVal(v);
		a.setE(e);
		a.setToken(token);
		return a;		
	}
	public static ASSIGN createAssign(REF v, EXP e){
		return createAssign(v, e, null);
	}

	public static NEG createNeg(EXP p1, CommonTree token){
		NEG o = SyntaxTree_ClassModelFactory.eINSTANCE.createNEG();
		o.setE(p1);
		o.setToken(token);
		return o;
	}
	public static NEG createNeg(EXP p1){
		return createNeg(p1, null);
	}
	
	public static FORALL createForall(EXP p1, TYPE t, VARNAME v, CommonTree token){
		FORALL p = SyntaxTree_ClassModelFactory.eINSTANCE.createFORALL();
		p.setE(p1);
		p.setType(t);
		p.setVar(v);
		p.setToken(token);
		return p;
	}
	public static FORALL createForall(EXP p1, TYPE t, VARNAME v){
		return createForall(p1, t, v, null);
	}

	public static ITERATOR createIterator(EXP body, TYPE type, VARNAME varname, EXP range, CommonTree token) {
		ITERATOR i = SyntaxTree_ClassModelFactory.eINSTANCE.createITERATOR();
		i.setE(body);
		i.setType(type);
		i.setVar(varname);
		i.setRange(range);
		i.setToken(token);
		return i;
	}
	public static ITERATOR createIterator(EXP body, TYPE type, VARNAME varname, EXP range) {
		return createIterator(body, type, varname, range, null);
	}

	public static EXIST createExist(EXP p1, TYPE t, VARNAME v, CommonTree token){
		EXIST p = SyntaxTree_ClassModelFactory.eINSTANCE.createEXIST();
		p.setE(p1);
		p.setType(t);
		p.setVar(v);
		p.setToken(token);
		return p;
	}
	public static EXIST createExist(EXP p1, TYPE t, VARNAME v){
		return createExist(p1, t, v, null);
	}
	
	public static INSTANCEOF createInstanceOf(REF e, CNAME c, CommonTree token){
		INSTANCEOF iof = SyntaxTree_ClassModelFactory.eINSTANCE.createINSTANCEOF();
		iof.setClass_(c);
		iof.setRef(e);
		iof.setToken(token);
		return iof;
	}
	public static INSTANCEOF createInstanceOf(REF e, CNAME c){
		return createInstanceOf(e, c, null);
	}

	static <T extends BINEXP> T createBINEXP(T be, EXP l, EXP r, CommonTree token) {
		be.setLeft(l);
		be.setRight(r);
		be.setToken(token);
		return be;
	}

	public static RELOP createRelOp(RelOps t, EXP e1, EXP e2, CommonTree token){
		RELOP op = createBINEXP(SyntaxTree_ClassModelFactory.eINSTANCE.createRELOP(),e1,e2,token);
		op.setOp(t);
		return op;	
	}
	public static RELOP createRelOp(RelOps t, EXP e1, EXP e2){
		return createRelOp(t, e1, e2, null);
	}

	public static LOGOP createLogOp(LogOps t, EXP e1, EXP e2, CommonTree token){
		LOGOP op = createBINEXP(SyntaxTree_ClassModelFactory.eINSTANCE.createLOGOP(),e1,e2,token);
		op.setOp(t);
		return op;	
	}
	public static LOGOP createLogOp(LogOps t, EXP e1, EXP e2){
		return createLogOp(t, e1, e2, null);
	}
	
	public static AOP createAOp(AOps t, EXP e1, EXP e2, CommonTree token){
		AOP op = createBINEXP(SyntaxTree_ClassModelFactory.eINSTANCE.createAOP(),e1,e2,token);
		op.setOp(t);
		return op;	
	}
	public static AOP createAOp(AOps t, EXP e1, EXP e2){
		return createAOp(t, e1, e2, null);
	}
	
	public static RET createRet(EXP e){
		RET r = SyntaxTree_ClassModelFactory.eINSTANCE.createRET();
		r.setE(e);
		return r;
	}

	public static boolean isTrue(EXP e) {
		return (e instanceof BASEBOOL && ((BASEBOOL)e).isBVal());
	}

	public static boolean isFalse(EXP e) {
		return (e instanceof BASEBOOL && !((BASEBOOL)e).isBVal());
	}
}
