package edu.unu.iist.rcos.spec.antlr;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.LinkedList;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.Callable;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.ANTLRFileStream;
import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;
import org.antlr.runtime.tree.CommonTreeNodeStream;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;

import org.eclipse.uml2.uml.*;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Package;

import edu.unu.iist.rcos.mm.ocl.Class_OCL;
import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.mm.util.Utility.MyPort;
import edu.unu.iist.rcos.profile.*;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.parser.rCOSTreeGetClasses;
import edu.unu.iist.rcos.spec.antlr.parser.rCOSTreeToUML;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;
import edu.unu.iist.rcos.spec.tc.*;
import edu.unu.iist.rcos.spec.tc.RuleTCMembers.*;
import edu.unu.iist.rcos.uml.EmptyModel;
import edu.unu.iist.rcos.uml.NotAnRCOSModelException;

public class MyRCOSModel extends EmptyModel {

	/**
	 * @author vs
	 */
	public boolean incompleteClass = false; // for debugging the example models
											// like VQ it's nice to relax the
											// types a bit.

	protected TCAble tc = new TCAble() {
		public boolean isEnabled() {
			return false;
		}
	};
	protected LinkedList<Callable<Void>> tcQ = new LinkedList<Callable<Void>>(); // throws
																					// RecognitionException	
	protected Set<String> topLevelComponentsRegister;
	
	final protected RuleTCMembers tcrm;
	final protected RuleOperationGuard tcg;
	final protected RuleOperationDesign tcd;
	final protected RuleClassInvariant tci;
	public final TypeHandler th; // has to be here because of packaging/circular
									// imports

	/**
	 * Create a new empty UML model
	 * 
	 * @param file
	 *            : URI
	 * @param model_name
	 *            : Name of top-level element
	 * @throws NotAnRCOSModelException
	 * @throws IOException
	 */
	public MyRCOSModel(URI file, String model_name)
			throws NotAnRCOSModelException, IOException {
		super(file, model_name);
		// Get the package containing all the rCOS primitive types

		// setup the rule framework:
		th = new TypeHandler(rcos, incompleteClass);
		tcrm = new RuleTCMembersTH(th);
		tcg = tcrm.new RuleOperationGuard();
		tcd = tcrm.new RuleOperationDesign();
		tci = tcrm.new RuleClassInvariant();
		topLevelComponentsRegister = new TreeSet<String>();
	}

	// Create wrappers for ANTLR so that we do not need to bother the user.
	public static class RCOSFromFile extends ANTLRFileStream {
		public RCOSFromFile(final String s) throws IOException {
			super(s, "utf-8");
		}
	}

	public static class RCOSFromString extends ANTLRStringStream {
		public RCOSFromString(final String s) {
			super(s);
		}
	}

	public MyRCOSModel(String xmi_file, String model_name, RCOSFromString buf)
			throws RecognitionException, NotAnRCOSModelException, IOException {
		this(xmi_file, model_name, buf, false);
	}

	public MyRCOSModel(String xmi_file, String model_name, RCOSFromString buf,
			boolean incomplete) throws RecognitionException,
			NotAnRCOSModelException, IOException {
		this(URI.createFileURI(xmi_file), model_name);
		this.incompleteClass = incomplete;
		helpMe(buf);
	}

	public MyRCOSModel(String xmi_file, String model_name, RCOSFromFile buf)
			throws IOException, RecognitionException, NotAnRCOSModelException {
		this(URI.createFileURI(xmi_file), model_name);
		helpMe(buf);
	}

	/**
	 * @param co
	 *            is the component containing the interface, null if it is a
	 *            top-level interface
	 * @param name
	 *            is the name of the interface
	 * @token contains all the information concerning the position of the token
	 * @return the interface "name"
	 * @throws TypingException
	 */
	public Component getComponent(String name, Component co, CommonTree token)
			throws RCOSModelCheckException {
		Component c = rcos.getComponent(name, co);
		if (c == null && token == null) {
			throw new RCOSModelCheckException("Component '" + name + "' not found");
		}
		else if (c == null) {
			throw new RCOSModelCheckException("Component '" + name
					+ "' not found", token);
		}
		else
			return c;
	}

	/**
	 * @param co
	 *            is the component containing the interface, null if it is a
	 *            top-level interface
	 * @param name
	 *            is the name of the interface
	 * @token contains all the information concerning the position of the token
	 * @return the interface "name"
	 * @throws TypingException
	 */
	public Interface getInterface(String name, Component co, CommonTree token)
			throws RCOSModelCheckException {
		Interface i = rcos.getInterface(name, co);
		if (i == null)
			throw new RCOSModelCheckException("Interface '" + name
					+ "' not found", token);
		else
			return i;
	}

	/**
	 * @param co
	 *            is the component containing the interface, null if it is a
	 *            top-level interface
	 * @param name
	 *            is the name of the class
	 * @token contains all the information concerning the position of the token
	 * @return the interface "name"
	 * @throws RCOSModelCheckException
	 *             if 'name' not found.
	 * @see Utility#getClass(String, Component, RCOSModel)
	 */
	public Class getClass(String name, Component co, CommonTree token)
			throws RCOSModelCheckException {
		Class c = rcos.getClass(name, co);
		if (c == null) {
			final String coName = co == null ? "" : (" in or above " + co
					.getQualifiedName());
			throw new RCOSModelCheckException("Class '" + name + "' not found"
					+ coName, token);
		} else
			return c;
	}

	/**
	 * @param co
	 *            is the component containing the interface, null if it is a
	 *            top-level interface
	 * @param name
	 *            is the name of the interface
	 * @token contains all the information concerning the position of the token
	 * @return the interface "name"
	 * @throws RCOSModelCheckException
	 * @see {@link Utility#getClassifierParam(String, Classifier)}
	 */
	public Class getClassifierParam(String name, Classifier cl, CommonTree token)
			throws RCOSModelCheckException {
		try {
			Class ret = Utility.getClassifierParam(name, cl);
			if (ret == null)
				throw new RCOSModelCheckException("Class '" + name
						+ "' not found in " + cl.getQualifiedName(), token);
			else
				return ret;

		} catch (WellformedException e) {
			throw new RCOSModelCheckException(e, token);
		}
	}

	/**
	 * @param co
	 *            is the component containing the contract, null if it is a
	 *            top-level contract
	 * @param name
	 *            is the name of the contract
	 * @token contains all the information concerning the position of the token
	 * @return the contract "name"
	 * @throws TypingException
	 */
	public ContractInterface getContract(String name, Component co,
			CommonTree token) throws RCOSModelCheckException {
		// co may be null
		ContractInterface ci = rcos.getContract(name, co);
		if (ci == null)
			throw new RCOSModelCheckException("Contract '" + name
					+ "' not found in component " + co.getQualifiedName(),
					token);
		else
			return ci;
	}

	/**
	 * Find contract 'name' in the inheritance chain of component 'co'.
	 */
	public ContractInterface getContractIn(String name, Component co,
			CommonTree token) throws RCOSModelCheckException {
		ContractInterface ci = rcos.getContractInherit(name, co);
		if (ci == null)
			throw new RCOSModelCheckException("Contract '" + name
					+ "' not found in component " + co.getQualifiedName(),
					token);
		else
			return ci;
	}

	public static void addDelegatedRequiredInterface(
			ContractInterface contract, ContractInterface contract2,
			Component co, CommonTree t) throws RCOSModelCheckException {
		final Interface i = contract.getBase_Interface();
		final Interface i2 = contract2.getBase_Interface();
		try {
			assert (co.getOwner() != null);
			Port p = MyPort.getUsagePort(co);
			if (p == null) {
				p = MyPort.createPort(co);
			}

			Dependency d = UMLFactory.eINSTANCE.createDependency();
			co.getPackagedElements().add(d);
			Delegation dl = Stereotypes.applyDelegation(d);
			d.setName("DI_" + i.getLabel() + "_" + co.getLabel() + "."
					+ i2.getLabel());
			dl.configure(p, contract2, contract);
		} catch (WellformedException e) {
			throw new RCOSModelCheckException(e, t);
		}
	}

	public static void addRequiredInterface(ContractInterface ctr,
			Component co, CommonTree t) throws RCOSModelCheckException {
		final Interface i = ctr.getBase_Interface();
		try {
			Usage ir = co.createUsage(i);
			co.getPackagedElements().add(ir); // TODO: read the docs of
												// createUsage: first it get's
												// created outside of 'co' and
												// is then moved inside
			ir.setName(i.getLabel() + "_Required");
			// If the component is not atomic, then we first look if a port
			// already requiring i exist
			// (that is created by [(C.I <- i)]
			if (!Utility.checkAtomicComponent(co)) {
				Port p = Utility.MyPort.getUsagePort(co);
				if (p == null) {
					p = MyPort.createPort(co);
				}
				ir.getClients().add(p);
			}
		} catch (WellformedException e) {
			throw new RCOSModelCheckException(e, t);
		}
	}

	//For rCOS type...
	private Parameter addPara(String name, TYPE type, Operation op,
			ParameterDirectionKind dir) throws RCOSModelCheckException {
	
		if (name.equals(Class_OCL.RESERVED_IDENTIFIER_THIS))
			throw new RCOSModelCheckException(
					"Can't declare a parameter with reserved name 'this'", type);
		
		try {
			Type t = getTypeFromTYPE(type, (Classifier) op.getOwner());
			if (t == null) {
				throw new RCOSModelCheckException("Type '"
						+ DesignUtility.typeToString(type) + "' not found",
						type);
			
			}
			else {
				return addPara(name, t, op, dir);
			}
		} catch (RCOSModelCheckException ex) {
			throw new RCOSModelCheckException("While defining operation "
					+ op.getQualifiedName() + ":\n" + ex.getMessage(), ex);
		}
	}
	
	//For UML type...
	public Parameter addPara(String name, Type type, Operation op,
			ParameterDirectionKind dir) throws RCOSModelCheckException {

		// We first need to know the class or the interface owning the operation
		Element e = op.getOwner();
		assert (e != null);

		Parameter par = op.createOwnedParameter(name, type);
		if (Utility.isSetType(type)) {
			par.setLower(0);
			par.setUpper(-1);
			par.setType(Utility.getBaseType(type));
		}
		par.setDirection(dir);
		return par;
	}

	public Type getTypeFromTYPE(TYPE type, Classifier owner)
			throws RCOSModelCheckException {
		try {
			return th.getTypeFromTYPE(type, owner);
		} catch (GetTypeException e) {
			throw new RCOSModelCheckException(e, type);
		}
	}

	public void addInPara(String name, TYPE type, Operation op)
			throws RCOSModelCheckException {
		addPara(name, type, op, ParameterDirectionKind.IN_LITERAL);
	}

	public void addOutPara(String name, TYPE type, Operation op)
			throws RCOSModelCheckException {
		addPara(name, type, op, ParameterDirectionKind.OUT_LITERAL);
	}

	private Property addAttribute(TYPE type, final String name, Type t,
			VisibilityKind visibility, Classifier c, EXP exp)
			throws RCOSModelCheckException {
		return addAttribute(type, name, t, visibility, 1, 1, c, exp);
	}

	private Property addAttribute(TYPE type, final String name, Type t,
			VisibilityKind visibility, int lower, int upper, Classifier c,
			EXP exp) throws RCOSModelCheckException {
		assert (name != null);
		assert (c != null);

		Property p;
		// We have to know whether the Classifier is an Interface or a Class,
		// because the
		// implementers of UML didn't put the method createOwnedAttribute in
		// Classifier ...
		if (c instanceof Class)
			p = ((Class) c).createOwnedAttribute(name, t);
		else if (c instanceof Component)
			p = ((Component) c).createOwnedAttribute(name, t);
		else if (c instanceof Interface)
			p = ((Interface) c).createOwnedAttribute(name, t);
		else
			throw new RCOSModelCheckException(
					"Trying to add a field to a classifier which is neither a Class nor an Interface",
					type);
		p.setVisibility(visibility);

		p.setLower(lower);
		p.setUpper(upper);

		if (exp != null) {
			final String expS = DesignUtility.expToString(c, exp);
			if (tc.isEnabled()) { // TC
				Classifier tcCtx = c; // use classifier instead of 'context'
										// from above
				try {
					if (!((!DesignUtility.hasPrimedVariables(exp)) && (th
							.getType(exp, tcCtx).equals(t) || (th.isBooleanExp(
							exp, tcCtx) && DesignUtility.getListId(exp)
							.contains("this." + name))))) {
						throw new RCOSModelCheckException("The expression '"
								+ expS + "' does not initialize attribute "
								+ name + " of type " + t.getQualifiedName(),
								type);
					}
				} catch (GetTypeException e) {
					throw new RCOSModelCheckException(e, type);
				}
			}
			p.setDefault(expS);
		}
		return p;
	}

	public void addField(String name, TYPE type, VisibilityKind v,
			Classifier c, EXP e) throws RCOSModelCheckException {

		//for (Property p : c.getAllAttributes()) {
		//	if (p.getName().equals(name))
		//	throw new RCOSModelCheckException(
		//			"Attribute with name " + name + " already exist in classifer "+ c.getName(), type);
		//}

		//for (Association a : c.getAssociations()) {
		//	for (Property p : a.getNavigableOwnedEnds()) {
		//		if (p.getName().equals(name))
		//			throw new RCOSModelCheckException(
		//			"Attribute with name " + name + " already exist in classifer "+ c.getName(), type);
		//	}
		//}
	
		if (name.equals(Class_OCL.RESERVED_IDENTIFIER_THIS))
			throw new RCOSModelCheckException(
					"Can't declare a field with reserved name 'this'", type);
		//check if type exist
		Type t = getTypeFromTYPE(type, c);
		if (t == null)
			throw new RCOSModelCheckException("Type "
					+ DesignUtility.typeToString(type) + " not found", type);
		assert (t != null);
		addField(name, t, v, c, e, type);
	}

	public void addField(String name, Type t, VisibilityKind v,
			Classifier c, EXP e, TYPE type) throws RCOSModelCheckException {

		if (PrimitiveTypes.isRCOSPrimitive(t))
			addAttribute(type, name, t, v, c, e);
		else if (Utility.isSetType(t)) {
			Type base = Utility.getBaseType(t);
			addAssociation(name, base, v, 0, -1, c, e);
		} else
			addAssociation(name, t, v, 1, 1, c, e);
	}

	private void addAssociation(String name, Type type, VisibilityKind v,
			int lower, int upper, Classifier c, EXP e)
			throws RCOSModelCheckException {
		// Type t = Utility.getTypeFromString(t, context, rm);
		AggregationKind ak = AggregationKind.NONE_LITERAL;
		if (lower == 0 && upper == -1) {
			ak = AggregationKind.COMPOSITE_LITERAL;
		}
		Association a = c.createAssociation(true, ak, name, lower, upper, type,
				false, AggregationKind.NONE_LITERAL,
				c.getLabel().toLowerCase(), 1, 1);
		a.setName("A_" + c.getLabel() + "_" + name);
		Property p = c.getAttribute(name, type);
		if (e != null) {
			// TODO: TC
			p.setDefault(DesignUtility.expToString(c, e));
		}
		a.getNavigableOwnedEnds().add(p);
	}

	public void addDesign(final DESIGN d, final Operation o)
			throws RCOSModelCheckException {
		assert o != null;
		// Don't typecheck on Interfaces (yet)? TODO. Maybe it's fixed?
		// There's another instance of this in addGuard() below.
		add(d, true, tcd, o);
		// TODO && (o.getClass_() != null)) {
	}

	public void addGuard(EXP e, Operation o) throws RCOSModelCheckException {
		// TODO: review, see above && (o.getClass_() != null)) {
		add(e, false, tcg, o);
	}

	private <T extends DESIGNEXP, NS extends Namespace> void add(final T t,
			boolean defer, final RuleTC<NS, ?, ?, T> rif, final NS ns)
			throws RCOSModelCheckException {
		if (tc.isEnabled()) {
			if (defer) {
				tcQ.add(new Callable<Void>() {
					public Void call() throws RCOSModelCheckException {
						try {
							rif.check(ns, t);
							return null;
						} catch (WellformedException e) {
							throw new RCOSModelCheckException(e, t);
						}
					}
				});
			} else
				try {
					rif.check(ns, t);
				} catch (WellformedException e) {
					throw new RCOSModelCheckException(e, t);
				}
		}
		try {
			// TODO: review
			// We don't add the DESIGN, otherwise UML cries ...
			// rs.getContents().add(d);
			// dop.setAst(d);
			rif.rf.create(ns, rif.toString(ns, t));
		} catch (WellformedException e1) {
			throw new RCOSModelCheckException(e1, t);
		}
	}

	public void addInvariant(final EXP e, final Class co)
			throws RCOSModelCheckException {
		add(e, true, tci, co);
	}

	public void addInvariant(EXP e, ContractInterface ctr)
			throws RCOSModelCheckException {
		Interface i = ctr.getBase_Interface();
		add(e, false, tci, i);
	}

	// Parses the file with tree walkers
	protected void helpMe(final ANTLRStringStream s)
			throws RecognitionException {
		CommonTree ret = Parse.getCommonTree(s);

		CommonTreeNodeStream nodes = new CommonTreeNodeStream(ret);
		rCOSTreeGetClasses walker = new rCOSTreeGetClasses(nodes);

		walker.m = this;
		walker.rCOSSpec();
		// We have to parse the stream again
		// System.err.println(ret.toStringTree());
		nodes = new CommonTreeNodeStream(ret);
		rCOSTreeToUML walker2 = new rCOSTreeToUML(nodes);
		walker2.m = this;
		walker2.rCOSSpec();
	}

	/**
	 * @author Charles Morisset
	 * @param c
	 * @param class1
	 * @param param
	 * @throws TypingException
	 */
	public void createGeneralization(Class c, Class class1, List<TYPE> param)
			throws RCOSModelCheckException {
		c.createGeneralization(class1);
		TemplateBinding tb = UMLFactory.eINSTANCE.createTemplateBinding();
		tb.setBoundElement(c);
		tb.setSignature(class1.getOwnedTemplateSignature());
		List<Type> ltype = new ArrayList<Type>();
		for (TYPE e : param) {
			ltype.add(getTypeFromTYPE(e, c));
		}
		for (int i = 0; i < ltype.size(); i++) {
			TemplateParameterSubstitution ps = tb.createParameterSubstitution();
			ps.setFormal((TemplateParameter) class1.getOwnedTemplateSignature()
					.getOwnedParameters().get(i));
			Utility.setActuals(ps, ltype.get(i));
		}
	}

	/**
	 * @author Charles Morisset
	 * @param name
	 *            the name of the interface
	 * @return the interface added to the component model
	 * @throws WellformedException
	 *             if an interface with 'name' already exists.
	 */
	public Interface addInterface(String name) throws RCOSModelCheckException {
		assert name != null;
		if (name.equals(""))
			throw new RCOSModelCheckException(
					"Empty interface name not permitted!");
		final Package componentPackage = rcos.getComponentModel()
				.getBase_Package();
		if (componentPackage.getOwnedMember(name, false,
				UMLPackage.eINSTANCE.getInterface()) != null
				&& Class_OCL.MEMBERS_UNIQUE_NAMES != null)
			throw new RCOSModelCheckException("Duplicate interface '" + name
					+ "' in component package");
		return componentPackage.createOwnedInterface(name);
	}

	/**
	 * @author Charles Morisset
	 * @param i
	 *            a class
	 * @param co
	 *            a component
	 * @return the class added to the component co
	 * @throws WellformedException
	 *             if a class with 'c.label' already exists in the component
	 */
	public ContractInterface addContract(String name)
			throws RCOSModelCheckException {
		Interface i = addInterface(name);
		return Stereotypes.applyContractInterface(i);
	}

	/**
	 * @param name
	 *            The name of the component added
	 * @param co
	 *            The component owning the component, null if it is a top-level
	 *            component
	 * @param process
	 *            <b>true</b> if the component is a process, <b>false</b> if it
	 *            is a service component
	 */
	public rCOSComponent addComponent(String name, rCOSComponent co,
			boolean process) {
		Component c = (Component) rcos
				.getComponentModel()
				.getBase_Package()
				.createPackagedElement(name,
						UMLPackage.eINSTANCE.getComponent());
		if (co != null)
			co.getBase_Component().getNestedClassifiers().add(c);
		topLevelComponentsRegister.add(name);
		rCOSComponent rc;
		if (process)
			rc = Stereotypes.applyProcessComponent(c);
		else
			rc = Stereotypes.applyServiceComponent(c);
		return rc;
	}

	public void addComposition(Component c, ContractInterface ci1,
			ContractInterface ci2) {
		Dependency d = UMLFactory.eINSTANCE.createDependency();
		d.getSuppliers().add(ci2.getBase_Interface());
		d.getClients().add(ci1.getBase_Interface());
		c.getPackagedElements().add(d);
		Stereotypes.applyComposition(d);
		d.setName(ci1.getBase_Interface() + "<->"
				+ ci2.getBase_Interface().getName());
	}

	/**
	 * @author Charles Morisset
	 * @param c
	 *            the class
	 * @param rm
	 *            the RCOS Model
	 * @return the class added to the class model of rm
	 * @throws WellformedException
	 *             if a class with 'c.label' already exists.
	 */
	public Class addClass(final Class c) throws RCOSModelCheckException {
		Package classPackage = rcos.getClassModel().getBase_Package();
		if (classPackage.getOwnedMember(c.getLabel(), false,
				UMLPackage.eINSTANCE.getClass_()) != null
				&& Class_OCL.MEMBERS_UNIQUE_NAMES != null)
			throw new RCOSModelCheckException("Duplicate class '"
					+ c.getLabel() + "' in class package");
		c.setPackage(classPackage);
		tcQ.add(new Callable<Void>() {
			public Void call() throws RCOSModelCheckException {
				try {
					RCOSModelValidationNoUML.checkClassImplements(c);
				} catch (WellformedException e) {
					throw new RCOSModelCheckException(e.getMessage(), e);
				}
				return null;
			}
		});
		return c;
	}

	/**
	 * @author Charles Morisset
	 * @param c
	 *            a class
	 * @param co
	 *            a component
	 * @return the class added to the component co
	 * @throws WellformedException
	 *             if a class with 'c.label' already exists in the component
	 */
	public Class addClass(final Class c, rCOSComponent co)
			throws RCOSModelCheckException {
		String name = c.getLabel();
		if (co.getBase_Component().getOwnedMember(name, false,
				UMLPackage.eINSTANCE.getClass_()) != null
				&& Class_OCL.MEMBERS_UNIQUE_NAMES != null)
			throw new RCOSModelCheckException("Duplicate class '" + name
					+ "' in component " + co.getBase_Component().getLabel());
		co.getBase_Component().getNestedClassifiers().add(c);
		// A class in a component may only be private
		c.setVisibility(VisibilityKind.PRIVATE_LITERAL);
		tcQ.add(new Callable<Void>() {
			public Void call() throws RCOSModelCheckException {
				try {
					RCOSModelValidationNoUML.checkClassImplements(c);
				} catch (WellformedException e) {
					throw new RCOSModelCheckException(e.getMessage(), e);
				}
				return null;
			}
		});
		return c;
	}

	/**
	 * @author Charles Morisset
	 * @param i
	 *            a class
	 * @param co
	 *            a component
	 * @return the class added to the component co
	 * @throws WellformedException
	 *             if a class with 'c.label' already exists in the component
	 */
	public static Interface addInterface(String name, rCOSComponent co)
			throws RCOSModelCheckException {
		assert co != null;
		String ok;
		if ((ok = canAddInterface(name, co)) != null)
			throw new RCOSModelCheckException(ok);
		return co.getBase_Component().createOwnedInterface(name);
	}

	public static String canAddInterface(String name, rCOSComponent co) {
		assert co != null;
		if (co.getBase_Component().getOwnedMember(name, false,
				UMLPackage.eINSTANCE.getInterface()) != null
				&& Class_OCL.MEMBERS_UNIQUE_NAMES != null)
			return "Duplicate interface '" + name + "' in component "
					+ co.getBase_Component().getLabel();
		else {
			Set<String> seen = new HashSet<String>();
			for (Interface i : co.getBase_Component().getProvideds()) {
				assert seen.add(i.getName());
			}
			return null;
		}
	}

	/**
	 * @author Charles Morisset
	 * @param i
	 *            a class
	 * @param co
	 *            a component
	 * @return the class added to the component co
	 * @throws WellformedException
	 *             if a class with 'c.label' already exists in the component
	 */
	public static ContractInterface addContract(String name, rCOSComponent co)
			throws RCOSModelCheckException {
		Interface i = addInterface(name, co);
		return Stereotypes.applyContractInterface(i);
	}

	public static InterfaceRealization addProvidedInterface(
			ContractInterface ci, Component co, NamedElement c)
			throws RCOSModelCheckException {

		Interface i = ci.getBase_Interface();
		InterfaceRealization ir = UMLFactory.eINSTANCE
				.createInterfaceRealization();
		ir.setName(i.getLabel() + "_Provided");
		// TODO: configure()?
		if (!(c instanceof Port)) // TODO: review
			ir.getClients().add(c);
		ir.getSuppliers().add(i);
		ir.setContract(i);
		ir.setImplementingClassifier(co);
		/*
		 * see addClass() -- if the class is on top-level, don't add
		 * realization:
		 */
		if (c instanceof Class
				&& c.getVisibility().equals(VisibilityKind.PRIVATE_LITERAL)) {
			// only add one component realization per class/component pair
			boolean cr_for_c_exists = false;
			for (DirectedRelationship cr : c
					.getTargetDirectedRelationships(UMLPackage.eINSTANCE
							.getComponentRealization())) {
				final EList<Element> sources = ((ComponentRealization) cr)
						.getSources();
				if (sources.contains(co)) {
					cr_for_c_exists = true;
					break;
				} else if (sources.size() > 0)
					throw new RCOSModelCheckException(
							c.getQualifiedName()
									+ " already provided by at least "
									+ ((NamedElement) sources.get(0))
											.getQualifiedName());
			}
			if (!cr_for_c_exists) {
				ComponentRealization cor = co.createRealization(c.getName()
						+ "_Real");
				cor.getRealizingClassifiers().add((Class) c);
			}
		}
		return ir;
	}

	// For delegated interfaces
	public static void addProvidedInterface(ContractInterface provided,
			Component co, ContractInterface ci2, Component co2)
			throws RCOSModelCheckException {
		Port p = MyPort.createPort(co);
		p.setType(provided.getBase_Interface());
		// InterfaceRealization ir = co.createInterfaceRealization("IR_" +
		// i.getLabel(), null);
		addProvidedInterface(provided, co, p);

		// TODO: could use 'i2.createDependency(p);', but that might yield a
		// different owner?
		Dependency d = UMLFactory.eINSTANCE.createDependency();
		co.getPackagedElements().add(d);
		Delegation dl = Stereotypes.applyDelegation(d);
		dl.configure(p, ci2, provided);
		Interface i2 = ci2.getBase_Interface();
		d.setName("DI_" + co2.getLabel() + "." + i2.getLabel() + " -> "
				+ provided.getBase_Interface().getLabel());
	}

	public static void addInterfaceImplementation(Interface i, Class class1) {
		// Before we had assert(i != null), but when this method is used when
		// considering "provided Ctr by C",
		// the Interface of Ctr is not necessarily known at this time, since an
		// interface is attached to
		// a contract only in the last step (rCOSTreeToUML)
		// 2 solutions:
		// 1 - Introduce an intermediary step between rCOSTreeGetClasses and
		// rCOSTreeToUML which attaches the interface to its contract
		// 2 - Do not try to add an interface implementation when it is not
		// know, which is done here.
		if (i != null) {
			assert (class1 != null);
			InterfaceRealization ir = class1.getInterfaceRealization(
					"IR " + i.getLabel() + " " + class1.getLabel(), i, false,
					true);
			ir.setContract(i);
		}
	}

	/* These methods are for Enumerations, not their literals */
	public Enumeration addEnumeration(String eName, CommonTree token)
			throws RCOSModelCheckException {
		assert eName != null;
		Package classPackage = rcos.getClassModel().getBase_Package();
		if (classPackage.getOwnedType(eName) != null)
			throw new RCOSModelCheckException("Duplicate enumeration/class "
					+ eName + " in class-model package.", token);
		return classPackage.createOwnedEnumeration(eName);
	}

	public Enumeration getEnum(String eName, CommonTree token)
			throws RCOSModelCheckException {
		assert eName != null;
		Package classPackage = rcos.getClassModel().getBase_Package();
		Type t = classPackage.getOwnedType(eName);
		if (!(t instanceof Enumeration))
			throw new RCOSModelCheckException(eName
					+ " is not an Enumeration declaration.", token);
		return (Enumeration) t;
	}
}
