package edu.unu.iist.rcos.spec.antlr;

import org.antlr.runtime.RecognitionException;
import org.antlr.runtime.tree.CommonTree;

import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.TOKEN;

/**
 * The main parent of every exception defined within the spec.antlr project. 
 * It extends RecognitionException in order to be able to be thrown from the parser.
 * Unfortunately RecognitionException() doesn't track messages, so we do it explicitly.
 * @author Charles Morisset
 */
public class RCOSModelCheckException extends RecognitionException {
	private static final long serialVersionUID = -2821726165314428543L;
	final protected String msg;

	protected RCOSModelCheckException(String m) {
		msg = m;
	}

	public RCOSModelCheckException(String m, Throwable t) {
		this(m);
		initCause(t);
	}

	/* 
	 * @see java.lang.Throwable#getMessage()
	 */
	@Override
	public String getMessage(){
		return msg;
	}
	
	public static String mkString(String str, TOKEN token) {
		String res = str;
		if (token != null && token.getToken() != null){
			CommonTree t = (CommonTree) token.getToken();
			res = mkString(res, t);  
		} 
		return res;
	}
	
	protected static String mkString(String res, CommonTree t) {
		res = "Line " + t.getLine() + ":" + t.getCharPositionInLine() + "\n" + res;
		return res;
	}

	public RCOSModelCheckException(String string, TOKEN token) {
		this(mkString(string, token));
	}

	public RCOSModelCheckException(Throwable e, TOKEN type) {
		this(mkString(e.getMessage(),type),e);
	}
	
	public RCOSModelCheckException(Throwable e, CommonTree type) {
		this(mkString(e.getMessage(),type),e);
	}

	public RCOSModelCheckException(String s, CommonTree type) {
		this(mkString(s,type));
	}
}
