package edu.unu.iist.rcos.spec.antlr;

import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Package;

import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.profile.*;
import edu.unu.iist.rcos.uml.RCOSLoader;

/**
 * @author vs
 * This class may run UML model validation before checking the rCOS aspects of the model,
 * if you use the validate(RCOSModel rm) method.
 */
public class RCOSModelValidation extends RCOSModelValidationNoUML {
	
	public RCOSModelValidation() { super("rCOS UML"); }
	
	/**
	 * @see edu.unu.iist.rcos.spec.antlr.RCOSModelValidationNoUML#validate(edu.unu.iist.rcos.profile.RCOSModel)
	 */
	@Override
	public boolean validate(RCOSModel mod) throws RCOSModelCheckException, WellformedException {
		boolean valid = true;
		
		// Place holder so that we don't forget about this:
		// we check the model structure first, maybe we should this if we can do this
		// on loading?
		valid &= RCOSLoader.validate(mod);
		valid &= super.validate(mod);
		
		return valid;
	}

	/**
	 * @deprecated This is now an OCL rule in Class_OCL.
	 * @see edu.unu.iist.rcos.mm.ocl.Class_OCL#ClassModel_No_Subpackages
	 */
	@Deprecated
	public boolean validateReq(RCOSModel mod) throws RCOSModelCheckException, WellformedException {
		return validate(mod) && isReqRCOSModel(mod);
	}

	private boolean isReqRCOSModel(RCOSModel mod) throws WellformedException {		
		ClassModel cm = mod.getClassModel();
		for (Element e : cm.getBase_Package().getOwnedElements()){
			if (e instanceof Package) {
				Package p = (Package)e;
				// WF False because sub-packages are not allowed in the class model of a requirement model
				if (diag == null)
					throw new WellformedException("Sub-packages are not allowed:"+p.getQualifiedName());
				else
					dmsg(ErrorType.WF,"Sub-packages are not allowed:"+p.getQualifiedName(),p);
			} 
		}
		return true;
	}
}
