package edu.unu.iist.rcos.spec.antlr;

import java.util.Iterator;
import java.util.Map;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.emf.common.util.Diagnostic;
import org.eclipse.emf.common.util.DiagnosticChain;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.Diagnostician;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.uml2.uml.*;
import org.topcased.validation.core.IValidator;

import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.profile.*;

/**
 * This validator is a bit awkward in the sense that it will get called for EACH UML element in the model, but only
 * do something ONCE for the top-level model element.
 * TODO: Maybe in the future we can extends the plugin mechanism to match on the profile level. 
 * @author stolz
 */
public class RCOSValidator implements IValidator {

	/**
	 * @see org.topcased.validation.core.IValidator#validate(org.eclipse.emf.ecore.EObject,
	 *      org.eclipse.emf.common.util.DiagnosticChain, org.eclipse.core.runtime.IProgressMonitor)
	 */
	public boolean validate(EObject eObject, DiagnosticChain diagnostics, final IProgressMonitor monitor) {
		int count = 0;
		for (Iterator<EObject> i = eObject.eAllContents(); i.hasNext(); i.next())
		  { ++count; }

		monitor.beginTask("rCOS Validation", count);

		final RCOSModelValidationNoUML rmv = new RCOSModelValidationNoUML(diagnostics);
		Diagnostician diagnostician = new Diagnostician() {

			@Override
			public boolean validate(EClass eClass, EObject eObj, DiagnosticChain diag, Map<Object,Object> context) {
				monitor.worked(1);
				boolean result = true;
				if (eObj instanceof Model) {
					RCOSModel rm = Stereotypes.isRCOSModelUnsafe((Model)eObj);
					if (rm != null) {
						monitor.setTaskName("Validation of "+rm.getBase_Model().getName());
						try {
							result &= rmv.validate(rm);
						// we shouldn't end up in the exceptions below:
						} catch (RCOSModelCheckException e) {
							throw new RuntimeException("Unhandled exception, please report a bug!",e);
						} catch (WellformedException e) {
							throw new RuntimeException("Unhandled exception, please report a bug!",e);
						}
					}
				}
				return result;
			}
		};

		monitor.setTaskName("Validation of " + EcoreUtil.getURI(eObject));

		Diagnostic result = diagnostician.validate(eObject);
		diagnostics.merge(result);

		return true;
	}
}
