package edu.unu.iist.rcos.spec.antlr;

import static edu.unu.iist.rcos.profile.Stereotypes.getRootModel;
import static edu.unu.iist.rcos.profile.Stereotypes.isRCOSModel;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Parameter;
import org.eclipse.uml2.uml.Type;

import edu.unu.iist.rcos.profile.RCOSModel;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.DesignPrinter;
import edu.unu.iist.rcos.spec.antlr.visitor.ExpPrinter;
import edu.unu.iist.rcos.spec.antlr.visitor.PrinterException;
import edu.unu.iist.rcos.spec.antlr.visitor.ExpPrinter.FieldIsShadowed;
import edu.unu.iist.rcos.spec.antlr.visitor.check.ExpGetType;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;
import edu.unu.iist.rcos.spec.antlr.visitor.toUTF8.UTF8ExpPrinter;
import edu.unu.iist.rcos.spec.antlr.visitor.toUTF8.UTF8Printer;

public class VardeclsFIS extends Vardecls implements FieldIsShadowed, DesignExpPrinter {

	public final TypeHandler th;
	final Classifier c;

	public VardeclsFIS(Classifier c, TypeHandler th) {
		assert c != null;
		this.c = c;
		this.th = th;
	}

	public VardeclsFIS(Classifier c, RCOSModel rm) {
		this(c,new TypeHandler(rm, false));
	}

	/**
	 * @param c must not be null, and be already part of a model.
	 */
	@Deprecated
	public VardeclsFIS(Classifier c) {
		this(c,isRCOSModel(getRootModel(c)));
	}

	@Deprecated
	public VardeclsFIS(Operation op) {
		this(op, new TypeHandler((Classifier) op.getOwner()));
	}

	public VardeclsFIS(Operation op, TypeHandler th) {
		this((Classifier) op.getOwner(), th);
		/* Now push all parameters into declarations */
		for (Parameter p : op.getOwnedParameters()) {
			new Vardecl(p.getName(),p.getType(), VarType.PAR);
		}
	}

	/**
	 * Copy constructor
	 */
	public VardeclsFIS(VardeclsFIS ctx) {
		this(ctx.c,ctx.th);
		boolean seen = false;
		Iterator<List<Vardecl>> it = ctx.vardecls.descendingIterator();
		while (it.hasNext()) {
			List<Vardecl> vs = it.next();
			seen = true;
			for (Vardecl v : vs) {
				new Vardecl(v);
			}
			vardecls.push(new ArrayList<Vardecl>());
		}
		if (seen) // pop last empty frame.
			vardecls.pop();
	}

	public VardeclsFIS(Operation op, RCOSModel rm) {
		this(op, new TypeHandler(rm));
	}

	@Override
	public Type getType(TYPE t) throws GetTypeException {
		return th.getTypeFromTYPE(t, c);
	}

	public boolean isShadowed(VARNAME da) throws GetTypeException {
		return isLocal(da);
	}

	/* Some helpers */
	public DesignPrinter<?> mkDesignPrinter() {
		if (DesignUtility.Unicode)
			return new UTF8Printer(this, new UTF8ExpPrinter(this));
		else
			return new DesignPrinter<ExpPrinter>(this, new ExpPrinter(this));
	}

	public String designToString(DESIGN d) {
		try {
			return mkDesignPrinter().run(d);
		} catch (PrinterException e) {
			throw new RuntimeException("Couldn't print design",e);
		}		
	}

	public ExpPrinter mkExpPrinter() {
		if (DesignUtility.Unicode)
			return new UTF8ExpPrinter(this);
		else
			return new ExpPrinter(this);
	}

	public String expToString(EXP e) { 
		try {
			return mkExpPrinter().run(e);
		} catch (PrinterException e1) {
			// TODO: relevant for rCOS 123
			throw new RuntimeException("Couldn't print expression",e1);
		}
	}

	public Type getType(EXP e) throws GetTypeException {
		return new ExpGetType(this).run(e);
	}

	public Classifier getClassifier() {
		return c;
	}
}