package edu.unu.iist.rcos.spec.antlr.commands;

import java.util.List;

import org.antlr.runtime.RecognitionException;
import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.common.command.IdentityCommand;
import org.eclipse.emf.common.command.UnexecutableCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.UMLPackage;

import edu.unu.iist.rcos.mm.RCOSCommand;
import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.profile.AssertNotNullParameter;
import edu.unu.iist.rcos.profile.ProfileFactory;
import edu.unu.iist.rcos.profile.SOE_Element;
import edu.unu.iist.rcos.spec.antlr.*;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DASHABLE;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.EXP;
import edu.unu.iist.rcos.spec.antlr.visitor.VisitorException;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement.RefinementException;

/**
 * TODO: Can't handle parameters (obviously).
 * @author stolz
 *
 */
public class AssertNotNullCommand extends RCOSCommand<AssertNotNullParameter> {

	private VardeclsFIS vdcls;

	public AssertNotNullCommand(){}

	public AssertNotNullCommand(EditingDomain domain, List<Throwable> errors, AssertNotNullParameter p) {
		super(domain,errors,p);
	}

	public static AssertNotNullCommand mkAssertNotNullCommand(EditingDomain domain, List<Throwable> errors, String name, Class c) {
		AssertNotNullParameter p = ProfileFactory.eINSTANCE.createAssertNotNullParameter();
		p.setDashable(name);
		SOE_Element soe = ProfileFactory.eINSTANCE.createSOE_Element();
		soe.setMetaclass(UMLPackage.eINSTANCE.getClassifier());
		soe.setElement(c);
		p.setContext(soe);
		return new AssertNotNullCommand(domain, errors, p);
	}

	@Override protected Command createCommand() {
		if (params.getContext() == null)
			return recordError(new IllegalArgumentException(params.toString()));
		// TODO: Only for classifiers at the moment.
		if (params.getContext().getMetaclass() != null && !UMLPackage.eINSTANCE.getClassifier().isSuperTypeOf(params.getContext().getMetaclass()))
			return recordError(new IllegalArgumentException());

		Classifier ctx = (Classifier) params.getContext().getElement();
		try {
			AssertNotNullVisitor av = new AssertNotNullVisitor(getDashable());

			final CompoundCommand cmd = new CompoundCommand("assert not null "+params.getDashable()+ " in "+ctx.getName());
			cmd.append(IdentityCommand.INSTANCE);

			DesignsED desEd = new DesignsED(domain);
			GuardsED guardsEd = new GuardsED(domain);
			for (Operation op : ctx.getOperations()) {
				final DESIGN designStr = Parse.getDesign(op);
				final VardeclsFIS vdOp = new VardeclsFIS(op, getVdcls().th);
				if (designStr != null) { // TODO: pattern
					DESIGN v = av.new DesignVisitor(ctx).run(designStr);
					if (v != null)
						cmd.append(desEd.update(op, vdOp.designToString(DesignUtility.simplify(v,ctx))));
				}
				cmd.append(mogrify(guardsEd, vdOp, ctx,av));
			}

			cmd.append(mogrify(new InvariantsED(domain), getVdcls(), ctx, av));
			return cmd.unwrap();
		} catch (RecognitionException e) {
			return recordError(e);
		} catch (WellformedException e) {
			return recordError(e);
		} catch (RefinementException e) {
			return recordError(e);
		} catch (VisitorException e) {
			return recordError(e);
		} catch (TransformationFailedException e) {
			return recordError(e);
		}
	}

	private static Command mogrify(EFRCOSRuleFeature ed, VardeclsFIS vdcls, Classifier ctx, AssertNotNullVisitor av) throws WellformedException, RecognitionException, RefinementException, VisitorException {
		final String inv = ed.get(ctx);
		if (inv != null) {
			EXP invExp = Parse.getExp(inv, vdcls);
			EXP invNew = av.run(invExp);

			// TODO: pattern
			invNew = DesignUtility.simplifyExpression(invNew);
			if (MyDesign.isTrue(invNew))
				return ed.destroy(ctx);
			else if (MyDesign.isFalse(invNew)) // XXX haha, you wish
				return UnexecutableCommand.INSTANCE;
			else
				return ed.update(ctx, vdcls.expToString(invNew));			
		} else
			return IdentityCommand.INSTANCE;
	}

	private VardeclsFIS getVdcls() {
		if (vdcls == null)
			vdcls = new VardeclsFIS((Classifier) params.getContext().getElement());
		assert vdcls != null;
		return vdcls;
	}

	protected DASHABLE getDashable() throws TransformationFailedException, RecognitionException, WellformedException {
		EXP exp = Parse.getExp(params.getDashable(), getVdcls());
		if (!(exp instanceof DASHABLE))
			throw new TransformationFailedException("Expression is not a field/navigation path or parameter.");
		return (DASHABLE) exp;
	}

	@Override protected Command recordError(Throwable e) {
		// TODO: how to set exception context here? :-/
		return super.recordError(e);
	}
}