package edu.unu.iist.rcos.spec.antlr.commands;

import java.util.Collection;
import java.util.Collections;

import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.StrictCompoundCommand;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.common.edit.command.ChangeCommand;
import org.eclipse.uml2.uml.Element;

public class EMFDeleteProvider implements DeleteProvider<Command> {
	final StrictCompoundCommand state = new StrictCompoundCommand() {{ isPessimistic = true; }};
	final EditingDomain domain;

	public EMFDeleteProvider(EditingDomain domain, Command cmd) {
		this.domain = domain;
		state.append(cmd);
	}

	public void delete(EObject e) {
		state.append(createDeleteCommand(domain, Collections.singleton(e)));
	}

	private static ChangeCommand createDeleteCommand(EditingDomain domain, final Collection<EObject> elements) {
		return new ChangeCommand(domain, new Runnable() {
			public void run() {
				for (EObject e : elements) {
					((Element)e).destroy();
				}
			}
		}){
			/* https://bugs.eclipse.org/bugs/show_bug.cgi?id=323719
			 */
			@Override public boolean canUndo() {return true;}
		};
	}

	public Command get() {
		return state;
	}

	public void add(Command cmd) {
		state.append(cmd);
	}
}