package edu.unu.iist.rcos.spec.antlr.commands;

import java.util.List;

import org.eclipse.emf.common.command.*;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.Class;

import edu.unu.iist.rcos.mm.RCOSCommand;
import edu.unu.iist.rcos.profile.ExpertPatternClassParameters;
import edu.unu.iist.rcos.profile.ProfileFactory;
import edu.unu.iist.rcos.spec.antlr.RCOSModelUtility;
import edu.unu.iist.rcos.spec.transformer.DesignTransformerEPGetED;
import edu.unu.iist.rcos.spec.transformer.DesignTransformerEPSetED;

public class ExpertPatternClassCommand extends RCOSCommand<ExpertPatternClassParameters> {

	public ExpertPatternClassCommand(EditingDomain ed, List<Throwable> errors,
			ExpertPatternClassParameters p) {
		super(ed, errors, p);
	}
	
	public ExpertPatternClassCommand(){}
	
	@Override public Command createCommand() {
		try {
			Class c = params.getClass_();
			CompoundCommand cmp = new StrictCompoundCommand("apply expert pattern on "+ c.getName()) {{ isPessimistic = true; }};
			cmp.append(RCOSModelUtility.transform(domain, c, new DesignTransformerEPSetED(domain)));
			cmp.append(RCOSModelUtility.transform(domain, c, new DesignTransformerEPGetED(domain)));
			return cmp.unwrap(); // micro-optimization
		} catch (Exception e) {
			return recordError(e);
		}
	}

	public static Command runExpertPattern(EditingDomain ed, List<Throwable> errors, Class c) {
		ExpertPatternClassParameters ep = ProfileFactory.eINSTANCE.createExpertPatternClassParameters();
		ep.setClass_(c);
		return new ExpertPatternClassCommand(ed, errors, ep);
	}
}
