package edu.unu.iist.rcos.spec.antlr.commands;

import java.util.List;

import org.antlr.runtime.RecognitionException;
import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.Operation;

import edu.unu.iist.rcos.mm.RCOSCommand;
import edu.unu.iist.rcos.mm.util.Designs;
import edu.unu.iist.rcos.mm.util.DesignsED;
import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.profile.Manualrefine;
import edu.unu.iist.rcos.profile.OperationQ;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.Parse;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement.RefinementException;
import edu.unu.iist.rcos.spec.tc.RuleTCMembersED;
import edu.unu.iist.rcos.spec.tc.RuleTCMembersED.RuleOperationDesign;

public class ManualRefineCommand extends RCOSCommand<Manualrefine> {

	public ManualRefineCommand(){}

	public ManualRefineCommand(EditingDomain ed, List<Throwable> errors,
			Manualrefine p) {
		super(ed, errors, p);
	}

	/**
	 * @param editingDomain
	 * @param p
	 * @return Command
	 * @throws RefinementException if parameters are missing.
	 */
	@Override
	public Command createCommand() {
		try {
			final OperationQ c_op = params.getOperation();
			if (c_op == null)
				throw new RefinementException("Operation not set on "+params.getBase_Class().getQualifiedName());
			Operation op = c_op.getOperation();
			setLabel("refine "+op.getQualifiedName());

			final String c_newDesign = params.getNewDesign();
			if (c_newDesign == null)
				throw new RefinementException("No new design specified!");
			final String c_oldDesign = params.getOldDesign();
			if (c_oldDesign != null) {
				DESIGN modelDesign = Parse.getDesign(op);

				// XXX This is not fool proof yet, more sanity checks required.

				// Workaround: if the old design is empty (not missing!), then any new design will do.
				// We compare the designs once parsed, to avoid the problem of white spaces
				// XXX ideally, this approach will eventually allow to consider alpha-conversion or commutativity
				if (!"".equals(c_oldDesign) && !DesignUtility.isDesignEqual(
						Parse.getDesign(c_oldDesign, op), 
						modelDesign)) {
					throw new RefinementException("Old design does not match for operation "+ op.getQualifiedName()+" (in Manual Refinement action "+params.getBase_Class().getQualifiedName()+")!\nExpected:"+c_oldDesign+
							"\nFound:   "+Designs.iNSTANCE.get(op));
				}
			}

			check(op, c_newDesign);

			// Just to get the pretty print of the new design 
			String newDesign = DesignUtility.designToString(op,Parse.getDesign(c_newDesign, op));
			return new DesignsED(domain).update(op, newDesign);
		}
		// We don't distinguish between syntax- and type errors:
		catch (Exception e) { return recordError(e); }
	}

	/**
	 * Check validity of new design. This method is a hook for overriding
	 * the type-checking, here with the default checker.
	 */
	protected void check(Operation op, final String c_newDesign)
			throws RecognitionException, WellformedException {
		RuleTCMembersED rtc = new RuleTCMembersED(domain);
		RuleOperationDesign chk = rtc.new RuleOperationDesign();
		chk.check(op, c_newDesign);
	}
}