package edu.unu.iist.rcos.spec.antlr.visitor;

import edu.unu.iist.rcos.mm.Visitor;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;

/**
 * @author charles
 * Same as ExpGenVisitor: only takes care of the cast. 
 * @param <E>
 * @param <EV>
 */
public abstract class DesignBaseVisitor<RD,RE,RT, 
	E  extends VisitorException,
 	TV extends TypeBaseVisitor<RT, E>, 
 	EV extends ExpGenVisitor<RE,RT,E,TV>> extends Visitor<RD,DESIGN,E> {
	
	// [vs] should this one be 'final'?
	/**
	 * If you override this method, remember to set "init" to true.
	 */
	public void visit(DESIGN d) throws E {
		init = true;
		if (d instanceof COMMAND){
			visit((COMMAND) d);
		} else if (d instanceof PP) {
			visit((PP) d);
		} else if (d instanceof IF){
			visit((IF) d);
		} else if (d instanceof SEQ){
			visit((SEQ) d);
		} else if (d instanceof CHOICE){
			visit((CHOICE) d);
		} else if (d instanceof WHILE){
			visit((WHILE) d);
		} else if (d instanceof FOR) {
			visit((FOR) d);
		}
	}

	abstract protected void visit(MCALL m) throws E;
	abstract protected void visit(PP d) throws E;
	abstract protected void visit(IF d) throws E;
	abstract protected void visit(CHOICE d) throws E;
	abstract protected void visit(SEQ d) throws E;
	abstract protected void visit(WHILE d) throws E;
	abstract protected void visit(FOR d) throws E;
	
	protected void visit(COMMAND d) throws E{
		if (d instanceof VARDECL){
			visit((VARDECL) d);
		} else if (d instanceof ENDDECL) {
			visit((ENDDECL) d);
		} else if (d instanceof ASSIGN){
			visit((ASSIGN) d);
		} else if (d instanceof SKIP){
			visit((SKIP) d);
		} else if (d instanceof CHAOS){
			visit((CHAOS) d);
		} else if (d instanceof MAGIC){
			visit((MAGIC) d);
		} else if (d instanceof STOP){
			visit((STOP) d);
		}else if (d instanceof MCALL){
			visit((MCALL) d);
		} else if (d instanceof RET) {
			visit((RET) d);
		} else if (d instanceof CHOKE) {
			visit((CHOKE) d);
		} else
			throw new IllegalArgumentException();
	}	
	
	abstract protected void visit(ASSIGN a) throws E;
	abstract protected void visit(VARDECL d) throws E;
	abstract protected void visit(ENDDECL d) throws E;
	abstract protected void visit(SKIP d) throws E;
	abstract protected void visit(CHAOS d) throws E;	
	abstract protected void visit(MAGIC d) throws E;
	abstract protected void visit(STOP d) throws E;
	abstract protected void visit(RET d) throws E;
	abstract protected void visit(CHOKE d) throws E;

	@Override public RD run(DESIGN e, RD r) {
		try {
			return run(e);
		} catch (VisitorException ex) {
			return r;
		}
	}
}
