package edu.unu.iist.rcos.spec.antlr.visitor;

import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import static edu.unu.iist.rcos.spec.antlr.MyDesign.*;

public class DesignGenCopy<VE extends VisitorException, EC extends ExpTypeVisit<VE, TypeCopy<VE>>> 
extends DesignVisitor<DESIGN,EXP,TYPE,VE,TypeCopy<VE>, EC> {
	
	public DesignGenCopy(EC e) {
		super(e);
	}

	protected EXP getCopy(EXP e) throws VE {
		return ev.run(e);
	}
	
	@Override
	protected void visit(PP d) throws VE { 
		EXP pre = getCopy(d.getPre());
		DESIGNEXP post = d.getPost() instanceof MCALL ? run((MCALL) d.getPost()) : getCopy(d.getPost());
		/* Someone may return a full design in post, like the AbsSetOps. In that case we create a new sequential composition.
		 * We handle MCALLs in the EXP-branch.
		 */
		if (post instanceof EXP)
			set(createPP(pre, (EXP)post));
		else
			set(createSeq(createPP(pre, createTrue()), (DESIGN)post));
	}
	
	protected void visit(CHOICE c) throws VE{
		DESIGN left = run(c.getLeft());
		DESIGN right = run(c.getRight());
		set(createChoice(left, right));
	}
	
	protected void visit(SEQ c) throws VE{
		DESIGN left = run(c.getLeft());
		assert left != null : c;
		DESIGN right = run(c.getRight());
		assert right != null : c;
		set(createSeq(left, right));
	}
	
	protected void visit(IF c) throws VE{
		DESIGN left = run(c.getLeft());
		DESIGN right = run(c.getRight());
		EXP b = getCopy(c.getB());
		set(createIf(b, left, right));
	}

	protected void visit(WHILE c) throws VE{
		DESIGN d1 = run(c.getDo());
		EXP b = getCopy(c.getB());
		set(createWhile(b, d1));
	}
	
	protected void visit(FOR d) throws VE {
		EXP val = getCopy(d.getVal());
		DESIGN d1 = run(d.getDo());
		set(createFor(ev.run(d.getType()), 
				createVarname(d.getVar()), (VAL) val, d1));
	}
	
	protected void visit(VARDECL vd) throws VE{
		set(createVarDecl(ev.run(vd.getType()), 
				createVarname(vd.getVar())));
	}
	
	protected void visit(ENDDECL ed){
		set(createEndDecl(createVarname(ed.getVar())));
	}
	
	protected void visit(ASSIGN a) throws VE{
		REF r = (REF) getCopy(a.getVal());
		set(createAssign(r, getCopy(a.getE())));
	}
	
	protected void visit(CHAOS c){
		set(createChaos());
	}
	
	protected void visit(MAGIC c){
		set(createMagic());
	}

	protected void visit(STOP c){
		set(createStop());
	}

	protected void visit(SKIP s){
		set(createSkip());
	}
	
	protected void visit(MCALL m) throws VE{
		set((DESIGN) getCopy((EXP) m));
	}
	
	protected void visit(RET r) throws VE {
		ev.set(null); // don't spam RET with an old value.
		set(createRet(r.getE() == null ? null : getCopy(r.getE())));
	}

	@Override
	protected void visit(CHOKE d) throws VE {
		set(SyntaxTree_ClassModelFactory.eINSTANCE.createCHOKE());
	}
}
