package edu.unu.iist.rcos.spec.antlr.visitor;

import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.spec.antlr.MyDesign;
import edu.unu.iist.rcos.spec.antlr.Vardecls;
import edu.unu.iist.rcos.spec.antlr.VardeclsFIS;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

public class DesignPrinter<EP extends ExpPrinter> extends DesignGenVisitor<String,String,String,PrinterException, TypePrinter<PrinterException>, EP> {

	protected String TSTILE; 
	private String tab = "";
	protected Vardecls vds; // TODO: Should be final

	private void incTab() {
		tab += "\t";
	}

	private void decTab() {
		if (! tab.equals(""))
			tab = tab.substring(1);
	}

	protected DesignPrinter(EP ep, Operation op) {
		this(new VardeclsFIS(op),ep);
	}

	protected DesignPrinter(EP ep, final Classifier c) {
		this(new VardeclsFIS(c),ep);
	}

	public DesignPrinter(Vardecls vfis, EP ep) {
		super(ep);
		TSTILE = "|-";
		state = "";
		vds = vfis;
	}

	/**
	 * This design printer does know nothing about types and fields/parameters, apart from what the parser told him.
	 * It should still be safe to use if you programatically transformed your AST, but you will get redundant "this."
	 * prefixes for field accesses.
	 */
	@Deprecated
	protected DesignPrinter(EP ep) {
		this(new Vardecls() {
			@Override
			public Type getType(TYPE t) { return null; }
		},ep);
	}

	protected String getString(EXP e) throws PrinterException{
		// TODO: should clear state!
		ev.set("");
		return ev.run(e);
	}

	@Override
	protected void visit(PP d) throws PrinterException{
		EXP pre = d.getPre();
		set(tab + "[ "+ (MyDesign.isTrue(pre) ? "" : getString(pre)) + TSTILE + getString(d.getPost()) + " ]"); 
	}

	@Override
	protected void visit(IF d) throws PrinterException {
		String ret = tab + "if ("+ getString(d.getB())+ ") then { \n";
		incTab();
		ret += run(d.getLeft()) + "\n "; 
		decTab();
		ret +=  tab + "} else {\n" ; 
		incTab();
		ret += run(d.getRight()) + "\n ";
		decTab();
		ret += tab + "}";
		set(ret);
	}

	@Override
	protected void visit(SEQ d) throws PrinterException{
		set(run(d.getLeft())+ ";\n" + run(d.getRight()));
	}

	@Override
	protected void visit(CHOICE d) throws PrinterException{
		final String l = run(d.getLeft());
		String r = run(d.getRight());
		if (d.getRight() instanceof SEQ)
			r = "("+r+")";
		set(l+ " \\/  "+ r);
	}

	@Override
	protected void visit(WHILE d) throws PrinterException{
		String ret = tab + "while ("+ getString(d.getB())+ ") { \n";
		incTab();
		ret += run(d.getDo());
		decTab();
		ret += "\n" + tab + "}";
		set(ret);
	}

	@Override
	protected void visit(FOR d) throws PrinterException {		
		String ret = tab + "for (" + ev.run(d.getType()) + " " + d.getVar().getName() + " : " + getString(d.getVal())+ ") {\n";
		vds.pushDeclFrame();
		try {
			vds.new Vardecl(d);
		} catch (GetTypeException e) {
			// XXX: review?!
//			throw new PrinterException(e);
		}
		incTab();
		ret += run(d.getDo()) + "\n";
		decTab();
		ret += tab + "}";
		set(ret);
		vds.popDeclFrame();
	}

	@Override
	protected void visit(ASSIGN a) throws PrinterException{
		set (tab + getString(a.getVal())+ " := "+ getString(a.getE()));
	}

	@Override
	protected void visit(VARDECL d) throws PrinterException {
		try {
			vds.new Vardecl(d);
		} catch (GetTypeException e) {
//			throw new PrinterException(e);
		}
		set(tab + "var " + ev.run(d.getType()) + " "+ d.getVar().getName());
	}

	@Override
	protected void visit(ENDDECL d) throws PrinterException {
		set(tab + "end " + d.getVar().getName());
		try {
			vds.remove(d.getVar());
		} catch (GetTypeException e) {
//			throw new PrinterException(e);
		}
	}

	@Override
	protected void visit(SKIP d) {
		set(tab + "skip");
	}

	@Override
	protected void visit(CHAOS d) throws PrinterException {
		set(tab + "chaos");
	}			

	@Override
	protected void visit(MAGIC d) throws PrinterException {
		set(tab + "magic");
	}

	@Override
	protected void visit(RET r) throws PrinterException {
		final String e;
		if (r.getE() == null)
			e = "";
		else
			e = " " + getString(r.getE());
		set(tab + "return" + e);
	}

	@Override
	protected void visit(MCALL m) throws PrinterException {
		set(tab + ev.run(m));
	}

	@Override
	protected void visit(STOP d) throws PrinterException {
		set(tab + "stop");
	}

	@Override
	protected void visit(CHOKE d) throws PrinterException {
		throw new UnsupportedOperationException();
	}
}
