package edu.unu.iist.rcos.spec.antlr.visitor;

import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;

/**
 * A deep visitor of designs, based on DesignGenVisitor. Doesn't do anything, apart from 
 * going deep down the design.
 * @param <RD>
 * @param <RE>
 * @param <RT>
 * @param <E>  extends VisitorException
 * @param <TV> extends TypeBaseVisitor RT,E
 * @param <EV> extends ExpVisitor RE,RT,E,TV
 * @author charles
 */
public abstract class DesignVisitor<RD,RE,RT,
	E extends VisitorException, 
	TV extends TypeBaseVisitor<RT, E>,
	EV extends ExpVisitor<RE,RT,E,TV>> extends DesignGenVisitor<RD,RE,RT,E,TV,EV> {
	
	protected DesignVisitor(EV ev) {
		super(ev);
	}

	protected void visit(MCALL m) throws E{
		visit((EXP) m);
	}	

	protected void visit(PP d) throws E{
		visit(d.getPre());
		if (d.getPost() instanceof MCALL)
			visit((MCALL)d.getPost());
		else
			visit(d.getPost());
	}
	
	protected void visit(IF d) throws E{
		visit(d.getB());
		visit(d.getLeft());
		visit(d.getRight());
	}
	
	protected void visit(CHOICE d) throws E{
		visit(d.getLeft());
		visit(d.getRight());
	}
	
	protected void visit(SEQ d) throws E {
		assert d.getLeft() != null;
		visit(d.getLeft());
		assert d.getRight() != null;
		visit(d.getRight());
	}
	protected void visit(WHILE d) throws E{
		visit(d.getB());
		visit(d.getDo());
	}
	
	protected void visit(FOR d) throws E {
		ev.visit(d.getType());
		ev.visit(d.getVar());
		visit(d.getVal());
		visit(d.getDo());
	}
	
	protected void visit(ASSIGN a) throws E{
		visit(a.getVal());
		visit(a.getE());
	}
	
	protected void visit(VARDECL d) throws E{
		ev.visit(d.getType());
		ev.visit(d.getVar());
	}	

	protected void visit(ENDDECL d) throws E{
		ev.visit(d.getVar());
	}

	/* TODO: These defaults here ARE DANGEROUS! */
	@Override protected void visit(SKIP d) throws E { return; }
	
	@Override protected void visit(MAGIC d) throws E { return; }

	@Override protected void visit(CHAOS d) throws E { return; }

	@Override protected void visit(STOP d) throws E { return; }
	
//	@Override protected void visit(CHOKE d) throws E { return; }

	@Override
	protected void visit(RET d) throws E {
		visit(d.getE());
	}
}
