package edu.unu.iist.rcos.spec.antlr.visitor;

import edu.unu.iist.rcos.mm.Visitor;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;

/**
 * A general visitor for expressions. Only takes care of the cast of general classes. 
 * @author charles
 * @param <E>
 */
public abstract class ExpGenVisitor<R, RT, E extends VisitorException, TV extends Visitor<RT, TYPE, E>> extends Visitor<R,EXP,E> {

	protected TV tv;

	/**
	 * If you override this method, remember to set "init" to true.
	 */
	public void visit(EXP d) throws E {
		init = true;
		if (d instanceof BINEXP){
			visit((BINEXP) d);
		} else if (d instanceof INSTANCEOF){
			visit((INSTANCEOF) d);  
		} else if (d instanceof NEG){
			visit((NEG) d); 
		} else if (d instanceof VAL){
			visit((VAL) d); 
		} else if (d instanceof PREDICATE) {
			visit((PREDICATE) d);
		} else if (d instanceof PREFIX) {
			visit((PREFIX) d);
		} else
		{ /* NADA: visiting null */ }
	}
	
	protected void visit(BINEXP e) throws E{
		if (e instanceof RELOP){
			visit((RELOP) e);
		} else if ( e instanceof AOP){
			visit((AOP) e);
		} else if (e instanceof LOGOP) {
			visit((LOGOP) e);
		} else
			throw new UnsupportedOperationException();	
	}	
	
	protected void visit(VAL d) throws E {
		if (d instanceof CONST){
			visit((CONST) d);
		} else if (d instanceof REF) {
			visit((REF) d);
		} else
			throw new UnsupportedOperationException(); 
	}
	
	protected void visit(CONST c) throws E{
		if (c instanceof BASEBOOL)
			visit((BASEBOOL) c);
		else if (c instanceof BASEINT)
			visit((BASEINT) c);
		else if (c instanceof BASESTRING)
			visit((BASESTRING) c);
		else if (c instanceof BASEDOUBLE)
			visit((BASEDOUBLE) c);
		else if (c instanceof BASEENUM)
			visit((BASEENUM) c);
		else
			throw new UnsupportedOperationException();
	}
	

	protected void visit(REF d) throws E {
		if (d instanceof MCALL){
			visit((MCALL) d);
		} else if (d instanceof DASHABLE){
			visit((DASHABLE) d);
		} else if (d instanceof NULL) {
			visit((NULL) d);
		} else if (d instanceof SUPER) {
			visit((SUPER) d);
		} else if (d instanceof NEW) {
			visit((NEW) d);
		} else if (d instanceof CAST) {
			visit((CAST) d);
		} else if (d instanceof THIS) {
			visit((THIS) d);
		} else
			throw new UnsupportedOperationException(); 
	}
	
	protected void visit(PREFIX p) throws E {
		if (p instanceof CNAME) { 
			visit((CNAME) p);
		} else if (p instanceof REF){
			visit((REF) p);
		} else
			throw new UnsupportedOperationException();
	}
	
	protected void visit(DASHABLE d) throws E {
		if (d instanceof PAR){
			visit((PAR) d);
		} else if (d instanceof VAR){
			visit((VAR) d);
		} else if (d instanceof FA){
			visit((FA) d);
		} else
			throw new UnsupportedOperationException();
	}

	protected void visit(TYPE t) throws E {
		tv.visit(t);
	}
	
	abstract protected void visit(RELOP e) throws E ;
	abstract protected void visit(LOGOP e) throws E ;
	abstract protected void visit(AOP e) throws E ;
	abstract protected void visit(INSTANCEOF d) throws E;
	abstract protected void visit(NEG d) throws E ;
	abstract protected void visit(VAR d) throws E ;
	abstract protected void visit(PAR d) throws E ;
	abstract protected void visit(FA d) throws E ;
	abstract protected void visit(THIS d) throws E ;
	abstract protected void visit(SUPER d) throws E ;
	abstract protected void visit(NULL d) throws E ;
	abstract protected void visit(NEW d) throws E ;
	abstract protected void visit(MCALL d) throws E ;
	abstract protected void visit(CAST d) throws E ;
	abstract protected void visit(PREDICATE d) throws E;
	abstract protected void visit(ITERATOR d) throws E;
	abstract protected void visit(CNAME c) throws E ;
	abstract protected void visit(VARNAME var) throws E ;
	abstract protected void visit(BASEBOOL d) throws E;
	abstract protected void visit(BASEINT d) throws E;
	abstract protected void visit(BASESTRING d) throws E;
	abstract protected void visit(BASEDOUBLE d) throws E;
	abstract protected void visit(BASEENUM d) throws E;

	public RT run (TYPE t) throws E {
		return tv.run(t);
	}
	
	@Override public R run(EXP e, R r) {
		try {
			return run(e);
		} catch (VisitorException ex) {
			return r;
		}
	}
}