package edu.unu.iist.rcos.spec.antlr.visitor;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.common.util.EList;

import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import static edu.unu.iist.rcos.spec.antlr.MyDesign.*;

public abstract class ExpTypeVisit<VE extends VisitorException, TV extends TypeGenVisitor<TYPE, VE>> extends ExpVisitor<EXP,TYPE,VE,TV> {

	protected ExpTypeVisit(TV tv) {
		super(tv);
	}
	
	@Override
	protected void visit(CNAME c) throws VE {
		set(createCname(c.getName()));
	}

	protected void visit(VAR v) throws VE{
		VARNAME vn = createVarname(v.getName());
		TYPE t = tv.run(v.getType());
		set(createVar(vn, t ,v.isDash()));
	}

	protected void visit(PAR p) throws VE{
		VARNAME vn = createVarname(p.getName());
		TYPE t = tv.run(p.getType());
		set(createPar(vn, t, p.isDash()));
	} 
	
	protected void visit(FA f) throws VE {
		VARNAME vn = createVarname(f.getName());
		set(createFA((PREFIX) run(f.getOwner()), vn, f.isDash(),null));
	} 
	
	protected void visit(NULL n) {
		set (createNull());
	}

	protected void visit(THIS t){
		set(createThis());
	} 
	
	protected void visit(SUPER t){
		set(createSuper());
	} 
	
	protected void visit(MCALL m) throws VE{
		VARNAME vn = createVarname(m.getName());
		PREFIX new_owner = (PREFIX) run(m.getOwner());
		List<EXP> mylexp = visitArgs(m.getIn());
		List<REF> lref = visitArgs(m.getOut());
		set(createMcall(new_owner, vn, mylexp, lref));
	}

	protected void visit(NEW m) throws VE{
		GTYPE i = (GTYPE) tv.run(m.getName());
		List<EXP> mylexp = visitArgs(m.getIn());
		List<REF> lref = visitArgs(m.getOut());
		set(createNew(i, mylexp, lref));
	}

	private <T extends EXP> List<T> visitArgs(EList<T> l) throws VE {
		List<T> lout = new ArrayList<T>();
		for (T o : l) {
				lout.add((T) run(o));
		}
		return lout;
	}

	@Override
	protected void visit(PREDICATE p) throws VE{
		if (p instanceof ITERATOR) {
			visit((ITERATOR)p);
			return;
		}
		EXP e = run(p.getE());
		if (p instanceof FORALL) 
			set(createForall(e, tv.run(p.getType()), 
					createVarname(p.getVar())));
		else if (p instanceof EXIST)
			set(createExist(e, tv.run(p.getType()), 
					createVarname(p.getVar())));
	}

	@Override
	protected void visit(ITERATOR p) throws VE{
		EXP e = run(p.getE());
		EXP r = run(p.getRange());
		set(createIterator(e, tv.run(p.getType()), 
				createVarname(p.getVar()), r));		
	}

	@Override
	protected void visit(INSTANCEOF d) throws VE {
		set(createInstanceOf((REF)run(d.getRef()), (CNAME)run(d.getClass_())));
	}

	protected void visit(NEG d) throws VE {
		set(createNeg(run(d.getE())));
	}

	@Override
	protected void visit(BASEBOOL d) throws VE {
		if(d.isBVal())
			set(createTrue());
		else 
			set(createFalse());
	}
	
	@Override
	protected void visit(BASESTRING d) throws VE {
		set(createBaseString(d.getVal()));
	}
	@Override
	protected void visit(BASEDOUBLE d) throws VE {
		set(createBaseDouble(d.getVal()));
	}
	@Override
	protected void visit(BASEINT d) throws VE {
		set(createBaseInt(d.getVal()));
	}
	@Override
	protected void visit(BASEENUM d) throws VE {
		set(createBaseEnum(d.getVal()));
	}

	protected void setBINEXP(BINEXP b, EXP left, EXP right) {
		if (b instanceof RELOP) {
			set(createRelOp(((RELOP)b).getOp(), left, right));
		} else if (b instanceof LOGOP) {
			set(createLogOp(((LOGOP)b).getOp(), left, right));
		} else if (b instanceof AOP) {
			set(createAOp(((AOP)b).getOp(), left, right));
		} else
			throw new RuntimeException("not reached");
	}

	@Override
	protected void visit(CAST c) throws VE {
		VAL v2 = (VAL) run(c.getVal());
		TYPE t2 = run(c.getType());
		set(createCast(t2, v2));
	}
}
