package edu.unu.iist.rcos.spec.antlr.visitor;

import org.eclipse.emf.common.util.EList;

import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;

/**
 * @author charles
 * A deep visitor of expressions, based on ExpGenVisitor. Doesn't do anything, apart from 
 * going deep down the expression. 
 * @param <E>
 */
public abstract class ExpVisitor<R, RT, E extends VisitorException, TV extends TypeBaseVisitor<RT, E>> extends ExpGenVisitor<R,RT,E,TV> {

	public ExpVisitor(TV t) {
		tv = t;
	}

	protected void visit(AOP e) throws E {	
		visit(e.getLeft());
		visit(e.getRight());
	}

	protected void visit(RELOP e) throws E {	
		visit(e.getLeft());
		visit(e.getRight());
	}

	protected void visit(LOGOP e) throws E {	
		visit(e.getLeft());
		visit(e.getRight());
	}

	protected void visit(INSTANCEOF d) throws E {
		visit(d.getRef());
		visit(d.getClass_());
	}

	protected void visit(NEG d) throws E {
		visit(d.getE());
	}

	protected void visit(VAR d) throws E {
		visit(d.getName());
	}

	protected void visit(PAR d) throws E {
		visit(d.getName());
	}

	protected void visit(FA d) throws E {
		visit(d.getOwner());
		visit(d.getName());
	}

	protected void visit(THIS d) throws E {
		return;
	}

	protected void visit(SUPER d) throws E {
		return;
	}

	protected void visit(NULL d) throws E {
		return;
	}

	protected void visit(NEW d) throws E {
		visit(d.getName());
		visit(d.getIn());
		visit(d.getOut());
	}

	protected void visit(MCALL d) throws E {
		visit(d.getOwner());
		visit(d.getName());
		visit(d.getIn());
		visit(d.getOut());
	}

	protected void visit(PREDICATE d) throws E {
		if (d instanceof ITERATOR)
			visit((ITERATOR)d);
		else {
			visit(d.getType());
			visit(d.getVar());
			visit(d.getE());
		}
	}

	protected void visit(ITERATOR d) throws E {
		visit(d.getType());
		visit(d.getVar());
		visit(d.getE());
		visit(d.getRange());
	}

	/**
	 * Visit a list of elements. Used in method calls and new expressions, to visit the parameters.
	 * It is not possible to directly define two methods visit(EList<EXP>) and visit(EList<VARNAME>), 
	 * because of the way generic types are handled in Java. 
	 * 
	 * @param l
	 * @throws VisitorException
	 */
	protected void visit(EList<?> l) throws E {
		for (Object o : l){
			if (o instanceof EXP) {
				visit((EXP) o);
			} else if (o instanceof VARNAME) {
				assert false;
				visit((VARNAME) o);
			} else
				assert false;
		}
	}

	/**
	 * @author Charles Morisset
	 * @param d
	 * @throws E
	 */
	@Override
	protected void visit(CAST d) throws E {
		visit(d.getType());
		visit(d.getVal());
	}
}