package edu.unu.iist.rcos.spec.antlr.visitor.check;

import java.util.ArrayList;
import java.util.Collection;

import org.apache.commons.collections.CollectionUtils;
import org.eclipse.uml2.uml.*;

import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.VardeclsFIS;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.*;

public class DesignCheckType extends DesignVisitor<Boolean,Type,Void,GetTypeException, 
TypeGenVisitor<Void, GetTypeException>, ExpGetType> {

	final private Type t_ret;
	public boolean query = true; // TODO: Should be in state
	protected Collection<String> outParams = new ArrayList<String>(); // TODO: should be in state; type...
	private final Operation op;

	private DesignCheckType(ExpGetType ev, Operation op) throws GetTypeException{
		super(ev);
		this.op = op;
		state = Boolean.TRUE;
		t_ret = ev.getReturnType(op);
	}

	public DesignCheckType(Operation op) throws GetTypeException{
		this(new ExpGetType(op), op);
	}
	
	public DesignCheckType(Operation op, VardeclsFIS vds) throws GetTypeException {
		this(new ExpGetType((Classifier)op.getOwner(),vds), op);
	}

	protected void visit(MCALL m) throws GetTypeException {
		ExpGetType ev2;
		(ev2 = new ExpGetType(ev) {
			@Override public boolean voidOK() { return true; }
		}).run(m);
		outParams.addAll(ev2.outParams);
		// TODO: Could be cached from previous run.
		Operation op = ev2.getOperationFromMcall(m);
		// Shouldn't happen if the run above succeeded:
		assert op != null;

		if (op.isQuery() && m.getOut().isEmpty())
			throw new GetTypeException("Don't call a pure function as a method!",m);
		/* A method is a query, if the OUTPUT ARGUMENTS are only local variables
		 * XXX That's of course a big simplification. RCOSTOOL-71
		 * TODO: Handle m(;this) :-)
		 */
		if (op.isQuery())
			for (REF p : m.getOut()) {
				Collection<DASHABLE> vars = getVariables(p);
				for (DASHABLE d : vars) {
					if (! (d instanceof VAR || d instanceof PAR))
						throw new ExpGetType.UnpureException("You can only pass local variables to pure methods, but not: "+ev.expToString(d),m);
				}
			}

		set(Boolean.TRUE);
	}

	protected void visit(NEW m) throws GetTypeException {
		new ExpGetType(ev) {
			@Override public boolean voidOK() { return true; }
		}.run(m);
		set(Boolean.TRUE);
	}

	@Override
	protected void visit(ASSIGN a) throws GetTypeException {
		final REF aVal = a.getVal();

		if (!DesignUtility.getPrimedVariables(a).isEmpty())
			throw new GetTypeException("Can't use primed variables in assignments", a);
		if (a.getVal() instanceof PAR)
			outParams.add(((PAR)a.getVal()).getName().getName());

		Type t1 = run(aVal);
		query &= a.getVal() instanceof VAR || a.getVal() instanceof PAR;

		if (t1 == null)
			throw new GetTypeException("Can't use 'null' on the left-hand side of an assignment",aVal);
		ev.set(null);
		Type t2 = run(a.getE());
		if (t1 == ev.prim.getVoid()) throw new GetTypeException("The expression " + ev.expToString(aVal)+ " is of type void " +
				" and so cannot be used in an assignment", aVal);
		if (t2 == ev.prim.getVoid()) throw new  GetTypeException("The expression " + ev.expToString(a.getE())+ " is of type void " +
				" and so cannot be used in an assignment", a.getE());
		// TODO: check subtyping (the two types don't have to be equal, t1 can be a super type of t2) 
		if (! ev.prim.isSubType(t2,t1)) throw new GetTypeException("Both sides of an assignment should have the same type:\nleft: "+t1.getQualifiedName()+", right: "+t2.getQualifiedName(), a);
	}

	@SuppressWarnings("unchecked")
	@Override
	protected void visit(IF i) throws GetTypeException {
		Type t1 = run(i.getB());
		if (t1 != ev.prim.getBoolean()) throw new GetTypeException("The type of the expression \"" + ev.expToString(i.getB()) + 
				"\" is not boolean", i.getB());
		pushDeclFrame();
		Collection<String> oldOut = outParams;
		outParams = new ArrayList<String>();
		visit(i.getLeft());
		Collection<String> outLeft = outParams;
		popDeclFrame();

		pushDeclFrame();
		outParams = new ArrayList<String>();
		visit(i.getRight());
		Collection<String> outRight = ev.outParams;
		popDeclFrame();
		
		// intersection of outParams:
		oldOut.addAll(CollectionUtils.intersection(outLeft,outRight));
		outParams = oldOut;
	}

	private Type run(EXP b) throws GetTypeException {
		ev = new ExpGetType(ev);
		ev.queryOps = new ArrayList<Operation>();
		ev.outParams = new ArrayList<String>();
		Type t = ev.run(b);
		query &= ev.queryOps.isEmpty();
		return t;
	}

	@Override
	protected void visit(PP d) throws GetTypeException {
		visit(d.getPre());
		// check for purity.
		// TODO: isPure() should be removed eventually once we are sure that queryOps works as expected.
		if (! ev.queryOps.isEmpty()) {
			isPure(d.getPre(), ev.currentClass);
			throw new GetTypeException(ev.queryOps.toString(),d);
		} else {
			isPure(d.getPre(), ev.currentClass);
			if (! ev.queryOps.isEmpty())
				throw new GetTypeException(ev.queryOps.toString(),d);
		}

		if (d.getPost() instanceof MCALL) {
			ExpGetType ev2;
			(ev2 = new ExpGetType(ev) {
				@Override public boolean voidOK() { return true; }
			}).run((MCALL) d.getPost());
			query &= ev2.queryOps.isEmpty();
		} else {
			visit(d.getPost());


			// A post-condition that is not an MCALL must not have side-effects:
			if (!ev.queryOps.isEmpty())
				throw new GetTypeException("Unpure method called: "+ev.queryOps.toString(), d);
			for (DASHABLE da : getPrimedVariables(d.getPost())) {
				query &= da instanceof VAR || da instanceof PAR; // TODO: Review
			}
		}
		// end
		outParams.addAll(ev.outParams); // relies on visit(d.getPost()) having been done.

		if (hasPrimedVariables(d.getPre()))
			throw new GetTypeException("Primed variables in pre-condition are not allowed", d);

	}

	// XXX Yikes.
	@Override
	protected void visit(EXP e) throws GetTypeException {
		run(e); // Hopefully...
	}

	protected void visit(RET r) throws GetTypeException {
		if (!DesignUtility.getPrimedVariables(r).isEmpty())
			throw new GetTypeException("Can't use primed variables in return statement", r);
		if (r.getE() != null) {

			/* If there's one result parameter, mark it as used.
			 * Only look at the first not-input parameter.
			 */
			for (Parameter p : op.getOwnedParameters()) {
				if (p.getDirection() != ParameterDirectionKind.IN_LITERAL)
					{ outParams.add(p.getName()); break; }
			}

			Type t_exp = run(r.getE());
			if (! ev.prim.isSubType(t_exp,t_ret)) throw new GetTypeException("The argument of the return statement is not a subtype of the return type", r);
		} else
		{ /* NADA: void return, simply leave execution. */ }
	}

	@Override
	protected void visit(VARDECL r) throws GetTypeException {
		ev.checkVarname(r.getVar());
		// check the type:
		ev.run(r.getType());
		ev.vdecl.new Vardecl(r);
	}

	@Override
	protected void visit(ENDDECL r) throws GetTypeException {
		ev.vdecl.remove(r.getVar());
	}

	@Override
	protected void visit(FOR d) throws GetTypeException {
		if (!DesignUtility.getPrimedVariables(d.getVal()).isEmpty())
			throw new GetTypeException("Can't use primed variables in statement.", d);
		ev.checkVarname(d.getVar());
		TYPE var = d.getType();
		ev.run(var);
		Type tv = ev.checkType(var);
		VAL val = d.getVal();
		ev.checkVarAgainstSet(d, tv, val);
		pushDeclFrame();
		ev.vdecl.new Vardecl(d);
		visit(val);
		visit(d.getDo());
		popDeclFrame();
	}

	private void pushDeclFrame() {
		ev.vdecl.pushDeclFrame();
	}

	private void popDeclFrame() {
		ev.vdecl.popDeclFrame();
	}

	@Override
	protected void visit(CHOICE d) throws GetTypeException {
		pushDeclFrame();
		Collection<String> oldOut = outParams;
		outParams = new ArrayList<String>();
		visit(d.getLeft());
		Collection<String> outLeft = outParams;
		popDeclFrame();

		pushDeclFrame();
		outParams = new ArrayList<String>();
		visit(d.getRight());
		Collection<String> outRight = outParams;
		popDeclFrame();
		
		// intersection of outParams:
		oldOut.addAll(CollectionUtils.intersection(outLeft,outRight));
		outParams = oldOut;
	}

	@Override
	protected void visit(WHILE d) throws GetTypeException {
		if (!DesignUtility.getPrimedVariables(d.getB()).isEmpty())
			throw new GetTypeException("Can't use dashed variables in statement", d);
		visit(d.getB());
		pushDeclFrame();
		visit(d.getDo());
		popDeclFrame();
	}

	@Override
	protected void visit(CHOKE d) throws GetTypeException {
		// NADA
	}

	public Collection<String> getOutParams() {
		return outParams;
	}
}