package edu.unu.iist.rcos.spec.antlr.visitor.check;

import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.DesignBaseVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.VisitorException;

public class DesignEqual extends DesignBaseVisitor<Boolean,Boolean,Boolean, VisitorException, TypeEqual, ExpEqual> {

	protected DESIGN des;
	final boolean strong;
	
	/**
	 * Compares this design with another. If 'strong' equality is selected,
	 * dashes are significant.
	 */
	public DesignEqual(DESIGN d, boolean strong){
		if (d == null)
			throw new IllegalArgumentException();
		des = d;
		state = false;
		this.strong = strong;
	}
	
	protected DESIGN getDesign(){
		return des;
	}

	/**
	 * Override the static method of DesignUtility, to pick the correct one w.r.t. strong
	 * @author Charles Morisset
	 * @param e1
	 * @param e2
	 * @return 
	 */
	private boolean isExpEqual(EXP e1, EXP e2) {
		assert e1 != null;
		assert e2 != null;
		if (strong)
			return DesignUtility.isExpEqual(e1, e2);
		else 
			return DesignUtility.isExpWeakEqual(e1, e2);
	}
	
	private boolean isDesignEqual(DESIGN d1, DESIGN d2) throws VisitorException {
		DESIGN oldDes = des;
		des = d1;
		visit(d2);
		des = oldDes;
		return state;
	}
	
	@Override
	protected void visit(PP d) throws VisitorException {
		DESIGN d2 = getDesign();
		if (d2 instanceof PP){
			PP pp2 = (PP) d2;
			boolean pre = isExpEqual(d.getPre(), pp2.getPre());
			boolean post = isExpEqual(d.getPost(), pp2.getPost());
			set(pre && post);
		} else set(false);
	}

	@Override
	protected void visit(IF d) throws VisitorException {
		DESIGN d2 = getDesign();
		if (d2 instanceof IF){
			IF if2 = (IF) d2;
			set(isExpEqual(if2.getB(), d.getB()) && 
				isDesignEqual(if2.getLeft(), d.getLeft()) &&
				isDesignEqual(if2.getRight(), d.getRight()));
		} else set(false);
	}

	@Override
	protected void visit(CHOICE d) throws VisitorException {
		DESIGN d2 = getDesign();
		if (d2 instanceof CHOICE){
			CHOICE c2 = (CHOICE) d2;
			set(isDesignEqual(c2.getLeft(), d.getLeft()) &&
				isDesignEqual(c2.getRight(), d.getRight()));
		} else set(false);
	}

	@Override
	protected void visit(SEQ d) throws VisitorException {
		DESIGN d2 = getDesign();
		if (d2 instanceof SEQ){
			SEQ s2 = (SEQ) d2;
			set(isDesignEqual(s2.getLeft(), d.getLeft()) &&
				isDesignEqual(s2.getRight(), d.getRight()));
		} else set(false);
	}

	@Override
	protected void visit(WHILE d) throws VisitorException {
		DESIGN d2 = getDesign();
		if (d2 instanceof WHILE){
			WHILE w2 = (WHILE) d2;
			set(isExpEqual(w2.getB(), d.getB()) && 
				isDesignEqual(w2.getDo(), d.getDo()));
		} else set(false);
	}

	@Override
	protected void visit(FOR d) throws VisitorException {
		DESIGN d2 = getDesign();
		if (d2 instanceof FOR){
			FOR f2 = (FOR) d2;
			set(DesignUtility.isTypeEqual(f2.getType(), d.getType()) &&
				f2.getVar().getName().equals(d.getVar().getName()) && 
				isDesignEqual(f2.getDo(), d.getDo()));
		} else set(false);
	}

	@Override
	protected void visit(ASSIGN a) throws VisitorException {
		DESIGN d2 = getDesign();
		if (d2 instanceof ASSIGN){
			ASSIGN a2 = (ASSIGN) d2; 
			set( isExpEqual(a2.getE(), a.getE()) && isExpEqual(a2.getVal(), a.getVal()));
		} else set(false);
	}

	@Override
	protected void visit(VARDECL d) throws VisitorException {
		DESIGN d2 = getDesign();
		if (d2 instanceof VARDECL){
			VARDECL vd2 = (VARDECL) d2; 
			set(DesignUtility.isTypeEqual(vd2.getType(), d.getType()) &&
				vd2.getVar().getName().equals(d.getVar().getName()));
		} else set(false);
	}

	@Override
	protected void visit(ENDDECL d) throws VisitorException {
		DESIGN d2 = getDesign();
		if (d2 instanceof ENDDECL){
			ENDDECL ed2 = (ENDDECL) d2; 
			set(ed2.getVar().getName().equals(d.getVar().getName()));
		} else set(false);
	}

	@Override
	protected void visit(SKIP d) throws VisitorException {
		set(getDesign() instanceof SKIP);
	}

	@Override
	protected void visit(CHAOS d) throws VisitorException {
		set(getDesign() instanceof CHAOS);
	}
	
	@Override
	protected void visit(MAGIC d) throws VisitorException {
		set(getDesign() instanceof MAGIC);
	}

	@Override
	protected void visit(STOP d) throws VisitorException {
		set(getDesign() instanceof STOP);
	}

	@Override
	protected void visit(RET d) throws VisitorException {
		DESIGN d2 = getDesign();
		if (d2 instanceof RET){
			RET r2 = (RET) d2; 
			set(isExpEqual(r2.getE(), d.getE())); 
		} else set(false);
	}
	
	protected void visit(MCALL m) throws VisitorException {
		DESIGN d2 = getDesign();
		if (d2 instanceof MCALL){
			MCALL m2 = (MCALL) d2;
			set(isExpEqual(m2,m));
		} else set(false);
	}

	@Override
	protected void visit(CHOKE d) throws VisitorException {
		throw new UnsupportedOperationException();
	}
}
