package edu.unu.iist.rcos.spec.antlr.visitor.check;

import org.eclipse.emf.common.util.EList;

import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.ExpGenVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.VisitorException;

public class ExpEqual extends ExpGenVisitor<Boolean,Boolean, VisitorException, TypeEqual> {

	EXP exp;
	final boolean strong;

	public ExpEqual(EXP e2, boolean s){
		if (e2 == null)
			throw new IllegalArgumentException();
		exp = e2;
		state = false;
		strong = s;
	}

	private EXP getExp(){
		return exp;
	}

	private boolean isExpEqual(EXP e1, EXP e2) throws VisitorException {
		EXP oldExp = exp;
		exp = e1;
		visit(e2);
		exp = oldExp;
		return state;
	}

	@Override
	protected void visit(RELOP e) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof RELOP){
			RELOP re2 = (RELOP) e2;
			if (e.getOp().equals(re2.getOp())){
				set(isExpEqual(e.getLeft(), re2.getLeft()) &&
						isExpEqual(e.getRight(), re2.getRight()));
			} else {
				set(false);
			}
		} else set(false);

	}

	@Override
	protected void visit(LOGOP e) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof LOGOP){
			LOGOP re2 = (LOGOP) e2;
			if (e.getOp().equals(re2.getOp())){
				set(isExpEqual(e.getLeft(), re2.getLeft()) &&
						isExpEqual(e.getRight(), re2.getRight()));
			} else {
				set(false);
			}
		} else set(false);

	}

	@Override
	protected void visit(AOP e) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof AOP){
			AOP re2 = (AOP) e2;
			if (e.getOp().equals(re2.getOp())){
				set(isExpEqual(e.getLeft(), re2.getLeft()) &&
						isExpEqual(e.getRight(), re2.getRight()));
			} else {
				set(false);
			}
		} else set(false);

	}

	@Override
	protected void visit(INSTANCEOF d) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof INSTANCEOF){
			INSTANCEOF iof2 = (INSTANCEOF) e2;
			if (d.getClass_().getName().equals(iof2.getClass_().getName())){
				set(isExpEqual(d.getRef(), iof2.getRef()));
			} else {
				set(false);
			}
		} else set(false);

	}

	@Override
	protected void visit(NEG d) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof NEG){
			set(isExpEqual(d.getE(), ((NEG) e2).getE()));
		} else {
			set(false);
		}

	}

	@Override
	protected void visit(VAR d) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof VAR){
			set(DesignUtility.isTypeEqual(((VAR) e2).getType(), d.getType()) &&
					((VAR) e2).getName().equals(d.getName()) && 
					((! strong) || ((VAR) e2).isDash() == d.isDash()));
		} else {
			set(false);
		}
	}

	@Override
	protected void visit(PAR d) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof PAR){
			set(DesignUtility.isTypeEqual(((PAR) e2).getType(), d.getType()) &&
					((PAR) e2).getName().equals(d.getName()) &&
					((! strong) || ((PAR) e2).isDash() == d.isDash()));
		} else {
			set(false);
		}
	}

	@Override
	protected void visit(FA d) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof FA){
			FA f2 = (FA) e2;
			if (isExpEqual(d.getOwner(), f2.getOwner())){
				set(d.getName().equals(f2.getName()) && 
						((! strong) || ((FA) e2).isDash() == d.isDash()));
			} else {
				set (false);
			}
		} else {
			set(false);
		}

	}

	@Override
	protected void visit(THIS d) throws VisitorException {
		set(getExp() instanceof THIS);
	}

	@Override
	protected void visit(SUPER d) throws VisitorException {
		set(getExp() instanceof SUPER);
	}


	@Override
	protected void visit(NULL d) throws VisitorException {
		set(getExp() instanceof NULL);
	}

	@Override
	protected void visit(NEW d) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof NEW){
			NEW m2 = (NEW) e2;
			if (DesignUtility.isTypeEqual(m2.getName(), d.getName())){
				EList<EXP> in_list1 = d.getIn();
				EList<EXP> in_list2 = m2.getIn();
				if (in_list1.size() == in_list2.size()){
					for (int i = 0; i < in_list1.size(); i++){
						if (! isExpEqual(in_list1.get(i), in_list2.get(i))){
							set(false); return;
						}
					}
				} else { 
					set(false);
					return;
				}

				EList<REF> out_list1 = d.getOut();
				EList<REF> out_list2 = m2.getOut();
				if (out_list1.size() == out_list2.size()){
					for (int i = 0; i < out_list1.size(); i++){
						if (! isExpEqual(out_list1.get(i), out_list2.get(i))){
							set(false); return;
						}
					}
				} else {
					set(false);
					return;
				}
				// If we arrive to this point, that means that all the parameters were equals.
				set(true); return;

			} else set(false);
		} else set (false);
	}


	@Override
	protected void visit(MCALL d) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof MCALL){
			MCALL m2 = (MCALL) e2;
			if (isExpEqual(m2.getOwner(), d.getOwner()) && m2.getName().equals(d.getName())){

				EList<EXP> in_list1 = d.getIn();
				EList<EXP> in_list2 = m2.getIn();
				if (in_list1.size() == in_list2.size()){
					for (int i = 0; i < in_list1.size(); i++){
						if (! isExpEqual(in_list1.get(i), in_list2.get(i))){
							set(false); return;
						}
					}
				} else { 
					set(false);
					return;
				}

				EList<REF> out_list1 = d.getOut();
				EList<REF> out_list2 = m2.getOut();
				if (out_list1.size() == out_list2.size()){
					for (int i = 0; i < out_list1.size(); i++){
						if (! isExpEqual(out_list1.get(i), out_list2.get(i))){
							set(false); return;
						}
					}
				} else {
					set(false);
					return;
				}
				// If we arrive to this point, that means that all the parameters were equals.
				set(true); return;

			} else set(false);
		} else set(false);
	}

	@Override
	protected void visit(CAST d) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof CAST){
			CAST c2 = (CAST) e2;
			if (isExpEqual(c2.getVal(), d.getVal())){
				set(isTypeEqual(c2.getType(), d.getType()));
			} else set(false);
		} else set(false);
	}

	@Override
	protected void visit(CNAME c) throws VisitorException {
		throw new IllegalStateException();
	}

	@Override
	protected void visit(VARNAME var) throws VisitorException {
		throw new IllegalStateException();
	}

	@Override
	protected void visit(BASEBOOL d) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof BASEBOOL) {
			set(((BASEBOOL) e2).isBVal() == d.isBVal());
		} else set(false);
	}

	@Override
	protected void visit(BASEINT d) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof BASEINT) {
			set(((BASEINT) e2).getVal().equals(d.getVal()));
		} else set(false);
	}

	@Override
	protected void visit(BASESTRING d) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof BASESTRING) {
			set(((BASESTRING) e2).getVal().equals(d.getVal()));
		} else set(false);
	}

	@Override
	protected void visit(BASEDOUBLE d) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof BASEDOUBLE) {
			set(((BASEDOUBLE) e2).getVal().equals(d.getVal()));
		} else set(false);
	}

	@Override
	protected void visit(BASEENUM d) throws VisitorException {
		// XXX Does Evidal actually exist?
		EXP e2 = getExp();
		if (e2 instanceof BASEENUM) {
			set(((BASEENUM)e2).getVal().equals(d.getVal()));
		} else
			set(false);
	}

	@Override
	protected void visit(PREDICATE d) throws VisitorException {
		if (d instanceof ITERATOR)
			visit((ITERATOR) d);
		else {
			EXP e2 = getExp();
			if ((d instanceof FORALL && e2 instanceof FORALL) || 
					(d instanceof EXIST && e2 instanceof EXIST)){
				PREDICATE p2 = (PREDICATE) e2;
				set (isExpEqual(p2.getE(), d.getE()) && isTypeEqual(p2.getType(), d.getType()) &&
						p2.getVar().equals(d.getVar()));
			} else set(false);
		}
	}

	@Override
	protected void visit(ITERATOR d) throws VisitorException {
		EXP e2 = getExp();
		if (e2 instanceof ITERATOR) {
			ITERATOR p2 = (ITERATOR) e2;
			set (isExpEqual(p2.getE(), d.getE()) && isTypeEqual(p2.getType(), d.getType()) &&
					p2.getVar().equals(d.getVar()) && isExpEqual(p2.getRange(),d.getRange()));
		} else set(false);
	}
}
