package edu.unu.iist.rcos.spec.antlr.visitor.check;

import java.util.ArrayList;
import java.util.Collection;

import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.TypeHandler;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.ExpVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.TypeVisitor;

/**
 * @author charles
 * Based on ExpGetType to return all the Operation corresponding to a method call
 */
public class ExpGetAllOperations extends ExpVisitor<Collection<Operation>,Void, GetTypeException, TypeVisitor<Void, GetTypeException>> {

	final Classifier currentClass;
	final TypeHandler th;
	final protected boolean voidOk;
	
	public ExpGetAllOperations(Classifier cl){
		this(cl, false);
	}

	public ExpGetAllOperations(Classifier cl, boolean voidOk){
		this(cl, new ArrayList<Operation>(), voidOk);
	}

	protected ExpGetAllOperations(Classifier cl, Collection<Operation> col, boolean voidOk){
		super(new TypeVisitor<Void, GetTypeException>());
		if (cl == null)
			throw new IllegalArgumentException();
		if (col == null)
			throw new IllegalArgumentException();
		state = col;
		currentClass = cl;
		th = new TypeHandler(cl);
		this.voidOk = voidOk;
	}

	/**
	 * @author Charles Morisset
	 * @param m
	 * @throws TypingException
	 */
	protected void visit(MCALL m) throws GetTypeException {
		ExpGetType egt = new ExpGetType(currentClass, th) {
			@Override public boolean voidOK() { return voidOk; }
		};
		PREFIX p = m.getOwner();
		if (! (p instanceof CNAME)) 
			visit(m.getOwner());	
		final Operation op = egt.getOperationFromMcall(m);
		assert op != null : "Couldn't find "+DesignUtility.expToString(currentClass, m) + ", voidOk: "+voidOk;
		state.add(op);
	}

	@Override protected void visit(CNAME c) throws GetTypeException {}
	@Override protected void visit(VARNAME var) throws GetTypeException {}
	@Override protected void visit(BASEBOOL d) throws GetTypeException {}
	@Override protected void visit(BASEINT d) throws GetTypeException {}
	@Override protected void visit(BASESTRING d) throws GetTypeException {}
	@Override protected void visit(BASEDOUBLE d) throws GetTypeException {}
	@Override protected void visit(BASEENUM d) throws GetTypeException {}
}
