package edu.unu.iist.rcos.spec.antlr.visitor.check;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.uml2.uml.*;
import org.eclipse.uml2.uml.Class;

import edu.unu.iist.rcos.mm.ocl.Class_OCL;
import edu.unu.iist.rcos.mm.util.PrimedVariableForbiddenException;
import edu.unu.iist.rcos.mm.util.TypingException;
import edu.unu.iist.rcos.mm.util.Utility;
import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.profile.ContractInterface;
import edu.unu.iist.rcos.profile.PrimitiveTypes;
import edu.unu.iist.rcos.profile.Stereotypes;
import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.MyRCOSModel;
import edu.unu.iist.rcos.spec.antlr.RCOSModelCheckException;
import edu.unu.iist.rcos.spec.antlr.TypeHandler;
import edu.unu.iist.rcos.spec.antlr.Vardecls;
import edu.unu.iist.rcos.spec.antlr.VardeclsFIS;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.Vardecls.Vardecl;
import edu.unu.iist.rcos.spec.antlr.visitor.ExpVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.TypeGenVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.TypeVisitor;

// TODO: Should be an ExpGenVisitor, but that requires subsequent changes :-/
public class ExpGetType extends ExpVisitor<Type,Void, GetTypeException, TypeGenVisitor<Void, GetTypeException>>
implements TypeHandler.VoidOK {

	final Classifier currentClass;
	protected final PrimitiveTypes prim;
	protected final TypeHandler th;
	protected final Component co;
	public final Vardecls vdecl;
	protected Collection<Operation> queryOps; // TODO: should be final.
	protected Collection<String> outParams; // TODO: should be final, in state; type...

	public boolean voidOK() { return false; }; // If we are coming from the design visitor...Evil hack. 

	public ExpGetType(final Classifier cl, final TypeHandler th) throws GetTypeException {
		this(cl, th, new Vardecls() {

			@Override
			public Type getType(TYPE t) throws GetTypeException {
				return th.getTypeFromTYPE(t, cl);
			}});
	}

	/**
	 * Check a TYPE from the AST against the current model.
	 * Never returns null.
	 * Static because used from a constructor below.
	 * @throws GetTypeException if TYPE is not defined.
	 * @see #getTypeFromTYPE(TYPE, Classifier)
	 */
	private static Type checkType(TypeHandler th, TYPE type, Classifier cl) throws GetTypeException {
		Type t = th.getTypeFromTYPE(type, cl);
		if (t == null)
			throw new GetTypeException("Type '"+DesignUtility.typeToString(type)+"' is not defined.", type);
		return t;
	}

	/**
	 * Proxy.
	 * @see TypeHandler#checkType(TYPE, Classifier)
	 */
	public Type checkType(TYPE type) throws GetTypeException {
		return checkType(th,type,currentClass);
	}

	protected ExpGetType(ExpGetType egt) throws GetTypeException {
		this(egt.co, egt.currentClass, egt.th, egt.vdecl);
		queryOps = egt.queryOps;
		outParams = egt.outParams;
	}

	/**
	 * @deprecated Use {@link #ExpGetType(Component, Classifier, TypeHandler, Vardecls)} instead,
	 * otherwise the code will try to infer the component.
	 */
	protected ExpGetType(final Classifier cl, final TypeHandler th, Vardecls vdecl) throws GetTypeException {
		this(getScope(cl),cl,th,vdecl);
	}

	public ExpGetType(VardeclsFIS vdecl) throws GetTypeException {
		this(vdecl.getClassifier(),vdecl.th,vdecl);
	}

	public ExpGetType(final Component co, final Classifier cl, final TypeHandler th, Vardecls vdecl) {
		super(new TypeVisitor<Void, GetTypeException>() {  // TODO: fishy! Why do we need the tv?

			@Override
			protected void visit(GTYPE type) throws GetTypeException {
				super.visit(type);
				ExpGetType.checkType(th,type,cl);
			}

		});
		if (cl == null)
			throw new IllegalArgumentException();
		state = null;
		currentClass = cl;
		this.th = th;
		this.prim = th; // TODO
		this.vdecl = vdecl;
		this.co = co;
		// XXX Should be part of state?
		this.queryOps = new ArrayList<Operation>();
		this.outParams = new ArrayList<String>();
	}

	@Deprecated
	public ExpGetType(Classifier cl) throws GetTypeException{
		this(cl, new TypeHandler(cl), new VardeclsFIS(cl));
	}
	
	@Deprecated
	public ExpGetType(Operation op) throws GetTypeException{
		this((Classifier) op.getOwner(), new VardeclsFIS(op));
	}
	
	protected ExpGetType(Classifier cl, VardeclsFIS vs) throws GetTypeException {
		this(cl, vs.th, vs);
	}

	public boolean isBooleanType() throws GetTypeException {
		return (getResult() == prim.getBoolean());
	}

	@Override
	protected void visit(BASEBOOL c){
		set(prim.getBoolean());
	}

	@Override
	protected void visit(BASESTRING c){
		set(prim.getString());
	}

	@Override
	protected void visit(BASEDOUBLE c){
		set(prim.getDouble());
	}

	@Override
	protected void visit(BASEINT c){
		set(prim.getInt());
	}
	
	@Override
	protected void visit(BASEENUM c) throws GetTypeException{
		// Look in all Enums for the literal:
		EnumerationLiteral lit = th.findEnumLiteral(c.getVal());
		if (lit == null)
			throw new GetTypeException("Enumeration literal not found: "+c.getVal(), c);
		set((Enumeration)lit.getOwner());
	}

	@Override
	protected void visit(NULL n) {
		set(null); // ?
	}

	protected void visit(THIS d){
		set(currentClass);
	}

	@Override
	protected void visit(SUPER s) throws GetTypeException {
		if (currentClass.getGenerals().isEmpty())
			throw new GetTypeException("No super class to call super on!", s);
		if (currentClass.getGenerals().size() > 1)
			throw new GetTypeException("Can't call super on a class with more than one general!", s);
		set(currentClass.getGenerals().get(0));
	}

	@Override
	protected void visit(VAR v) throws GetTypeException {
		Type t = vdecl.getType(v.getType());
		if (t != null)
			set(t);
		else 
			throw new GetTypeException("The variable " + v.getName().getName() + " is declared as of type " 
					+ typeToString(v.getType()) +
					", which does not exist", v);
	}

	protected void visit(PAR v) throws GetTypeException{
		Type t = vdecl.getType(v.getType());
		if (t != null)
			set(t);
		else 
			throw new GetTypeException("The parameter " + v.getName().getName() + " is declared as of type " + 
					typeToString(v.getType()) + ", which does not exist", v);
		if (v.isDash())
			outParams.add(v.getName().getName());
	}

	protected void visit(FA f) throws GetTypeException{
		PREFIX pr = f.getOwner();

		final Type t;
		if (pr instanceof CNAME) {
			t = getRequiredInterface(f, (CNAME)pr);
		} else 
			t = run((REF) pr);
		final String fieldname = f.getName().getName();

		//XXX Void type
		if (t == prim.getVoid()) throw new GetTypeException("The expression \"" + expToString(f.getOwner()) 
				+ "\" is of type void and so does not have any field \"" + fieldname + "\"", f);
		if (PrimitiveTypes.isRCOSPrimitive(t)) throw new GetTypeException("The expression \"" + expToString(f.getOwner()) 
				+ "\" is of type " + t.getName() + " and so does not have any field \"" + fieldname + "\"", f);

		EList<Property> attrs = new BasicEList<Property>();
		EList<Association> assocs = new BasicEList<Association>();
		if (t instanceof Classifier) {
			Classifier ct = (Classifier) t;
			attrs.addAll(ct.getAllAttributes());
			assocs.addAll(ct.getAssociations());
			for (Classifier cg : ct.getGenerals()) {
				assocs.addAll(cg.getAssociations());
			}
		} else {
			throw new IllegalArgumentException("Don't know how to handle "+t.getQualifiedName()+" ["+t.eClass().getName()+"]");
		}

		for (Association a : assocs){
			// XXX Only works for binary associations
			if (a.getMembers().size() != 2) throw new GetTypeException("The association " + a.getQualifiedName() + " is not binary",f);
			Property p1 = a.getMemberEnds().get(0);
			Property p2 = a.getMemberEnds().get(1);
			if (prim.isSubType(t, p1.getType()) && a.getNavigableOwnedEnds().contains(p2)) attrs.add(p2);
			if (prim.isSubType(t, p2.getType()) && a.getNavigableOwnedEnds().contains(p1)) attrs.add(p1);
		}

		for (Property p : attrs){
			if (p.getName().equals(fieldname)){
				// If it's a simple multiplicity
				// XXX Need to consider the case 0..1
				if (p.getUpper() == 1)	
					set(p.getType());
				else {
					Type t2 = th.findSetType(p.getType(),co);
					if (t2 == null) {
						// No implementation for sets of this class
						throw new GetTypeException("Class Set<" + p.getType().getName() + "> not found",f);
					} else {
						set(t2);
					}
				}
				return;
			}
		}
		// Field not found
		throw new GetTypeException("In the expression \"" + expToString(f) + "\", the sub-expression  \"" + expToString(f.getOwner()) + 
				"\" is of type " + t.getQualifiedName() + " which does not have an attribute \"" + fieldname + "\" (or the association is not navigable)", f);
	}

	protected Type getTypeFromString(CNAME n) throws GetTypeException {
		return th.getTypeFromString(n, null, currentClass);
	}

	/**
	 * @author Charles Morisset
	 * @param m
	 * @throws TypingException
	 */
	protected void visit(MCALL m) throws GetTypeException{
		// We do not permit dashed variables in input positions:
		for (EXP p : m.getIn()) {
			if (p instanceof DASHABLE && (((DASHABLE)p).isDash()))
				throw new GetTypeException(new PrimedVariableForbiddenException("Input argument must not be dashed!"));
		}

		for (EXP o : m.getOut()) {
			if (o instanceof DASHABLE && (((DASHABLE)o).isDash()))
				throw new GetTypeException(new PrimedVariableForbiddenException("Output argument must not be dashed!"));
			if (o instanceof PAR) {
				PAR par = (PAR) o;
				outParams.add(par.getName().getName());
			}
		}
		// No dashes in owner:
		Collection<DASHABLE> dvars = DesignUtility.getPrimedVariables(m.getOwner());
		if (dvars.size() > 0)
			throw new GetTypeException(new PrimedVariableForbiddenException("Can't use dashed variables in target!"));

		// If we came here through a design, we need a different visitor for void-handling:
		ExpGetType ev = new ExpGetType(this);
		final Type t = ev.getOwnerType(m);

		final String mName = m.getName().getName();
		if (t == prim.getVoid()) throw new GetTypeException("The expression \"" + expToString(m.getOwner()) 
				+ "\" is of type void and so does not have any method \"" + mName + "\"", m);
		if (PrimitiveTypes.isRCOSPrimitive(t)) throw new GetTypeException("The expression \"" + expToString(m.getOwner()) 
				+ "\" is of type " + t.getName() + " and so does not have any method \"" + mName + "\"", m);

		List<Type> paramList = ev.new MkInParams().mkInParams(m);

		/* Only look for void methods if there are no output parameters */
		final Operation operation = getOperation(mName, paramList, (Classifier) t, m.getOwner() instanceof SUPER, voidOK() && m.getOut().isEmpty());
		if (operation != null) {
			if (! operation.isQuery())
				queryOps.add(operation);
			else
			/* XXX: The check for purity is too strong. What we are
			 * doing here is actually abusing queryOps: m() maybe pure,
			 * but the call with actual parameters is not correct. Maybe it
			 * would make sense to have a field indicating what we are looking
			 * for just like in DesignCheckType? 
			 */
				for (REF p : m.getOut()) {
					if (!(p instanceof VAR || p instanceof PAR) && (p instanceof DASHABLE && ((DASHABLE) p).isDash())) {
						queryOps.add(operation); break;
					}
				}

			Type t_ret = getReturnType((Classifier) t,operation);
			assert t_ret != null;
			set(t_ret);
		} else {
			String list_type = "";
			for (Type t1 : paramList){
				list_type += t1.getName() + ",";
			}

			// We remove the last comma
			if (list_type.length() > 0 ) list_type = list_type.substring(0, list_type.length() - 1);

			final String queryMethod; 
			String msgSuffix = "";
			if (!(voidOK() && m.getOut().isEmpty())) {
				queryMethod = "query";
				final Operation operation2 = getOperation(mName, paramList, (Classifier) t, m.getOwner() instanceof SUPER, true);
				if (operation2 != null)
					msgSuffix=" (but there is a method with that name NOT marked as a query)";
			} else
				queryMethod = "method";
			//XXX: The error message may be wrong, because it may have such a method, but of type void
			throw new GetTypeException("The expression  \"" + expToString(m.getOwner()) + 
					"\" is of type " + t.getQualifiedName() + " which does not have a "+queryMethod+" \"" + mName + "(" + list_type + ")\""+msgSuffix, m);
		}
	}

	public Type getReturnType(Operation op) throws GetTypeException {
		return getReturnType(currentClass, op);
	}

	private Type getReturnType(Classifier t, Operation op) throws GetTypeException {
		List<Parameter> para_out = Utility.getOwnedOutParameters(op); 
		int size = para_out.size();
		if (size > 0) {
			final Parameter p = para_out.get(size-1);
			Type pT;
			if (p.getType().getOwner() instanceof ClassifierTemplateParameter) {
				pT = getActualType(t, p.getType());
				if (pT == null) { // We are checking a generic definition, so the type parameter is unbound.
					pT = p.getType();
				}
			}
			else
				pT = p.getType();
			
			if (p.getUpper() == -1 && p.getLower() == 0) {
				Type t2 = th.findSetType(pT, co);
				assert t2 != null;
				return t2;
			} else {
				assert p.getUpper() == 1 && p.getLower() == 1;
				return pT;
			}
		} else {
			// The method is of void type
			return th.getVoid();
		}

	}

	/**
	 * A helper to map the list of actual parameters into types.
	 * We use this to lookup the correct operation---we may have operations 
	 * with the same method name, but different signature.
	 * Unfortunately, this approach doesn't work so well for 'null', which doesn't have
	 * a 'proper' type by itself. In this case, we would have to first lookup the method by name,
	 * and then match the parameters.  Since we know that the caller doesn't expect this YET,
	 * we bail out with an exception here. FIXME: RCOSTOOL-230 
	 */
	class MkInParams {
		public List<Type> mkInParams(MCALL m) throws GetTypeException {
			return mkInParams(m.getIn(),m);
		}

		public List<Type> mkInParams(NEW m) throws GetTypeException {
			return mkInParams(m.getIn(),m);
		}

		private List<Type> mkInParams(EList<EXP> in, TOKEN t) throws GetTypeException {
			List<Type> paramList = new ArrayList<Type>(in.size());

			for (EXP e : in){
				assert e != null;
				final Type p = run(e);
				/* FIXME 'p' can be null for the literal 'null'! */
				if (p == null)
					throw new GetTypeException("Please cast 'null' to the desired type, sorry.", t);
				paramList.add(p);
			}
			return paramList;
		}
	}

	private Type getOwnerType(MCALL m) throws GetTypeException {
		final Type t;
		PREFIX p = m.getOwner();
		if (p instanceof CNAME) {
			t = getRequiredInterface(m, (CNAME) p);
		} else 
			t = run((REF) p);
		return t;
	}

	public Type getRequiredInterface(TOKEN t, CNAME p) throws GetTypeException {
		final Interface i = getInterfaceFromString(p.getName(), t);
		if (i == null) throw new GetTypeException("Required interface \"" + p.getName() + "\" not found ", t);
		ContractInterface ci = Stereotypes.isContractInterface(i);
		if (ci != null)
			return ci.getInterface();
		else
			return i; /* ? */
	}

	/**
	 * @author Charles Morisset
	 * @param name the name of the required contract  
	 * @return the Interface of the required contract
	 * @see MyRCOSModel#addProvidedInterface
	 * @throws RCOSModelCheckException 
	 */
	public Interface getInterfaceFromString(String name, TOKEN t) throws GetTypeException {
		if (co == null)
			throw new GetTypeException("For "+ this.currentClass.getQualifiedName()+": no component context found given for type checking (while looking for '"+name+"'." +
					" Do you need to add any \"implements\" declarations to the class and mark it as private? )!", t);
		try {
			for (ContractInterface ci : Utility.getRequiredContracts(co)) {
				if (ci.getBase_Interface().getName().equals(name))
					return ci.getInterface();
			}
		} catch (WellformedException e) {
			throw new GetTypeException("The required interfaces of " + co.getLabel() + " are not well-formed", e);
		}
		throw new GetTypeException("The component " + co.getLabel() + " does not require an interface " + 
				name, t);
	}

	/**
	 * @author Charles Morisset
	 * @param cl any classifier
	 * @param t_ret any type
	 * @return if cl is a binding of A<X> with X->P and t_ret = X then P, else null
	 */
	private static Type getActualType(Classifier cl, Type t_ret) {
		assert cl != null;
		assert t_ret != null;
		TemplateBinding tb = Utility.getTemplateBinding(cl);
		// If not template binding, then exit
		if (tb == null) return null;
		for (TemplateParameterSubstitution tp : tb.getParameterSubstitutions()) 
			if (tp.getFormal().getParameteredElement() == t_ret)
				return (Type) Utility.getActuals(tp);								
		return null;
	}

	protected void visit(NEW n) throws GetTypeException {
		// XXX [cm] It is not clear what is the owner of an IFCALL
		// I probably have to look through this and change code
		Type t = checkType(n.getName());
		String cname = n.getName().getName().getName();

		// Basically, the type of the new is the type of the class (cname)
		// We just need to check that there exists a corresponding constructor

		List<Type> paramList = new MkInParams().mkInParams(n);
		Operation op = getOperation(cname, paramList, (Class) t, false, true);

		if (op != null) {
			// If no exception, then we found the constructor
			set(t);
		} else {
			// For pretty-print:
			String list_type = "";
			for (Type t1 : paramList)
				list_type += t1.getName() + ",";

			// We remove the last comma
			if (list_type.length() > 0 ) list_type = list_type.substring(0, list_type.length() - 1);

			throw new GetTypeException("The class " + cname + " does not have a constructor \"" + cname + "(" + list_type + ")\"", n);
		}
	}

	/**
	 * @author Charles Morisset
	 * @param name the name of the operation 
	 * @param typeList
	 * @param cl
	 * @return The operation, or null.
	 * @throws GetTypeException 
	 */
	private Operation getOperation(String name, List<Type> typeList, Classifier cl, boolean onlyGenerals, boolean voidOk) throws GetTypeException {
		assert name != null;
		assert cl != null;
		TemplateBinding tb = Utility.getTemplateBinding(cl);
		for (Operation op : cl.getAllOperations()){
			if (op.getName().equals(name) && (! (onlyGenerals && cl.getOperations().contains(op)))) {
				EList<Parameter> paramList = Utility.getOwnedInParameters(op);
				if (paramList.size() == typeList.size()){ // short cut
					boolean ok = true;
					for (int i = 0; i < paramList.size(); i++){
						Type t = paramList.get(i).getType();
						if (tb != null) {
							for (TemplateParameterSubstitution tp : tb.getParameterSubstitutions()) {
								if (tp.getFormal().getParameteredElement() == t) {
									t = (Type) Utility.getActuals(tp);
								}
							}
						} 
						if (! prim.isSubType(typeList.get(i), t))
							ok = false; // TODO: break?
					}
					// If this is a function, we don't allow it in designs
					final Type returnType = getReturnType(cl, op);
					assert returnType != null;
					if (voidOk)
						ok = returnType == prim.getVoid();
					else
						ok = returnType != prim.getVoid();
					if (ok) {
						return op;
					}
				}
			}
		}
		
		/* If we were preferring a method (voidOk) but couldn't find any, let's see
		 * if we can find a boolean(!) function since we might(!) be in a postcondition.
		 * Sheesh, this needs cleaning up. XXX
		 */
		if (voidOk) {
			Operation op = getOperation(name, typeList, cl, onlyGenerals, false);
			if (op != null && getReturnType(cl,op) == prim.getBoolean())
				return op;
			else
				return null;
		} else
			return null;
	}

	/**
	 * @author Charles Morisset
	 */
	public Operation getOperationFromMcall(MCALL m) throws GetTypeException {
		// getType of the owner
		Type tOwner = getOwnerType(m);
		assert tOwner != null;
		final ExpGetType ev = new ExpGetType(this);
		List<Type> paramList = ev.new MkInParams().mkInParams(m);

		Operation op = getOperation(m.getName().getName(), paramList, (Classifier) tOwner, m.getOwner() instanceof SUPER, voidOK() && m.getOut().isEmpty());
		/* Sanity check. XXX Review! */
		if (op != null && op.getOwnedParameters().size() != m.getIn().size() + m.getOut().size() + (m.getOut().isEmpty() && !voidOK() ? 1 : 0))
			throw new GetTypeException("Mismatched number of args on method call to "+op.getQualifiedName(), m);
		return op;
	}	

	protected void visit(RELOP r) throws GetTypeException{
		Type left = run(r.getLeft());
		if (left == prim.getVoid()) throw new GetTypeException("The expression " + expToString(r.getLeft())+ " is of type void" +
				" and so cannot be used with " + r.getOp(), r.getLeft());
		visit(r.getRight());
		Type right = getResult();
		if (right == prim.getVoid()) throw new GetTypeException("The expression " + expToString(r.getRight())+ " is of type void" +
				" and so cannot be used with " + r.getOp(), r.getRight());

		if (prim.isSubType(left,right) || prim.isSubType(right, left)){
			// For <,<=,>,>=, we impose the type to be int, double
			RelOps op = r.getOp();
			if (op == RelOps.GT || op == RelOps.GEQ || op == RelOps.LT || op == RelOps.LEQ) {
				if (left == prim.getInt() || left == prim.getDouble() || left == prim.getUnlimitedNatural())
					set(prim.getBoolean());
				else 
					throw new GetTypeException("The operation \"" + r + "\" is not allowed over the type " + (left == null ? "left "+expToString(r) : left.getQualifiedName()), r);
			} else { // op = "=" || "!="
				set(prim.getBoolean());
			}
		}
		else 
			throw new GetTypeException("The types of each side of " + expToString(r) + " should be compatible ("+left.getQualifiedName()+" / "+right.getQualifiedName()+")", r);
	}

	protected void visit(LOGOP r) throws GetTypeException{
		Type left = run(r.getLeft());
		if (left == prim.getVoid()) throw new GetTypeException("The expression " + expToString(r.getLeft())+ " is of type void " +
				" and so cannot be used with " + r.getOp(), r.getLeft());
		visit(r.getRight());
		Type right = getResult();
		if (right == prim.getVoid()) throw new GetTypeException("The expression " + expToString(r.getRight())+ " is of type void " +
				" and so cannot be used with " + r.getOp(), r.getRight());
		if (left == prim.getBoolean() && right == prim.getBoolean()){
			set(prim.getBoolean());
		}
		else 
			throw new GetTypeException("Logical operations only allowed over boolean expressions", r);
		//			throw new TypingException("the types should be the same");
	}

	protected void visit(AOP r) throws GetTypeException{
		Type left = run(r.getLeft());
		if (left == prim.getVoid()) throw new GetTypeException("The expression " + expToString(r.getLeft())+ " is of type void " +
				" and so cannot be used with " + r.getOp(), r.getLeft());
		set(null);
		Type right = run(r.getRight());
		if (right == prim.getVoid()) throw new GetTypeException("The expression " + expToString(r.getRight())+ " is of type void " +
				" and so cannot be used with " + r.getOp(), r.getRight());
		// TODO: Cleanup. Should use something like Integral in uml model?
		if ((left == prim.getInt() || left == prim.getUnlimitedNatural()) && (right == prim.getInt() || right == prim.getUnlimitedNatural()))
			set(prim.getInt()); // XXX Nat?
		else if ((left == prim.getInt() || left == prim.getDouble()  || left == prim.getUnlimitedNatural()) && (right == prim.getDouble() || right == prim.getInt() || right == prim.getUnlimitedNatural()))
			set(prim.getDouble());
		else if ((left == prim.getString() && right == prim.getString()))
			set(prim.getString());
		else if (left == null ||  right == null)
			throw new GetTypeException("The operation \"" + r.getOp() + "\" cannot take null as an operand ", r);
		else
			throw new GetTypeException("The operation \"" + r.getOp() + "\" is not allowed over the types " + left.getQualifiedName()+" / "+right.getQualifiedName(), r);
	}

	@Override
	protected void visit(CAST c) throws GetTypeException {
		Type t1 = checkType(c.getType());
		// Only to check that the type of the expression is correct
		visit(c.getVal());
		set(t1);
	}

	@Override
	protected void visit(INSTANCEOF d) throws GetTypeException {
		run(d.getRef());
		run(d.getClass_());
		set(prim.getBoolean());
		assert state != null;
	}

	@Override
	protected void visit(NEG d) throws GetTypeException {
		visit(d.getE()); // ...or set to bool directly.
	}

	@Override
	protected void visit(CNAME c) throws GetTypeException {
		Class res;
		set(res = th.getModel().getClass(c.getName(), co));
		if (res == null)
			throw new GetTypeException("Type '"+c.getName()+"' not found." + (co == null? "" : co.getQualifiedName()) , c);
	}

	@Override
	protected void visit(VARNAME var) throws GetTypeException {
		set(vdecl.find(var));
		assert state != null;
	}

	@Override
	protected void visit(PREDICATE d) throws GetTypeException {
		if (d instanceof ITERATOR) {
			visit((ITERATOR)d);
			return;
		}
		checkVarname(d.getVar());
		// TODO: run on d.getType() twice?
		Vardecl v = vdecl.new Vardecl(d);
		tv.run(d.getType());
		Type e = run(d.getE());
		if (!(PrimitiveTypes.isBoolean(e))) {
			throw new GetTypeException("Quantified expression must be of type boolean!", d);
		}
		if (! queryOps.isEmpty())
			throw new UnpureException(print(queryOps),d);
		set(prim.getBoolean());
		vdecl.remove(v);
	}

	@Override
	protected void visit(ITERATOR d) throws GetTypeException {
		checkVarname(d.getVar());
		// TODO: run on d.getType() twice?
		tv.run(d.getType());
		Type vt = vdecl.getType(d.getType());
		checkVarAgainstSet(d, vt, d.getRange());
		Vardecl v = vdecl.new Vardecl(d);
		if (!DesignUtility.isPure(d.getRange(),currentClass) || (d.getRange() instanceof DASHABLE && ((DASHABLE)d.getRange()).isDash()))
			throw new GetTypeException("Range expression must be side-effect free!", d);
		Type e = run(d.getE());
		if (!(PrimitiveTypes.isBoolean(e))) {
			throw new GetTypeException("Quantified expression must be of type boolean!", d);
		}
		if (! queryOps.isEmpty())
			throw new UnpureException(print(queryOps),d);
		set(prim.getBoolean());
		vdecl.remove(v);
	}

	private static String print(Collection<Operation> queryOps2) {
		StringBuffer b = new StringBuffer();
		for (Operation op : queryOps2) {
			b.append(op.getQualifiedName());
			b.append(' ');
		}
		return b.toString();
	}

	@SuppressWarnings("serial")
	public static class UnpureException extends GetTypeException {
		public UnpureException(String s, TOKEN t) {
			super("Unpure methods called: "+s,t);
		}
	}

	@Override
	protected void visit(EList<?> l) throws GetTypeException {
		assert false;
		throw new UnsupportedOperationException();
	}

	protected String expToString(EXP e) {
		return DesignUtility.expToString(currentClass,e);
	}

	/**
	 * Static helper, only to be called once from constructor.
	 */
	private static Component getScope(Classifier cl) throws GetTypeException {
		try {
			Component sc = Utility.getScope(cl);
			return sc;
		} catch (TypingException e) {
			throw new GetTypeException(e);
		}
	}
	
	protected void checkVarname(VARNAME var) throws GetTypeException {
		if (var.equals(Class_OCL.RESERVED_IDENTIFIER_THIS))
			throw new GetTypeException("Variable declaration with reserved indentifier 'this' not allowed.",var);
	}

	public void checkVarAgainstSet(TOKEN d, Type tv, EXP val) throws GetTypeException {
		Type tit = th.getType(val,currentClass);
		if ((!(tit instanceof Class)) || ((Class)tit).getTemplateBindings().isEmpty())
			throw new GetTypeException("Range must be a set!", d);
		Class cit = (Class)tit;
		if (cit.getTemplateBindings().size() != 1)
			throw new GetTypeException("Generic instance malformed?!", d);
		TemplateBinding tb = cit.getTemplateBindings().get(0);
		if (tb.getParameterSubstitutions().size() != 1)
			throw new GetTypeException("Generic instance malformed?!", d);
		ParameterableElement actual = tb.getParameterSubstitutions().get(0).getActual();
		if (!(actual instanceof Class) || !prim.isSubType((Class)actual, tv))
			throw new GetTypeException("Type of iteration variable "+tv.getName()+" not compatible with range expression!", d);
	}
}
