package edu.unu.iist.rcos.spec.antlr.visitor.check;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.uml2.uml.Association;
import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Property;
import org.eclipse.uml2.uml.Type;

import edu.unu.iist.rcos.profile.PrimitiveTypes;
import edu.unu.iist.rcos.spec.antlr.TypeHandler;
import edu.unu.iist.rcos.spec.antlr.Vardecls;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.CNAME;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.FA;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.PREFIX;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.REF;

public class ExpGetTypeNoSet extends ExpGetType {

	public ExpGetTypeNoSet(Classifier cl) throws GetTypeException {
		super(cl);
		// TODO Auto-generated constructor stub
	}
	
	public ExpGetTypeNoSet(Classifier cl, TypeHandler th, Vardecls vdecl) throws GetTypeException {
		super(cl,th,vdecl);
	}
	
	public ExpGetTypeNoSet(final Classifier cl, final TypeHandler th) throws GetTypeException {
		super(cl,th);
	}
	
	@Override
	protected void visit(FA f) throws GetTypeException{
		PREFIX pr = f.getOwner();
		Type t;
		if (pr instanceof CNAME) {
			t = getTypeFromString((CNAME) pr);
		} else 
			t = run((REF) pr);
		final String fieldname = f.getName().getName();

		//XXX Void type
		if (t == prim.getVoid()) throw new GetTypeException("The expression \"" + expToString(f.getOwner()) 
				+ "\" is of type void and so does not have any field \"" + fieldname + "\"", f);
		if (PrimitiveTypes.isRCOSPrimitive(t)) throw new GetTypeException("The expression \"" + expToString(f.getOwner()) 
				+ "\" is of type " + t.getName() + " and so does not have any field \"" + fieldname + "\"", f);

		EList<Property> attrs = new BasicEList<Property>();
		EList<Association> assocs = new BasicEList<Association>();
		if (t instanceof Classifier) {
			Classifier ct = (Classifier) t;
			attrs.addAll(ct.getAllAttributes());
			assocs.addAll(ct.getAssociations());
			for (Classifier cg : ct.getGenerals()) {
				assocs.addAll(cg.getAssociations());
			}
		} else {
			throw new IllegalArgumentException("Don't know how to handle "+t.getQualifiedName()+" ["+t.eClass().getName()+"]");
		}

		for (Association a : assocs){
			// XXX Only works for binary associations
			if (a.getMembers().size() != 2) throw new GetTypeException("The association " + a.getQualifiedName() + " is not binary",f);
			Property p1 = a.getMemberEnds().get(0);
			Property p2 = a.getMemberEnds().get(1);
			if (prim.isSubType(t, p1.getType()) && a.getNavigableOwnedEnds().contains(p2)) attrs.add(p2);
			if (prim.isSubType(t, p2.getType()) && a.getNavigableOwnedEnds().contains(p1)) attrs.add(p1);
		}

		for (Property p : attrs){
			if (p.getName().equals(fieldname)){				
				// just get type, do not do something more --LD, 2010-1-3				
				set(p.getType());
				return;
			}
		}
		// Field not found
		throw new GetTypeException("In the expression \"" + expToString(f) + "\", the sub-expression  \"" + expToString(f.getOwner()) + 
				"\" is of type " + t.getQualifiedName() + " which does not have an attribute \"" + fieldname + "\"", f);
	}	

}
