package edu.unu.iist.rcos.spec.antlr.visitor.requiredProtocol;

import java.util.*;

import org.antlr.runtime.RecognitionException;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.uml2.uml.CallEvent;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Pseudostate;
import org.eclipse.uml2.uml.Region;
import org.eclipse.uml2.uml.State;
import org.eclipse.uml2.uml.StateMachine;
import org.eclipse.uml2.uml.Transition;
import org.eclipse.uml2.uml.Trigger;
import org.eclipse.uml2.uml.UMLFactory;
import org.eclipse.uml2.uml.Vertex;

import edu.unu.iist.rcos.mm.util.Utility;
import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.profile.ContractInterface;
import edu.unu.iist.rcos.profile.RCOSModel;
import edu.unu.iist.rcos.spec.antlr.Parse;
import edu.unu.iist.rcos.spec.antlr.visitor.requiredProtocol.RequiredProtocolEXP.RequiredProtocolDesignVisitor;

public class RequiredProtocolStateMachine {

	private StateMachine provide = null;

	private Region region = null;

	private Map<Vertex,Vertex> stateMap = new HashMap<Vertex,Vertex>();

	public RequiredProtocolStateMachine() {
		super();
		//provide = (StateMachine) ci.getProtocol().getStatemachine();
		//assert provide.getRegions().size() == 1;
		//region = provide.getRegions().size() == 0 ? null : provide.getRegions()
	   //			.get(0);
	}

	public StateMachine initializeRequiredProtocolStateMachine(ContractInterface ci){
		provide = (StateMachine) ci.getProtocol().getStatemachine();
		StateMachine sm = UMLFactory.eINSTANCE.createStateMachine();
		region = sm.createRegion("MyRegion");
		assert provide.getRegions().size() == 1;
		/* Deep-copy provided SM into required SM first */
		List<Vertex> vertex = provide.getRegions().get(0).getSubvertices();
		List<Transition> transitions = provide.getRegions().get(0).getTransitions();
		for (Vertex v : vertex) {
			Vertex vv;
			region.getSubvertices().add(vv = (Vertex) EcoreUtil.copy(v));
			stateMap.put(v, vv);
		}
		for (Transition t : transitions) {
			Transition tNew = region.createTransition(t.getName());
			tNew.setSource(stateMap.get(t.getSource()));
			tNew.setTarget(stateMap.get(t.getTarget()));
			/* XXX Tricky, special case initial state/transition */
			if (t.getSource() instanceof Pseudostate)
				continue;
			assert t.getTriggers().size() == 1 : t ; // WF
			Trigger tr = t.getTriggers().get(0);
			Trigger trNew = tNew.createTrigger(tr.getName());
			CallEvent ev = (CallEvent) EcoreUtil.copy(tr.getEvent());
			trNew.setEvent(ev);
		}
		return provide;
	}
	
	// print out the states
	private void printState(State state) {
		List<Transition> unvisited = new ArrayList<Transition>();
		List<Transition> visited = new ArrayList<Transition>();
		unvisited.addAll(state.getIncomings());
		unvisited.addAll(state.getOutgoings());
		while (unvisited.size() != 0) {
			Transition current = unvisited.get(0);
			unvisited.remove(0);
			visit(current, unvisited, visited);
			if (!visited.contains(current))
				visited.add(current);
			Vertex source = current.getSource();
			for (Transition transition : source.getIncomings())
				if (!visited.contains(transition)
						&& !unvisited.contains(transition))
					unvisited.add(transition);
			for (Transition transition : source.getOutgoings())
				if (!visited.contains(transition)
						&& !unvisited.contains(transition))
					unvisited.add(transition);
			Vertex target = current.getTarget();
			for (Transition transition : target.getIncomings())
				if (!visited.contains(transition)
						&& !unvisited.contains(transition))
					unvisited.add(transition);
			for (Transition transition : target.getOutgoings())
				if (!visited.contains(transition)
						&& !unvisited.contains(transition))
					unvisited.add(transition);
		}
	}

	// print out the transitions
	private void visit(Transition current, List<Transition> unvisited,
			List<Transition> visited) {
		Vertex source = current.getSource(), target = current.getTarget();
		System.out.println(source + "-->" + target);
		Trigger trigger = current.getTriggers().size() == 0 ? null : current
				.getTriggers().get(0);
		CallEvent callEvent = trigger.getEvent() instanceof CallEvent ? (CallEvent) trigger
				.getEvent()
				: null;
		if (callEvent != null)
			System.out.println(callEvent.getOperation());
	}

	/**
	 * 
	 * @return requiredProtocolStateMachine
	 * @throws WellformedException
	 * @throws RecognitionException
	 * @throws RequiredProtocolVisitorException
	 */
	public StateMachine calculateRequiredStateMachine(Class c, RCOSModel model)
			throws RequiredProtocolVisitorException, RecognitionException,
			WellformedException {

		/* For each state in the provided SM, create a corresponding state in
		 * the required SM, and remember their relation in the map.
		 * Don't forget that we deep-copied everything first from provided to required SM.
		 * To avoid a ConcurrentModificationException, deep-copy the list as well.
		 */
		Collection<Transition> transitions = new ArrayList<Transition>(region.getTransitions());
		for (Transition t : transitions) {
			// create a new fragment state machine for the transition t
			RequiredProtocolDesignVisitor v = new RequiredProtocolEXP(
					(RCOSModel) model, c).new RequiredProtocolDesignVisitor();

			if (t.getTriggers().size() == 0) {
				assert t.getSource() instanceof Pseudostate;
				continue;
			}
			assert t.getTriggers().size() == 1;

			CallEvent ev = (CallEvent) t.getTriggers().get(0).getEvent();
			Operation op = ev.getOperation();
			assert op != null;
			
			// obtain the calculated fragment (design)

			final Operation iOp = Utility.implementsOperation(c,op);
			assert iOp != null : op.getQualifiedName();
			State frag = v.run(Parse.getDesign(iOp));
			//printState(frag);
			// combine the new fragment with the required protocol
			combine(frag, t.getSource(), t.getTarget(), t);
		}
		// return the required state machine
		return region.getStateMachine();
	}

	/**
	 * for each transition and state from the frag copy to the region
	 * 
	 * @param frag
	 * @param srcR
	 * @param tgtR
	 */
	private void combine(State frag, Vertex srcR, Vertex tgtR, Transition trans) {
		Collection<Vertex> vertices = getAllConnectedStates(frag);
		if (vertices.size() == 1)
			return;
		// for all states in the fragment do
		System.out.println(vertices.size());

		/* Patch up the provided transition:
		 * it has to point to the new, single beginning state.
		 * Also remember the ending state. We have to do it first, since the number of
		 * outgoing transitions CHANGES as we move them "from" (they haven't been really in there)
		 * into the required SM.
		 */
		Vertex finalState = null;
		boolean changed = false;
		for (Vertex s : vertices) {
			if (s.getIncomings().isEmpty()) {
				assert !changed : "More than one beginning state in fragment?!";
				changed = true;
				trans.setTarget(s);
			}
			if (s.getOutgoings().isEmpty()) {
				assert finalState == null : "More than one end state in fragment?!";
				finalState = s;
			}
		}

		for (Vertex s : vertices) {
			// for all transitions in the fragment do
			for (Transition t : s.getOutgoings()) {
				// it is the first state of the fragment
				// it is the final state of the fragment
				if (t.getTarget() == finalState) {
					t.setTarget(tgtR);
				}
				region.getTransitions().add(t);
			}
			region.getSubvertices().add(s);
		}
		// remove the transition from the provided protocol
		// [vs] 'region' is not the provided protocol, and we shouldn't be modifying it anyway.
//		region.getTransitions().remove(trans);
//		srcR.getOutgoings().remove(trans);
//		tgtR.getIncomings().remove(trans);
	}

	private Collection<Vertex> getAllConnectedStates(State frag) {
		Set<Vertex> result = new HashSet<Vertex>();
		Set<Vertex> unvisited = new HashSet<Vertex>();
		unvisited.add(frag);
		while (unvisited.size() != 0) {
			Vertex current = unvisited.iterator().next();
			unvisited.remove(current);
			result.add(current);
			for (Transition transition : current.getOutgoings()) {
				Vertex t = transition.getTarget();
				if (!result.contains(t))
					unvisited.add(t);
			}
		}
		return result;
	}

	// get the original provided state machine
	public StateMachine getProvide() {
		return provide;
	}

	// add state
	public void addState(Vertex e) {
		region.getSubvertices().add(e);
	}

	// add a transition
	public void addTransition(Transition t) {
		region.getTransitions().add((org.eclipse.uml2.uml.Transition) t);
	}

}
