package edu.unu.iist.rcos.spec.antlr.visitor.toUTF8;

import org.antlr.runtime.RecognitionException;
import org.eclipse.core.runtime.Assert;
import org.eclipse.emf.common.command.*;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.profile.RCOSModel;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import static edu.unu.iist.rcos.spec.antlr.Parse.*;

public class UTF8Converter {

	final private EditingDomain domain;
	final private EFRCOSRuleFeature invEd;
	final private EFRCOSRuleFeature guardEd;
	final private EFRCOSRuleFeature designEd;
	private final CompoundCommand cmds;
	private final RCOSModel rm;

	public UTF8Converter(final EditingDomain d, RCOSModel rm){
		assert(d != null);
		domain = d;
		invEd = new InvariantsED(domain);
		guardEd = new GuardsED(domain);
		designEd = new DesignsED(domain);
		this.rm = rm;
		// the name of the command will show up in the Edit->Undo menu:
		cmds = new CompoundCommand("convert model to UTF-8");
		cmds.append(IdentityCommand.INSTANCE);
	}

	public Command toUTF8() throws WellformedException {
		assert(rm != null);

		TreeIterator<EObject> tree = rm.eResource().getAllContents();
		while(tree.hasNext()) {
			EObject e = tree.next();
			if (e instanceof Operation) {
				// translate guard
				translateGuard((Operation)e);
				// translate design
				translateDesign((Operation)e);
			} else if (e instanceof Class) {
				// translate invariant
				translateInv((Class)e);
			} else {
				// NADA
			}
		}
		return cmds.unwrap();
	}

	interface TIf {
		String translate(String in) throws RecognitionException,WellformedException;
	}
	
	private void translate(TIf tif, EFRCOSRuleFeature rif, Namespace op) throws WellformedException {
		final  String x = rif.get(op);
		if (x != null) {
			try {
				cmds.append(rif.update(op,tif.translate(x)));
			} catch (RecognitionException e) {
				throw new WellformedException(e);
			}
		}
	}

	private void translateInv(final Class c) throws WellformedException {
		translate(new TIf(){ public String translate(String in) throws RecognitionException {return new UTF8ExpPrinter(rm,c).run(getExpFromString(in),null); }},
				invEd, c);
	}

	private void translateGuard(final Operation op) throws WellformedException {
		translate(new TIf(){ public String translate(String in) throws RecognitionException,WellformedException {
			return new UTF8ExpPrinter(rm,op).run(getGuard(in,op),null);
		}},
		guardEd, op);
	}

	private void translateDesign(final Operation op) throws WellformedException {
		translate(new TIf(){ public String translate(String in) throws RecognitionException,WellformedException {
			DESIGN old;
			String result = new UTF8Printer(rm,op).run(old = getDesign(in,op),null);
			DESIGN n = getDesign(in,op);
			Assert.isTrue(DesignUtility.isDesignEqual(old, n));
			return result;
		}},
		designEd, op);
	}
}
