package edu.unu.iist.rcos.spec.antlr.visitor.transformation;

import java.util.List;

import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.DesignGenCopy;
import edu.unu.iist.rcos.spec.antlr.visitor.VisitorException;

import static edu.unu.iist.rcos.spec.antlr.MyDesign.*;

public class DesignAbstraction extends DesignGenCopy<VisitorException,ExpAbstraction> {
	
	public DesignAbstraction(List<String> lfa){
		super(new ExpAbstraction(lfa));
	}
	
	public DesignAbstraction(List<String> lfa, List<String> lvar, List<String> lpar) {
		super(new ExpAbstraction(lfa, lvar, lpar));
	}

	/**
	 * Renaming getCopy, for a better understandability of the code
	 * @author Charles Morisset
	 * @param e
	 * @return
	 * @throws VisitorException 
	 * @see getCopy
	 */
	
	EXP getAbstract(EXP e) throws VisitorException {
		return ev.run(e);
	}

	protected void visit(PP p) throws VisitorException{
		EXP pre = getAbstract(p.getPre());
		EXP post = getAbstract(p.getPost());

		if (pre == null){
			pre = createTrue();
		}
		if (post == null){
			post = createTrue();
		}
		set(createPP(pre, post));
	}
	
	@Override
	protected
	void visit (SEQ d) throws VisitorException{
		DESIGN left = run(d.getLeft());
		DESIGN right = run(d.getRight());
		if (left == null)
			left = createSkip();
	
		if (right == null)
			right = createSkip();
		set(createSeq(left, right));
	}	

	protected void visit (CHOICE d) throws VisitorException{
		DESIGN left = run(d.getLeft());
		DESIGN right = run(d.getRight());
		if (left == null)
			left = createSkip();
	
		if (right == null)
			right = createSkip();
		set(createChoice(left, right));
	}

	protected void visit (IF d) throws VisitorException{

		DESIGN left = run(d.getLeft());
		DESIGN right = run(d.getRight());
		if (left == null)
			left = createSkip();
	
		if (right == null)
			right = createSkip();
		
		EXP b = getAbstract(d.getB());
//		for (FA f : ev.listFA) System.err.println("Fields to keep " + f.getName().getName());
		if (b != null) {
			IF new_if = createIf(b, left, right);
//			System.err.println("Transforming '" + DesignUtility.designToString(d) + "' in '"+ DesignUtility.designToString(new_if) + "'");
			set(new_if);
		} else {
			set(createChoice(left, right));
		}
	}

	protected void visit (ASSIGN a) throws VisitorException{
		REF r = (REF) getAbstract(a.getVal());
		EXP e = getAbstract(a.getE());
		if (r == null || e == null)
			set(createSkip());
		else 
			set(createAssign(r, e));
	}
	
	@Override
	protected void visit(FOR f) throws VisitorException {
		EXP e = getAbstract(f.getVal());
		if (e == null) {
			set(createSkip());
		}
		else {
			DESIGN d1 = run(f.getDo());
			set(createFor(ev.run(f.getType()), 
					createVarname(f.getVar()), (VAL) e, d1));
		}
	}
	
}
