package edu.unu.iist.rcos.spec.antlr.visitor.transformation;

import org.eclipse.uml2.uml.Classifier;

import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import static edu.unu.iist.rcos.spec.antlr.MyDesign.*;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement.DesignRefinement;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement.RefinementException;

/**
 * @author charles
 *
 */
public class DesignSimplification extends DesignRefinement<RefinementException,ExpSimplification>{

	@Deprecated
	public DesignSimplification() {
		// we don't need a classifier:
		super(new ExpSimplification());
	}
	
	public DesignSimplification(Classifier cl) {
		super(new ExpSimplification(),cl);
	}

	protected void visit(SEQ s) throws RefinementException {
		DESIGN left = run(s.getLeft());
		DESIGN right = run(s.getRight());
		if (left instanceof SKIP)
			setModified(right);
		else if (left instanceof CHAOS)
			setModified(left);
		else if (left instanceof STOP)
			setModified(left);
		else if (right instanceof SKIP)
			setModified(left);
		else if(left instanceof SEQ) {
			/*           SEQ
			 *       SEQ      r2
			 *       l1 r1
			 *          ===>
			 *           SEQ
			 *         l1   SEQ
			 *             r1  r2
			 * We avoid creating new SEQ node by reusing the old one.
			 */
			SEQ leftS = (SEQ) left;
			s.setRight(leftS);
			s.setLeft(leftS.getLeft());
			leftS.setLeft(leftS.getRight());
			leftS.setRight(right);
			setModified(s); // restart
		} else if (left instanceof VARDECL && right instanceof ENDDECL) {
			VARDECL vd = (VARDECL) left;
			ENDDECL ed = (ENDDECL) right;
			if (vd.getVar().equals(ed.getVar()))
				setModified(createSkip());
			else super.visit(s);
		} else if (left instanceof VARDECL && right instanceof SEQ && ((SEQ) right).getLeft() instanceof ENDDECL) {
			VARDECL vd = (VARDECL) left;
			ENDDECL ed = (ENDDECL) ((SEQ) right).getLeft();
			if (vd.getVar().equals(ed.getVar()))
				setModified(((SEQ) right).getRight());
			else super.visit(s);
		} else if (left instanceof VARDECL && (right instanceof CHAOS || right instanceof STOP)) {
			// TODO: band-aid, extend in general for left.var unused.
			setModified(right);
		} else super.visit(s);
	}

	@Override
	protected void visit(CHOICE c) throws RefinementException {
		super.visit(c);
		DESIGN l = ((CHOICE)getResult()).getLeft();
		DESIGN r = ((CHOICE)getResult()).getRight();

		/* micro-optimization: faster tests first */
		if (l instanceof SKIP || r instanceof CHAOS)
			set(r);
		else if (l instanceof CHAOS || r instanceof SKIP)
			set(l);
		else
			// TODO: deeper nestings: CHOICE A (CHOICE A B)
			if (DesignUtility.isDesignEqual(l, r))
				set(l);
			else
			{ /* NADA */ }
	}

	protected void visit(PP pp) throws RefinementException {
		/* Call the super visitor and inspect result. */
		/* TODO: check the other classes for the same super.visit() calls. */
		super.visit(pp);
		pp = (PP) getResult(); /* cast is safe */
		EXP pre = pp.getPre();
		EXP post = pp.getPost();
		if (isTrue(pre) && isTrue(post))
			setModified(createSkip());
		else if (isFalse(pre))
			setModified(createChaos());
		else if (isFalse(post))
			setModified(createMagic());
		else
		{ /* NADA */ }
	}

	protected void visit(IF i) throws RefinementException {
		super.visit(i);
		IF j = (IF)getResult();
		if (isTrue(j.getB())) {
			setModified(j.getLeft());
		} else if (isFalse(j.getB())) {
			setModified(j.getRight());
		} else if (j.getLeft() instanceof SKIP &&
				j.getRight() instanceof SKIP) {
			setModified(j.getLeft());
		}
	}
}
