package edu.unu.iist.rcos.spec.antlr.visitor.transformation;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;

import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import static edu.unu.iist.rcos.spec.antlr.MyDesign.*;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.*;

public class ExpAbstraction extends ExpTypeVisit<VisitorException, TypeCopy<VisitorException>> {

	final List<VAR> listVar;
	final List<FA> listFA;
	final List<PAR> listPar;
	
	/**
	 * Renaming getCopy, for a better understandability of the code
	 * @author Charles Morisset
	 * @param e
	 * @return
	 * @throws VisitorException 
	 * @see getCopy
	 */

	public ExpAbstraction(List<String> lfa){
		super(new TypeCopy<VisitorException>());
		listVar = new ArrayList<VAR>();
		listFA = new ArrayList<FA>();
		listPar = new ArrayList<PAR>();
		if (lfa != null)
			for (String name : lfa ){
				FA f = SyntaxTree_ClassModelFactory.eINSTANCE.createFA(name, SyntaxTree_ClassModelFactory.eINSTANCE.createTHIS());
				listFA.add(f);
			}
	}

	// Field accesses and variables
	public ExpAbstraction(List<String> lfa, List<String> lvar){
		this(lfa);
		if (lvar != null)
			for (String name : lvar ){
				VAR v = SyntaxTree_ClassModelFactory.eINSTANCE.createVAR(name);
				listVar.add(v);
			}
	}

	/**
	 * Field accesses and variables and parameters
	 * @param lfa
	 * @param lvar
	 * @param lpar
	 */
	public ExpAbstraction(List<String> lfa, List<String> lvar, List<String> lpar){
		this(lfa, lvar);
		if (lpar != null)
			for (String name : lpar ){
				PAR v = SyntaxTree_ClassModelFactory.eINSTANCE.createPAR(name);
				listPar.add(v);
			}
	}

	private boolean keepVal(VAL v){
		if (v instanceof VAR) {
			for (VAR x : listVar){
				if (x.getName().equals(((VAR) v).getName()))
					return true;				
			}
		} else if (v instanceof PAR) {
			for (PAR x : listPar){
				if (x.getName().equals(((PAR) v).getName()))
					return true;				
			}
		} else if (v instanceof FA) {
			for (FA x : listFA) {
				if (isExpWeakEqual(x, v))
					return true;
			}
		}
		return false;
	}

	@Override
	protected void visit(LOGOP e) throws VisitorException {
		EXP left = run(e.getLeft());
		if (left == null) left = createTrue();
		EXP right = run(e.getRight());
		if (right == null) right = createTrue();
		set(createLogOp(e.getOp(), left, right));
	}

	@Override
	protected void visit (AOP e) throws VisitorException{
		EXP left = run(e.getLeft());
		EXP right = run(e.getRight());
		
		if (left == null || right == null)
			set(null);
		else {
			set(createAOp(e.getOp(), left, right));
		}
	}

	@Override
	protected void visit (RELOP e) throws VisitorException{
		EXP left = run(e.getLeft());
		EXP right = run(e.getRight());
		if (left == null || right == null)
			set(null);
		else {
			set(createRelOp(e.getOp(), left, right));
		}
	}

	protected void visit(VAR d) throws VisitorException {
		if (keepVal(d)) {
			super.visit(d);
			set(state);
		} else {
			set(null);
		} 
	}

	protected void visit(PAR d) throws VisitorException{
		if (keepVal(d)) {
			super.visit(d);
			set(state);
		} else {
			set(null);
		} 
	}

	protected void visit(FA d) throws VisitorException{
		if (keepVal(d)) {
			super.visit(d);
			set(state);
		} else {
			set(null);
		} 
	}

	protected void visit(MCALL d) throws VisitorException{
		PREFIX r = d.getOwner();
		EList<EXP> in = d.getIn();
		EList<REF> out = d.getOut();
		if (r instanceof THIS || (r instanceof VAL && keepVal((VAL) r))) { 
			EList<EXP> new_in = new BasicEList<EXP>();
			EList<REF> new_out = new BasicEList<REF>();
			for (EXP e : in) {
				EXP e2 = run(e);
				if (e2 != null) 
					new_in.add(e2);
			}
			
			MCALL new_m = createMcall((REF) run(r), createVarname(d.getName()), new_in, null);
			set(new_m);
			// XXX Need to change the method call to directly put the corresponding
			// "type": VAR, PAR or FA
		} else 
			set(null);
	}

	/* 
	 * Types can't be empty in rCOS, so any forall T x :true can be removed.
	 */
	@Override
	protected void visit(PREDICATE r) throws VisitorException {
		EXP c = ((PREDICATE)r).getE();
		c = DesignUtility.simplifyExpression(c);
		/* reuse sub-expressions if possible */
		if (c instanceof BASEBOOL)
			set(c);
	}

	@Override
	protected void visit(NEW d) {
		set(null);
	}

	@Override
	protected void visit(VARNAME var) throws VisitorException {}
}
