package edu.unu.iist.rcos.spec.antlr.visitor.transformation.associationtorequiredinterface;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.spec.antlr.MyDesign;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement.*;

public class AssocToReqIfExpVisitor extends ExpRefinement<RefinementException> {
	final Property candidate;
	private String context;
	final String cname;

	public AssocToReqIfExpVisitor(Property candidate, String cname) {
		assert candidate != null;
		assert cname != null;
		this.cname = cname;
		this.candidate = candidate;
	}

	public void setExceptionContex(String ctx) {
		assert ctx != null;
		context=ctx; // for exceptions
	}

	public class AssocToReqIfDesignVisitor extends DesignRefinement<RefinementException, ExpRefinement<RefinementException>> {
		public AssocToReqIfDesignVisitor(Classifier cl) {
			super(AssocToReqIfExpVisitor.this, cl);
		}

		@Override protected void visit(MCALL mc) throws RefinementException {
			if (mc.getOwner() instanceof FA) {
				FA faOwner = (FA)mc.getOwner();
				if (faOwner.getOwner() instanceof THIS && checkName(faOwner)) {
					// poke:
					mc.setOwner(MyDesign.createCname(cname));
					set(mc);
					return;
				}
			}
			super.visit(mc);
		}
	}

	@Override protected void visit(MCALL mc) throws RefinementException {
		/* Only rewrite an MCALL when it is directly contained in a design, like
		 *    if (comp.m()) ...
		 * but not when it is in a compound expression, or an n-ary design.
		 * Well, I guess in a unary operator could also be okay, like
		 *    if (-comp.m()) ...
		 * or rather, when there is only a single call with a side-effect.
		 * But that needs work on the typechecker, so let's not worry about that now.
		 * (TODO: RCOSTOOL-266)
		 */
		final EObject eContainer = mc.eContainer();
		if (eContainer instanceof ASSIGN
				|| (eContainer instanceof IF && mc.eContainingFeature() == SyntaxTree_ClassModelPackage.eINSTANCE.getIF_B())
				|| (eContainer instanceof WHILE && mc.eContainingFeature() == SyntaxTree_ClassModelPackage.eINSTANCE.getWHILE_B())
				|| (eContainer instanceof FOR && mc.eContainingFeature() == SyntaxTree_ClassModelPackage.eINSTANCE.getFOR_Val())
		)
			// TODO: refactor with above...but how?
			if (mc.getOwner() instanceof FA) {
				FA faOwner = (FA)mc.getOwner();
				if (faOwner.getOwner() instanceof THIS && checkName(faOwner)) {
					// poke:
					mc.setOwner(MyDesign.createCname(cname));
					set(mc);
					return;
				}
			}
		super.visit(mc);
	}

	@Override protected void visit(FA fa) throws RefinementException {
		/* TODO: check earlier --- or could we run the transformation with a temporary name first,
		 * and patch in the real name later if it would succeed? */
		if (fa.getOwner() instanceof THIS && checkName(fa))
			throw new SideEffectException("Cannot convert field access in "+context+" to component-call.");
		else
			super.visit(fa);
	}

	protected boolean checkName(FA fa) {
		return fa.getName().equals(candidate.getName());
	}

	private class SideEffectException extends RefinementException {
		private static final long serialVersionUID = 1L;
		public SideEffectException(String msg) { super(msg +" (candidate: "+candidate.getName()+")"); }
	}
}