package edu.unu.iist.rcos.spec.antlr.visitor.transformation.expert;

import java.util.*;

import org.eclipse.emf.common.command.*;
import org.eclipse.emf.common.util.*;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.mm.util.WellformedException;
import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import edu.unu.iist.rcos.spec.antlr.FreshNameGenerator;
import static edu.unu.iist.rcos.spec.antlr.MyDesign.*;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.commands.CreateOperationCommand;
import edu.unu.iist.rcos.spec.antlr.commands.CreateOperationCommand.DesignProviderI;

import edu.unu.iist.rcos.spec.antlr.visitor.*;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

/**
 * @author Charles Morisset
 * This design visitor creates all the setters that are needed, but does NOT change the existing code. 
 * For every path.x := e, where type of path is C, the method createSetter returns a command that 
 * creates in C a setter for x. 
 * This class works together with DesignReplaceSetter, which assumes that the setters are created, through
 * the getRequired method of DesignTransformerEPSetED 
 *
 */
public class DesignCreateSetter extends 
DesignVisitor<CompoundCommand, EXP, TYPE, ExpertPatternException,TypeCopy<ExpertPatternException>, ExpCopy<ExpertPatternException>> {

	final protected EditingDomain domain;
	final protected EPUtility epu;

	public DesignCreateSetter(EditingDomain ed, Classifier cl) {
		super(new ExpCopy<ExpertPatternException>());
		state = new StrictCompoundCommand(){{ isPessimistic = true;}};
		state.append(IdentityCommand.INSTANCE);
		domain = ed;
		epu = new EPUtility(cl);
	}

	@Override
	protected void visit(ASSIGN a) throws ExpertPatternException {
		REF r = (REF) copyExpression(a.getVal());

		if (r instanceof FA && ((PATH) r).getOwner() instanceof REF) {
			REF o = (REF) ((FA) r).getOwner();
			if (! (o instanceof THIS)) {
				FA fa = (FA) r;
				fa.setOwner(createThis());
				state.append(createSetter(o, fa));
			} else {
				super.visit(a);
			}
		} else {
			super.visit(a);
		}
	}

	private Command createSetter(final REF owner, final FA fa) throws ExpertPatternException {
		return new CommandWrapper() {
			@Override public Command createCommand() {
				try {
					//XXX improve by defining set_x where x is defined -> need to check for the name conflict 
					// We first get the type of the owner
					final Class c = (Class) epu.getType(owner);

					// Then the type of the path, in the context of the type of the owner
					final Type t_path = epu.getType(fa, c, true);

					// If the assignment is well-typed, t_path shouldn't be void
					if (epu.isVoid(t_path)) // WF,TC
						throw new ExpertPatternException("The field access " + expToString(c,fa) + " is of type void");

					// We generate an input parameter, of the type of the path
					FreshNameGenerator pg = new FreshNameGenerator(epu,c);
					final String para = pg.get(t_path);

					// Creation of the lists containing the input parameter, to look for an already existing setter
					final EList<String> paras_list = new BasicEList<String>();
					final EList<Type> type_list = new BasicEList<Type>();
					paras_list.add(para);
					type_list.add(t_path);

					int num = 0;
					do {
						// The name of the method 
						String name = epu.createSetter(fa, num);

						// Then we check the existing methods of c
						Operation op1 = c.getOwnedOperation(name, paras_list, type_list);		
						if (op1 == null) { 
							// There is no such a method, so we create it

							final CreateOperationCommand cc = new CreateOperationCommand(domain, new LinkedList<Throwable>(), name, c) {
								@Override protected EList<String> getNames() {return paras_list;}
								@Override protected EList<Type> getTypes() {return type_list;}
								@Override public void execute() {
									super.execute();
									Parameter para2 = getCreatedOperation().getOwnedParameters().get(0);
									para2.setDirection(ParameterDirectionKind.IN_LITERAL);									
								}
							};

							PAR para_exp = createPar(para, epu.createTYPE(t_path));
							final ASSIGN a = createAssign(fa, para_exp);

							/* We have to wait for "op" to become contained in the model for the pretty printer to work properly:
							 * I wonder if there's some requirement on strictness here. [vs]
							 */
 							return cc.createDesign(new DesignProviderI() {
								public String getDesign(Operation op) {
									return designToString(epu,op,a);
								}
							});
						} else if (epu.isGeneratedSetter(op1, fa)) {
							// We do not need to create a new method, so we just say that there is no conflict
							return new IdentityCommand();
						}
						// There is already a method with this name, but it's not one we generated, so we 
						// redo the same process with a different name. 
						num++;
					} while (true);

				} catch (GetTypeException ex) {
					throw new WrappedException(ex);
				} catch (WellformedException we) {
					throw new WrappedException(we);
				} catch (ExpertPatternException e) {
					throw new WrappedException(e);
				}
			}
		};
	}

	@Override
	protected void visit(CHOKE d) throws ExpertPatternException {
		throw new UnsupportedOperationException();
	}
}
