package edu.unu.iist.rcos.spec.antlr.visitor.transformation.expert;

import java.util.List;

import org.antlr.runtime.RecognitionException;
import org.eclipse.emf.common.util.EList;
import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Parameter;
import org.eclipse.uml2.uml.Type;

import edu.unu.iist.rcos.mm.util.Designs;
import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.profile.PrimitiveTypes;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.FreshNameGenerator;
import edu.unu.iist.rcos.spec.antlr.MyDesign;
import edu.unu.iist.rcos.spec.antlr.Parse;
import edu.unu.iist.rcos.spec.antlr.TypeHandler;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

/**
 * @author Charles Morisset
 *
 */
public class EPUtility extends TypeHandler {
	
	public final Classifier cl;

	@Deprecated
	public EPUtility(Classifier c) {
		super(c); // TODO
		cl = c;
	}

	/**
	 * @author charles
	 * @param path an expression with the form foo(e1, e2, ..., en).bar(f1, f2, ..., fm), which is transformed to foo(x1, x2, ..., xn).bar(y1, y2, ..., ym)
	 * @param pg a parameter generator which associate to each ei (resp. fi) a formal parameter xi (resp yi) of type ti (resp ui)
	 * @param exp_in is transformed to exp_in @ [f1;f2;...;fm] @ [e1;e2;...;en]
	 * @param paras_list is transformed to paras_list @ [y1;y2;...;ym] @ [x1;x2;...;xn]
	 * @param type_list is transformed to type_list @ [u1;u2;...;um] @ [t1;t2;...;tn]
	 * @throws GetTypeException
	 * @throws ExpertPatternException 
	 */
	public void getPathParameters(REF path, FreshNameGenerator pg, List<EXP> exp_in, EList<String> paras_list, EList<Type> type_list)
	throws GetTypeException, ExpertPatternException {
		if (path instanceof MCALL){
			MCALL m_path = (MCALL) path;
			parameterAbstraction(m_path, pg, exp_in, paras_list, type_list);

			if (m_path.getOwner() instanceof MCALL) {
				MCALL m_path_owner = (MCALL) m_path.getOwner();
				parameterAbstraction(m_path_owner, pg, exp_in, paras_list, type_list);
			}
		}
	}

	/**
	 * @author charles
	 * @param path is an expression with the form foo(e1, e2, ..., en) and is transformed to foo(x1, x2, ..., xn)
	 * @param pg is a parameter generator which associate to each ei  a formal parameter xi of type ti
	 * @param exp_in is transformed to exp_in @  [e1;e2;...;en]
	 * @param paras_list is transformed to paras_list @  [x1;x2;...;xn]
	 * @param type_list is transformed to type_list @ [t1;t2;...;tn]
	 * @throws GetTypeException 
	 * @throws ExpertPatternException 
	 */
	private void parameterAbstraction(MCALL m_path, FreshNameGenerator pg, List<EXP> exp_in, EList<String> paras_list, EList<Type> type_list)
	throws GetTypeException, ExpertPatternException {
		for (int i = 0; i < m_path.getIn().size(); i++){
			EXP e = m_path.getIn().get(i);
			exp_in.add(e);
			Type t = getType(e, cl);
			type_list.add(t);
			String par = pg.get(t);
			paras_list.add(par);
			try {
				m_path.getIn().set(i, MyDesign.createPar(par, createTYPE(t)));
			} catch (WellformedException e1) {
				throw new ExpertPatternException(e1);
			} 
		}
	}

	/**
	 * @author charles
	 * @param op: a DesignOperation
	 * @param path: an expression
	 * @return true if the design of op is "return path"
	 * @throws VisitorException
	 */
	public boolean isGeneratedGetter(Operation op, REF path, boolean voidOK) throws ExpertPatternException {
		String msg = null;
		try {
			msg = op.getQualifiedName()+"\n"+Designs.iNSTANCE.get(op);
			DESIGN d = Parse.getDesign(op);
			if (PrimitiveTypes.isVoid(getType(path, op.getClass_(),voidOK))){
				// void method				
				return DesignUtility.isExpEqual((MCALL) path, (MCALL) d);
			} else {
				if (d instanceof RET){
					RET r = (RET) d;
					return DesignUtility.isExpEqual(path, r.getE());
				} else {
					return false;
				}
			}
		} catch (RecognitionException e) { throw new ExpertPatternException(msg,e); }
		catch (WellformedException e) { throw new ExpertPatternException(msg,e); }
		catch (GetTypeException e) { throw new ExpertPatternException(msg,e); }
	}

	/**
	 * @author Charles Morisset
	 * @param op an operation with a parameter x
	 * @param fa a field
	 * @return true if the design of op is "fa := x"
	 * @throws ExpertPatternException
	 */
	public boolean isGeneratedSetter(Operation op, FA fa) throws ExpertPatternException {
		if (op.getOwnedParameters().size() == 1) {
			Parameter para = op.getOwnedParameters().get(0);
			String para_name = para.getLabel();
			PAR par;
			try {
				par = MyDesign.createPar(para_name, createTYPE(para.getType()));
				ASSIGN a = MyDesign.createAssign(fa, par);
				DESIGN d = Parse.getDesign(op);
				return DesignUtility.isDesignEqual(d, a);
			} catch (RecognitionException e) {
				throw new ExpertPatternException(e);
			} catch (WellformedException e) {
				throw new ExpertPatternException(e);
			}
		}
		else return false;
	}
	
	public String createGetter(REF path, int num) throws ExpertPatternException {
		return createMethodNameEP(path, "get", num);
	}
	
	public String createSetter(REF path, int num) throws ExpertPatternException {
		return createMethodNameEP(path, "set", num);
	}

	/**
	 * @author charles
	 * @param path: an expression of the form x.foo(e1,e2).z
	 * @param num: a positive or null integer 
	 * @return the string "get_x_foo_z" @ (num * "_")
	 * @throws VisitorException
	 */
	private String createMethodNameEP(REF path, String prefix, int num) throws ExpertPatternException{
		String underscore = "";
		for (int i = 0; i < num; i++) underscore += "_";
		return prefix + createMethodNameEP(path, prefix) + underscore;
	}

	/**
	 * @author charles
	 * @param path: an expression of the form x.foo(e1,e2).z
	 * @return the string "get_x_foo_z"
	 * @throws VisitorException
	 */
	private String createMethodNameEP(REF path, String prefix) throws ExpertPatternException {
		if (path instanceof THIS)
			return "";
		else if (path instanceof PATH && ((PATH) path).getOwner() instanceof REF){
			REF o = (REF) ((PATH) path).getOwner();
			String str = ((PATH) path).getName().getName();
			int length = prefix.length() + 1;
			if (str.length() > length && str.subSequence(0, length).equals(prefix + "_")){
				str = str.substring(length);
			}
			return createMethodNameEP(o, prefix) + "_" + str; 
		} else {
			throw new ExpertPatternException("The path " + DesignUtility.expToString(cl,path) + " is not well formed");
		}
	}

	public Type getType(REF owner) throws GetTypeException {
		return getType(owner, cl);
	}
}
