package edu.unu.iist.rcos.spec.antlr.visitor.transformation.expert;

import java.util.*;

import org.eclipse.emf.common.command.*;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.WrappedException;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.*;
import org.eclipse.uml2.uml.Class;

import edu.unu.iist.rcos.profile.PrimitiveTypes;
import edu.unu.iist.rcos.spec.antlr.DesignUtility;
import edu.unu.iist.rcos.spec.antlr.MyDesign;
import edu.unu.iist.rcos.spec.antlr.FreshNameGenerator;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.commands.CreateOperationCommand;
import edu.unu.iist.rcos.spec.antlr.commands.CreateOperationCommand.DesignProviderI;
import edu.unu.iist.rcos.spec.antlr.visitor.ExpVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.TypeVisitor;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

/**
 * @author Charles Morisset 
 * @author Volker Stolz     
 *  Go through an expression, and returns a CompoundCommand containing all the necessary creations 
 *  of getters. This is step is required by the ExpReplaceGetter
 */
public class ExpCreateGetter extends 
ExpVisitor<CompoundCommand, Void, ExpertPatternException, TypeVisitor<Void, ExpertPatternException>> {

	final protected EditingDomain domain;
	final protected EPUtility epu;
	private boolean voidOk = false;

	protected ExpCreateGetter(ExpCreateGetter ecg, boolean voidOk) {
		this(ecg.domain,ecg.epu.cl);
		this.voidOk = voidOk;
	}

	public ExpCreateGetter(EditingDomain ed, Classifier c){
		super(new TypeVisitor<Void, ExpertPatternException>());
		if (ed == null)
			throw new IllegalArgumentException();
		if (c == null)
			throw new IllegalArgumentException();
		domain = ed;
		epu = new EPUtility(c);
		state = new StrictCompoundCommand(){{ isPessimistic = true;}};
		state.append(IdentityCommand.INSTANCE);
		init = true;
	}

	@Override public void set(CompoundCommand e) {
		// Should never be called.
		throw new UnsupportedOperationException();
	}

	protected void visit(MCALL m) throws ExpertPatternException {
		visit((PATH) m);
	}

	protected void visit(FA m) throws ExpertPatternException{
		visit((PATH) m);
	}

	protected void visit(final PATH f) throws ExpertPatternException {

		if ((f.getOwner() instanceof REF)) {
			// If this is a method call, we have to make sure to call the expert pattern 
			// also for the arguments of the call

			if (f instanceof MCALL){
				MCALL m = (MCALL) f;

				for (EXP e : m.getIn()){
					ExpCreateGetter ecg = new ExpCreateGetter(this,false);
					for (Command c : ecg.run(e).getCommandList())
						state.append(c);
				}
			}
			PATH f_copy = (PATH) DesignUtility.copyExpression(f);
			REF r = (REF) f_copy.getOwner();
			if (r instanceof PATH && ((PATH) r).getOwner() instanceof REF){
				PATH o = (PATH) r;
				REF o2 = (REF) o.getOwner();
				if (! (o2 instanceof THIS)){
					o.setOwner(MyDesign.createThis());
					state.append(createGetter(o2, f_copy));
				} 
			}
		}
	}

	/**
	 * @author Charles Morisset
	 * @param owner an object different from THIS
	 * @param path a dotted expression x.y, where x and y can be field or method calls 
	 * @return the command creating the method get_x_y_*(in) (the number of underscores depends on the conflicts
	 * with existing methods or IdentityCommand if the getter already exists 
	 * @throws ExpertPatternException
	 */
	private Command createGetter(final REF owner, final REF path) {
		return new CommandWrapper() {
			@Override public Command createCommand() {
				try {
					// We first get the type of the owner
					final Class c = (Class) epu.getType(owner);
					assert(c != null);

					// We need to collect all the input parameters of the path, to propagate them.
					// exp_in contains the concrete arguments, which will be passed to the method call
					List<EXP> exp_in = new ArrayList<EXP>();

					// paras_list contains the formal parameters created for the method definition, using the 
					// parameter generator pg (see below).
					final EList<String> paras_list = new BasicEList<String>();
					//type_list[i] corresponds to the type of paras_list[i]
					final EList<Type> type_list = new BasicEList<Type>();

					final FreshNameGenerator pg = new FreshNameGenerator(epu,c);
					epu.getPathParameters(path, pg, exp_in, paras_list, type_list);	

					// [vs] Call the typechecker INSIDE the command to get up-to-date result
					// We get the type of the path
					final Type t_path = epu.getType(path, c, ExpCreateGetter.this.voidOk);
					// If the path is not void, then we need to add a return parameter
					final boolean tPathIsVoid = PrimitiveTypes.isVoid(t_path);
					if (! tPathIsVoid) {
						paras_list.add("ret");
						type_list.add(t_path);
					}

					// As long as we have a conflict, ie a method with the same signature but a different method body
					// we try we another name
					int num = 0;
					do {
						// The name of the method 
						final String name = epu.createGetter(path, num);

						// Then we check the existing methods of c
						Operation op1 = c.getOwnedOperation(name, paras_list, type_list);
						if (op1 == null) { 
							// There is no such a method, so we create it

							final CreateOperationCommand cc = new CreateOperationCommand(domain, new LinkedList<Throwable>() /* TODO */, name, c) {
								@Override protected EList<String> getNames() {return paras_list;}
								@Override protected EList<Type> getTypes() {return type_list;}
								@Override public void execute() {
									super.execute();
									Operation op = getCreatedOperation();
									assert op != null;
									// we can get away with creating the designs directly since no one is watching 'op' yet!
									if (! tPathIsVoid){
										// If the type is not void
										Parameter p = op.getOwnedParameter("ret", t_path);
										if (p != null) p.setDirection(ParameterDirectionKind.OUT_LITERAL);
									}
									op.setIsQuery(DesignUtility.isPureSafe(path, c, voidOk));
								}
							};
							return cc.createDesign(new DesignProviderI() {
								public String getDesign(Operation op) {
									String pathS = DesignUtility.expToString(op,path);
									final String arg;
									if (tPathIsVoid)
										arg = pathS;
									else
										arg = "return " + pathS;
									return arg;
								}});
						} else if (epu.isGeneratedGetter(op1, path,voidOk)) {
							// We do not need to create a new method, so we just say that there is no conflict
							return new IdentityCommand();
						} else {
							// NADA.
							// There is already a method with this name, but it's not one we generated, so we 
							// redo the same process with a different name.
							num++;
						}
					} while (true); // The loop is directly exited through return statements
				} catch (GetTypeException e) {
					throw new WrappedException(e);
				} catch (ExpertPatternException e) {
					throw new WrappedException(e);
				}
			}
		};}

	@Override protected void visit(CNAME c) throws ExpertPatternException {}
	@Override protected void visit(VARNAME var) throws ExpertPatternException {}
	@Override protected void visit(BASEBOOL d) throws ExpertPatternException {}
	@Override protected void visit(BASEINT d) throws ExpertPatternException {}
	@Override protected void visit(BASESTRING d) throws ExpertPatternException {}
	@Override protected void visit(BASEDOUBLE d) throws ExpertPatternException {}
	@Override protected void visit(BASEENUM d) throws ExpertPatternException {}
}
