package edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement;

import java.util.Collections;
import java.util.List;

import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.WrappedException;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Parameter;
import org.eclipse.uml2.uml.ParameterDirectionKind;
import org.eclipse.uml2.uml.Type;
import org.eclipse.uml2.uml.UMLFactory;

import edu.unu.iist.rcos.mm.util.Designs;
import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.profile.PrimitiveTypes;
import static edu.unu.iist.rcos.spec.antlr.DesignUtility.*;
import edu.unu.iist.rcos.spec.antlr.FreshNameGenerator;
import static edu.unu.iist.rcos.spec.antlr.MyDesign.*;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.ASSIGN;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.EXP;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.FA;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.MCALL;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.PAR;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.PATH;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.REF;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.THIS;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.expert.EPUtility;
import edu.unu.iist.rcos.spec.antlr.visitor.transformation.expert.ExpertPatternException;

/**
 * @author Charles Morisset
 * This class replaces all the path.x := e by path.set_x(e). The setters should be 
 * created first using DesignCreateSetter, otherwise an exception is thrown
 *
 */
public class DesignReplaceSetter extends DesignRefinement<ExpertPatternException, ExpRefinement<ExpertPatternException>> {

	final protected EPUtility epu;

	public DesignReplaceSetter(Classifier cl) {
		super(new ExpRefinement<ExpertPatternException>(), cl);
		epu = new EPUtility(cl);
	}

	@Override
	protected void visit(ASSIGN a) throws ExpertPatternException {
		REF r = (REF) copyExpression(a.getVal());

		if (r instanceof FA && ((PATH) r).getOwner() instanceof REF) {
			REF o = (REF) ((FA) r).getOwner();
			if (! (o instanceof THIS)) {
				FA fa = (FA) r;
				fa.setOwner(createThis());
				MCALL m = getMCallEPAssign(o, fa, copyExpression(a.getE()));
				setModified(m);
				
			} else {
				super.visit(a);
			}
		} else {
			super.visit(a);
		}
	}

	private MCALL getMCallEPAssign(REF owner, final FA fa, EXP e) throws ExpertPatternException {
		try {
			//XXX improve by defining set_x where x is defined -> need to check for the name conflict 
			// We first get the type of the owner
			final Class c = (Class) epu.getType(owner);

			// Then the type of the path, in the context of the type of the owner
			final Type t_path = epu.getType(fa, c);

			// If the assignment is well-typed, t_path shouldn't be void
			if (PrimitiveTypes.isVoid(t_path)) // WF,TC
				throw new ExpertPatternException("The field access " + expToString(c,fa) + " is of type void");

			// We generate an input parameter, of the type of the path
			FreshNameGenerator pg = new FreshNameGenerator(epu,c);
			final String para = pg.get(t_path);

			// Creation of the lists containing the input parameter, to look for an already existing setter
			final EList<String> paras_list = new BasicEList<String>();
			final EList<Type> type_list = new BasicEList<Type>();
			paras_list.add(para);
			type_list.add(t_path);

			// name is eventually the name of the correct setter, which is assumed to already exist
			String name = "";
			boolean conflict = false;
			
			int num = 0;
			do {
				try {
					// The name of the method 
					name = epu.createSetter(fa, num);

					// Then we check the existing methods of c
					Operation op1 = c.getOwnedOperation(name, paras_list, type_list);		
					if (op1 == null) { 
						throw new ExpertPatternException("The setter for '" + fa.getName().getName() + "' does not exist in '"
								+ c.getQualifiedName() + "'. Please create it first, for instance using DesignCreateSetter.");
					} else if (epu.isGeneratedSetter(op1, fa)) {
						// We do not need to create a new method, so we just say that there is no conflict
						conflict = false;
					}
				}
				catch (ExpertPatternException epe) {
					throw new WrappedException(epe);
				}
				// There is already a method with this name, but it's not one we generated, so we 
				// redo the same process with a different name. 
				num++;
			} while (conflict);

			// We finally create the method call which will replace the assignment
			List<EXP> list_in = Collections.singletonList(e);
			return createMcall(owner, name, list_in, null);
		} catch (GetTypeException ex) {
			throw new ExpertPatternException(ex);
		}
	}
}