package edu.unu.iist.rcos.spec.antlr.visitor.transformation.refinement;

import static edu.unu.iist.rcos.spec.antlr.MyDesign.*;
import org.eclipse.uml2.uml.Classifier;

import edu.unu.iist.rcos.mm.util.WellformedException;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.*;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

/**
 * @author Charles Morisset
 *  Transforms [p |- R] into [p |- true]; [true |- R]
 *  
 *  I have (Saeed) extended the code so that the design 
 *  [true |- r_1 && r_2 && ... && r_n], where each r_i is a RELOP 
 *  will be transformed to the design 
 *  [true |- p_k1 ; p_k2 ; ... ; p_kn], where each p_ki represents an r_i 
 */

public class DesignSplitPP extends DesignRefinement<RefinementException, ExpRefinement<RefinementException>> {

	final private EXPToComposition expToComp;

	public DesignSplitPP(Classifier aCl) {
		super(new ExpRefinement<RefinementException>(), aCl);
		expToComp = new EXPToComposition(currentClass);
	}

	protected void visit(PP d) throws RefinementException{ 		
		try {
			if ((! isTrue(d.getPre())) && (! isTrue(d.getPost()))) {
				EXP pre = getCopy(d.getPre());//It  is important to get a copy
				EXP post = getCopy(d.getPost());//It  is important to get a copy
				DESIGN d1 = createPP(pre, createTrue());
				DESIGN d2;
				if (post instanceof LOGOP){
					d2 = expToComp.postToSEQ((LOGOP)post);
				}else{
					d2 = createPP(createTrue(), post);
				}
				setModified(createSeq(d1, d2));
				return;
			};
			if (! isTrue(d.getPost())) {
				// Disable for now, not sure what's going wrong there.
				if (false && d.getPost() instanceof LOGOP){
					EXP post = getCopy(d.getPost());//It  is important to get a copy
					DESIGN d2 = expToComp.postToSEQ((LOGOP)post);
					if (expToComp.isSuccess()){
						setModified(d2);
						return;
					}
				}
			};
			super.visit(d);
		} catch (GetTypeException e) {
			throw new RefinementException(e);
		} catch (WellformedException e) {
			throw new RefinementException(e);
		}
	}
}
