package edu.unu.iist.rcos.spec.tc;

import org.antlr.runtime.RecognitionException;
import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Parameter;
import org.eclipse.uml2.uml.ParameterDirectionKind;

import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.spec.antlr.Parse;
import edu.unu.iist.rcos.spec.antlr.Vardecls;
import edu.unu.iist.rcos.spec.antlr.VardeclsFIS;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.EXP;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.PAR;
import edu.unu.iist.rcos.spec.antlr.visitor.check.GetTypeException;

public class RuleTCMembers {

	protected VardeclsFIS mkVardecls(Operation me) {
		return new VardeclsFIS(me);
	}

	public class RuleClassInvariant extends RuleTC<Classifier,Void,Invariants,EXP> {
		public RuleClassInvariant() {
			super(Invariants.iNSTANCE);
		}
		@Override
		public Classifier getContext(Classifier me) { return me; }
		@Override
		public void check(Classifier me, String doc) throws RecognitionException, WellformedException {
			checkBool(me,doc);
		}
		@Override
		public void check(Classifier me, EXP e) throws WellformedException {
			checkBool(me,e);
		}
		@Override
		protected VardeclsFIS mkVardecls(Classifier me) {
			return new VardeclsFIS(me);
		}
		@Override
		protected void checkPAR(Classifier me, Vardecls vs, PAR par) {
			// NADA
		};
	}

	public class RuleOperationDesign extends RuleTC<Operation,Void,Designs,DESIGN> {
		public RuleOperationDesign() {
			super(Designs.iNSTANCE);
		}
		@Override
		public Classifier getContext(Operation op) {
			Classifier ctx = (Classifier)op.getOwner();
			if (ctx == null)
				throw new IllegalArgumentException(op + " does not have an owner!");
			return ctx;
		}
		@Override
		public void check(Operation me, String doc) throws RecognitionException, WellformedException { checkDesign(me,doc); }
		@Override
		public void check(Operation me, DESIGN t) throws TypingException {
			checkDesign(me, t);
		}
		@Override
		protected VardeclsFIS mkVardecls(Operation me) {
			return RuleTCMembers.this.mkVardecls(me);
		}
		@Override
		protected void checkPAR(Operation me, Vardecls vs, PAR par) {
			// NADA
		}
	}

	public class RuleOperationGuard extends RuleTC<Operation,Void,Guards,EXP> {

		public RuleOperationGuard() {
			super(Guards.iNSTANCE);
		}

		@Override
		public Classifier getContext(Operation op) {
			Classifier ctx = (Classifier)op.getOwner();
			if (ctx == null)
				throw new IllegalArgumentException(op + " does not have an owner!");
			return ctx;
		}

		@Override
		public void check(Operation me, String doc) throws RecognitionException, WellformedException {
			checkBool(me,doc);
		}
		
		@Override
		protected void checkBool(Operation op, String doc) throws RecognitionException, WellformedException {
			// special version that pulls in the arguments from 'op' for type checking
			EXP e = Parse.getGuard(doc, op);
			check(op,e);
		}

		/**
		 * HERE BE DRAGONS -- you need to worry about the parameters of op!
		 * @throws WellformedException 
		 * @see edu.unu.iist.rcos.spec.tc.RuleTC#check(Namespace, TOKEN)
		 */
		@Override
		public void check(Operation me, EXP t) throws WellformedException {
			checkBool(me, t);
		}

		@Override
		protected VardeclsFIS mkVardecls(Operation me) {
			return RuleTCMembers.this.mkVardecls(me);
		}

		@Override
		protected void checkPAR(Operation me, Vardecls vs, PAR par)
				throws WellformedException {
			RuleTCMembers.checkPAR(me, vs, par);
		}
	}

	/**
	 * Static helper to check for output parameters in guards.
	 */
	public static void checkPAR(Operation me, Vardecls vs, PAR par) throws WellformedException {
		try {
			Parameter p = me.getOwnedParameter(par.getName().getName(), vs.getType(par.getType()));
			if (!p.getDirection().equals(ParameterDirectionKind.IN_LITERAL))
				throw new TypingException("Can't use output parameter '"+p.getName()+"' in a guard!");
		} catch (GetTypeException e) {
			throw new TypingException(e);
		}
	}
}
