package edu.unu.iist.rcos.spec.tc;

import org.antlr.runtime.RecognitionException;
import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.Classifier;
import org.eclipse.uml2.uml.Operation;

import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.spec.antlr.Parse;
import edu.unu.iist.rcos.spec.antlr.Vardecls;
import edu.unu.iist.rcos.spec.antlr.VardeclsFIS;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.DESIGN;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.EXP;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.PAR;

public class RuleTCMembersED {

	/* TODO: unify with RuleTCMembers, but that requires some refactoring higher up as well,
	* since e.g. Invariants and InvariantsED must have a common ancestor
	* 	IInvariants extends RCOSRuleFeature
	* You can get a default EditingDomain from EFRCOSRuleFeature.
	*/
	final protected EditingDomain d;
	
	public RuleTCMembersED(EditingDomain d) {
		this.d = d;
	}

	public class RuleClassInvariant extends RuleTC<Classifier,Command,InvariantsED,EXP> {
		public RuleClassInvariant() {
			super(new InvariantsED(d));
		}
		@Override
		public Classifier getContext(Classifier me) { return me; }
		@Override
		public void check(Classifier me, String doc) throws RecognitionException, WellformedException {
			checkBool(me,doc);
		}
		@Override
		public void check(Classifier me, EXP e) throws WellformedException {
			checkBool(me,e);
		}
		@Override
		protected VardeclsFIS mkVardecls(Classifier me) {
			return new VardeclsFIS(me);
		}
		@Override
		protected void checkPAR(Classifier me, Vardecls vs, PAR par) {
			// NADA
		};
	}

	public class RuleOperationDesign extends RuleTC<Operation,Command,DesignsED,DESIGN> {
		public RuleOperationDesign() {
			super(new DesignsED(d));
		}
		@Override
		public Classifier getContext(Operation op) {
			Classifier ctx = (Classifier)op.getOwner();
			if (ctx == null)
				throw new IllegalArgumentException(op + " does not have an owner!");
			return ctx;
		}
		@Override
		public void check(Operation me, String doc) throws RecognitionException, WellformedException { checkDesign(me,doc); }
		@Override
		public void check(Operation me, DESIGN t) throws TypingException {
			checkDesign(me, t);
		}
		@Override
		protected VardeclsFIS mkVardecls(Operation me) {
			return new VardeclsFIS(me);
		}
		@Override
		protected void checkPAR(Operation me, Vardecls vs, PAR par) {
			// NADA
		};
	}

	public class RuleOperationGuard extends RuleTC<Operation,Command,GuardsED,EXP> {

		public RuleOperationGuard() {
			super(new GuardsED(d));
		}

		@Override
		public Classifier getContext(Operation op) {
			Classifier ctx = (Classifier)op.getOwner();
			if (ctx == null)
				throw new IllegalArgumentException(op + " does not have an owner!");
			return ctx;
		}

		@Override
		public void check(Operation me, String doc) throws RecognitionException, WellformedException {
			checkBool(me,doc);
		}
		
		@Override
		protected void checkBool(Operation op, String doc) throws RecognitionException, WellformedException {
			// special version that pulls in the arguments from 'op' for type checking
			if (!doc.equals("")) { // ignore empty guards
				EXP e = Parse.getGuard(doc, op);
				check(op,e);
			}
		}

		/**
		 * HERE BE DRAGONS -- you need to worry about the parameters
		 * @throws WellformedException 
		 * @see edu.unu.iist.rcos.spec.tc.RuleTC#check(Namespace, TOKEN)
		 */
		@Override
		public void check(Operation me, EXP t) throws WellformedException {
			checkBool(me, t);
		}

		@Override
		protected VardeclsFIS mkVardecls(Operation me) {
			return new VardeclsFIS(me);
		}

		@Override
		protected void checkPAR(Operation me, Vardecls vs, PAR par)
				throws WellformedException {
			RuleTCMembers.checkPAR(me, vs, par);
		};
	}	
}
