package edu.unu.iist.rcos.spec.tc;

import java.util.List;

import org.antlr.runtime.RecognitionException;
import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.uml2.uml.*;

import edu.unu.iist.rcos.mm.ocl.StateMachine_OCL;
import edu.unu.iist.rcos.mm.util.*;
import edu.unu.iist.rcos.mm.util.statemachine.STMHelper;
import edu.unu.iist.rcos.profile.ContractInterface;
import edu.unu.iist.rcos.spec.antlr.Vardecls;
import edu.unu.iist.rcos.spec.antlr.VardeclsFIS;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.EXP;
import edu.unu.iist.rcos.spec.antlr.SyntaxTree.SyntaxTree_ClassModel.PAR;

public class RuleTransitionGuard extends RuleTC<Transition,Command,EFRCOSRuleFeature,EXP> {

	public RuleTransitionGuard(EditingDomain ed) {
		super(new GuardsED(ed));
	}

	@Override
	public Classifier getContext(Transition t) throws WellformedException {
		final Operation opIn = getOperation(t);
		List<ContractInterface> cis = STMHelper.getContractInterface(t.getContainer().getStateMachine());
		if (cis == null)
			throw new WellformedException("No contract interfaces for "+t.getQualifiedName());
		final Interface i;
		final Operation op;
		if (cis.size() == 1)
			op = Utility.implementsOperation(i = cis.get(0).getBase_Interface(), opIn);
		else
			throw new WellformedException("I'm confused.");
		if (op == null)
			throw new WellformedException("Can't find operation "+opIn.getName()+" in interface "+i.getQualifiedName());
		Classifier ctx = (Classifier)op.getOwner();
		assert ctx != null;
		return ctx;
	}

	private Operation getOperation(Transition t) throws WellformedException {
		if (!StateMachine_OCL.REGION_FORALL_EXACTLY_ONE_TRIGGER_R.check(t))
				throw new WellformedException(t,StateMachine_OCL.REGION_FORALL_EXACTLY_ONE_TRIGGER_R);
		CallEvent ev = (CallEvent)t.getTrigger(null).getEvent();
		if (ev == null)
			throw new WellformedException("Trigger must have a CallEvent!"); // WF
		final Operation opIn = ev.getOperation();
		if (opIn == null)
			throw new WellformedException(t,StateMachine_OCL.REGION_FORALL_CALLEVENT_EXACTLY_ONE_OPERATION_R);
		return opIn;
	}

	@Override
	protected VardeclsFIS mkVardecls(Transition me) throws WellformedException {
		return new VardeclsFIS(getOperation(me));
	}

	@Override
	public void check(Transition me, String doc) throws RecognitionException, WellformedException { checkBool(me,doc); }

	@Override
	public void check(Transition me, EXP t) throws WellformedException {
		checkBool(me, t);
	}

	@Override
	protected void checkPAR(Transition me, Vardecls vs, PAR par)
			throws WellformedException {
		RuleTCMembers.checkPAR(getOperation(me), vs, par);
	}
}